/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <tulip/PropertyAlgorithm.h>
#include <tulip/BooleanProperty.h>
#include <tulip/StringProperty.h>
#include <tulip/ColorProperty.h>
#include <tulip/LayoutProperty.h>
#include <tulip/SizeProperty.h>

#include <portgraph/PortGraph.h>
#include <portgraph/PortGraphRule.h>
#include <portgraph/PortGraphModel.h>
#include <portgraph/Bridge.h>
#include <portgraph/Trace.h>
#include <portgraph/PorgyTlpGraphStructure.h>

//#define NDEBUG

using namespace std;
using namespace tlp;

const static std::string SAVEID="saveId";

static node findNode(const int n, IntegerProperty *prop, Graph *g) {
    node n2;
    forEach(n2, g->getNodes()) {
        if(prop->getNodeValue(n2)==n)
            break;
    }
    return n2;
}

class ApplyRule:public tlp::Algorithm {

public:

    PLUGININFORMATION(PorgyConstants::APPLY_RULE, "Bruno Pinaud", "09/04/10", "Comments", "1.0.1", PorgyConstants::CATEGORY_NAME_INTERNAL)
    ApplyRule(const PluginContext *context):Algorithm(context) {
        addDependency(PorgyConstants::REDRAW_GRAPH, "1.0");
        addDependency(PorgyConstants::REDRAW_TRACE, "1.0");

        addInParameter<bool>(PorgyConstants::applyConnectedComponentPacking, "", "true");
        addInParameter<BooleanProperty>(PorgyConstants::POSITION, "", "", false);
        addInParameter<BooleanProperty>(PorgyConstants::BAN,"", "", false);
        addInParameter<PorgyConstants::state>(PorgyConstants::isCondition, "", "false");
        addInParameter<bool>(PorgyConstants::layoutNewModel, "", "false");

        addOutParameter<Graph *>(PorgyConstants::newData, "","",false);
    }

    bool check(string &errorMsg) {
        return (PorgyTlpGraphStructure::isPorgyGraph(graph, errorMsg)&&PorgyTlpGraphStructure::isModelGraph(graph->getSuperGraph()));
    }

    ///////////////
    //graph => l'instance du membre gauche de la règle à appliquer
    bool run() {
        pluginProgress->showPreview(false);
        pluginProgress->setComment("rewriting graph " + graph->getSuperGraph()->getName());

        Graph *old_graph = graph->getSuperGraph();

        //Est-ce que cette instance de la règle a déjà été appliquée avec les mêmes paramètres
        Trace traceobj(PorgyTlpGraphStructure::getMainTrace(graph));
        string new_model_name = traceobj.existEdge(old_graph->getName(), graph->getName());
        if(!new_model_name.empty()) {
            Graph *newG = PorgyTlpGraphStructure::findModel(graph, new_model_name);
            dataSet->set(PorgyConstants::newData, newG);
            cerr << __PRETTY_FUNCTION__ << ": Rule already applied before!" << endl;
            return true;
        }
        BooleanProperty *Pos=nullptr;
        BooleanProperty *Ban=nullptr;
        PorgyConstants::state st=PorgyConstants::state::OTHER;
        if(dataSet!=nullptr) {
            dataSet->get<BooleanProperty *>(PorgyConstants::POSITION, Pos);
            dataSet->get<BooleanProperty *>(PorgyConstants::BAN, Ban);
            dataSet->get(PorgyConstants::isCondition, st);
        }
        Graph *new_graph = PorgyTlpGraphStructure::clone(old_graph, (Pos!=nullptr)?Pos->getName():"", (Ban!=nullptr)?Ban->getName():"");
        Graph *_rule = PorgyTlpGraphStructure::findRule(graph,PorgyTlpGraphStructure::getRuleNameForSubGraph(graph));
        assert(_rule!=nullptr);
        //the saveId property is used to save the id of the RHS nodes to have a link between RHS and the graph being rewritten
        Observable::holdObservers();
        IntegerProperty *saveId=_rule->getLocalProperty<IntegerProperty>(SAVEID);
        node n;
        forEach(n, _rule->getNodes()) {
            saveId->setNodeValue(n, n.id);
        }
        edge e;
        forEach(e, _rule->getEdges()) {
            saveId->setEdgeValue(e, e.id);
        }

        //copying rule RHS to new_graph
        BooleanProperty newelt(_rule);
        newelt.setAllNodeValue(false);
        newelt.setAllEdgeValue(false);
        //select all elements of RHS
        PortGraphRuleDecorator dec(_rule);
        forEach(n, _rule->getNodes()) {
            if(dec.getSide(n) == PorgyConstants::SIDE_RIGHT)
                newelt.setNodeValue(n,true);
        }
        forEach(e, _rule->getEdges()) {
            if(dec.getSide(e) == PorgyConstants::SIDE_RIGHT) {
                newelt.setEdgeValue(e,true);
            }
        }

        BooleanProperty *new_strategy_application = new_graph->getLocalProperty<BooleanProperty>(PorgyConstants::NEW_STRATEGY_APPLICATION);
        BooleanProperty *newmodel_newelt = new_graph->getLocalProperty<BooleanProperty>(PorgyConstants::NEW);
        //copy RHS to the new graph
        tlp::copyToGraph(new_graph, _rule, &newelt, newmodel_newelt);
        //restore interface
        PortGraphRule PGMrule(_rule);
        IntegerProperty *oldId = new_graph->getLocalProperty<IntegerProperty>(SAVEID);
        IntegerProperty *tag_prop = graph->getProperty<IntegerProperty>(PorgyConstants::TAG);
        PortGraphModel lhs_instance_portgraph(graph);
        set<edge> to_restore; //use a set to avoid duplication of edges if both source and target are in the LHS instance
        Iterator<PortNode *> *it=lhs_instance_portgraph.getPortNodes();
        while(it->hasNext()) {
            PortNode *pn = it->next();
            Iterator<Port::PortPtr> *it2 = pn->getPorts();
            while(it2->hasNext()) {
                Port::PortPtr p2 = it2->next();
                node port_rule_left(tag_prop->getNodeValue(p2->getNode()));
                //check if the tag is valid (it should be a node in the rule)
                assert(_rule->isElement(port_rule_left));

                //copying properties of ports where the *_isused properties are set to false in RHS
                //RHS has been copied to the new graph. So all RHS properties are present in new_graph
                vector<string> matchingPropertiesList;
                if (graph->getRoot()->attributeExist(PorgyConstants::MATCHING_PROPERTIES)) {
                    graph->getRoot()->getAttribute<vector<string> >(PorgyConstants::MATCHING_PROPERTIES, matchingPropertiesList);
                    node right_equiv_rule(PGMrule.findRightEquiv(port_rule_left));
                    if(right_equiv_rule.isValid()) {
                        node right_equiv_model = findNode(right_equiv_rule,oldId, new_graph);
                        for(const auto& propertyName: matchingPropertiesList) {
                            BooleanProperty *prop_isused = new_graph->getProperty<BooleanProperty>(PorgyConstants::MATCHING_PROPERTIES_PREFIX + propertyName + PorgyConstants::MATCHING_PROPERTIES_SUFFIX);
                            if(prop_isused->getNodeValue(right_equiv_model)==false) {
                                PropertyInterface *prop = graph->getProperty(propertyName);
                                PropertyInterface *prop_new = new_graph->getProperty(propertyName);
                                prop_new->copy(right_equiv_model, p2->getNode(),prop);
                            }
                        }
                    }
                }

                //looking for edges
                //if the difference between the degree of the LHS instance nodes and the model nodes is greater than 0, there is edges to save
                if(old_graph->deg(p2->getNode())-graph->deg(p2->getNode())>0) {
//#ifndef NDEBUG
//                    cerr << "There are edges connected to port " << p2->getNode() << " in the LHS instance interface to preserve" << endl;
//#endif
                    Iterator<edge> *itp = old_graph->getInOutEdges(p2->getNode());
                    while(itp->hasNext()) {
                        edge e(itp->next());
                        //we do not want edges between center and port and edges present in the LHS instance
                        const pair<node, node> &ends = old_graph->ends(e);
                        if(!(PortNode::isCenter(ends.first, old_graph)||PortNode::isCenter(ends.second, old_graph)||graph->isElement(e))) {
                            to_restore.insert(e);
                        }
                    }
                    delete itp;
                }
            }
            delete it2;
        }
        delete it;

        //restore edges
        for(const auto& e: to_restore) {
#ifndef NDEBUG
            cerr << "restoring LHS interface for edge " << e << endl;
#endif
            const pair<node, node> &ends = old_graph->ends(e);
            node src(ends.first), dst(ends.second);
            bool green_edge=false;
            if(graph->isElement(src)) {
                if(graph->isElement(dst)) {
//#ifndef NDEBUG
//                    cerr << "edge (both source and target) is in the LHS instance but it is not part of the LHS instance" << endl;
//#endif
                    src = findNode(PGMrule.findRightEquiv(node(tag_prop->getNodeValue(src))), oldId, new_graph);
                    dst = findNode(PGMrule.findRightEquiv(node(tag_prop->getNodeValue(dst))), oldId, new_graph);
                }
                else {
//#ifndef NDEBUG
//                    cerr << "Only edge source is in the LHS instance" << endl;
//#endif
                    node tmp_node(tag_prop->getNodeValue(src));
                    node n1(PGMrule.findRightEquiv(tmp_node));
                    if(dec.getSide(n1) == PorgyConstants::SIDE_LEFT) {
                        cerr << "green edge detected" << endl;
                        green_edge=true;
                        src = findNode(n1, tag_prop, graph);
                        forEach(n, old_graph->getInOutNodes(src)) {
                            if(PortBase::isPort(n, old_graph)) {
                                src = n;
                                break;
                            }
                        }
                        if(new_graph->hasEdge(src, dst, false)) {
                            cerr << "edge already restored, skipping..." << endl;
                            continue;
                        }
                    }
                    else
                        src = findNode(n1,oldId, new_graph);
                }
            }
            else {
//#ifndef NDEBUG
//                cerr << "Only edge target is in the LHS instance " << endl;
//#endif
                node lhs_equivalent(tag_prop->getNodeValue(dst));
                node n1(PGMrule.findRightEquiv(lhs_equivalent));
                if(dec.getSide(n1) == PorgyConstants::SIDE_LEFT) {
                    cerr << "green edge detected" << endl;
                    green_edge=true;
                    dst = findNode(n1,tag_prop, graph);
                    node n;
                    forEach(n, old_graph->getInOutNodes(dst)) {
                        if(PortBase::isPort(n, old_graph)) {
                            dst = n;
                            break;
                        }
                    }
                    if(new_graph->hasEdge(src, dst, false)) {
                        cerr << "edge already restored, skipping..." << endl;
                        continue;
                    }
                }
                else
                    dst = findNode(n1,oldId, new_graph);
            }
            edge new_e = new_graph->addEdge(src, dst);
#ifndef NDEBUG
            cerr << "Adding edge " << new_e << " between " << src << " and " << dst << endl;
#endif
            if(green_edge) {
                newmodel_newelt->setEdgeValue(new_e, true);
                green_edge=false;
            }
            else {
                PropertyInterface *prop;
                forEach(prop, new_graph->getObjectProperties()) {
                    prop->copy(new_e,e,prop);
                }
            }
        }

        //handle M and N
        //if the M property is true, put elements inside the current P property
        //if the N property is true, put elements inside the current Ban property
        PortGraphModel pgm(new_graph);
        BooleanProperty *newP=nullptr, *newM=nullptr, *newN=nullptr, *newBan=nullptr;
        if(Pos!=nullptr) {
            newP = new_graph->getProperty<BooleanProperty>(Pos->getName());
            newM = new_graph->getProperty<BooleanProperty>(PorgyConstants::M);
        }
        if(Ban!=nullptr) {
            newN = new_graph->getProperty<BooleanProperty>(PorgyConstants::N);
            newBan = new_graph->getProperty<BooleanProperty>(Ban->getName());
        }

        forEach(e, pgm.getEdges()) {
            if(Pos!=nullptr)
                newP->copy(e, e, newM);
            if(Ban!=nullptr)
                newBan->copy(e, e, newN);
            if(newmodel_newelt->getEdgeValue(e) == true)
                new_strategy_application->setEdgeValue(e, true);
        }
        PortNode *pn=nullptr;
        forEach(pn, pgm.getPortNodes()) {
            if(Pos!=nullptr&&newM->getNodeValue(pn->getCenter())) {
                pn->select(true, newP);
            }
            if(Ban!=nullptr&&newN->getNodeValue(pn->getCenter())) {
                pn->select(true, newBan);
            }
            if(newmodel_newelt->getNodeValue(pn->getCenter()))
                pn->select(true, new_strategy_application);
        }

        set<node> setnewnode;
        forEach(n, newmodel_newelt->getNodesEqualTo(true)) {
            if(PortNode::isCenter(n, new_graph))
                setnewnode.insert(n);
        }

        LayoutProperty *layout_old_graph = graph->getProperty<LayoutProperty>("viewLayout");
        SizeProperty *size_old_graph = graph->getProperty<SizeProperty>("viewSize");
        StringProperty *label_old_graph = graph->getProperty<StringProperty>("viewLabel");

        StringProperty *label_new_graph = new_graph->getProperty<StringProperty>("viewLabel");
        SizeProperty *size_new_graph = new_graph->getProperty<SizeProperty>("viewSize");
        LayoutProperty *layout_new_graph = new_graph->getProperty<LayoutProperty>("viewLayout");

        //instead of leaving new nodes at (0,0,0), put them where the old ones are if a LHS node linked to a RHS node
        BooleanProperty used(graph);
        used.setAllNodeValue(false);
        Iterator<PortNode *> *itpo = lhs_instance_portgraph.getPortNodes();
        while(itpo->hasNext()) {
            PortNode *p = itpo->next();  //old node
            Iterator<Port::PortPtr> *itpop = p->getPorts();
            while(itpop->hasNext()) {
                Port::PortPtr po = itpop->next(); //old port
                node p_rhs(PGMrule.findRightEquiv(node(tag_prop->getNodeValue(po->getNode()))));
                if(p_rhs.isValid()) {
                    const PortNode* pp = PGMrule[p_rhs];
                    node new_node(findNode(pp->getCenter(), oldId, new_graph));  //replacing node from RHS
                    layout_new_graph->setNodeValue(new_node, layout_old_graph->getNodeValue(p->getCenter()));
                    size_new_graph->setNodeValue(new_node, size_old_graph->getNodeValue(p->getCenter()));
                    used.setNodeValue(p->getCenter(), true);
                    setnewnode.erase(new_node);
                    break;
                }
            }
            delete itpop;
        }
        delete itpo;

        //LHS and RHS are not linked. Try to put each new node at the same position of an old node with the same label
        vector<set<node>::iterator> toremove;
        for(set<node>::iterator it=setnewnode.begin();it!=setnewnode.end();++it) {
            node no(*it);
            //looking for an old node with the same label
            stableForEach(n, used.getNodesEqualTo(false)) {
                if(PortGraphDecorator::isCenter(graph,n)) {
                    if(label_old_graph->getNodeValue(n)==label_new_graph->getNodeValue(no)) {
                        layout_new_graph->setNodeValue(no, layout_old_graph->getNodeValue(n));
                        size_new_graph->setNodeValue(no, size_old_graph->getNodeValue(n));
                        used.setNodeValue(n, true);
                        toremove.push_back(it);
                        break;
                    }
                }
            }
        }
        //to avoid invalidating iterators, setnewnode must be updated after iterating on it
        for(auto it: toremove)
                setnewnode.erase(it);
        toremove.clear();

        //for the remaining nodes, put them at the same position of an old node
        for(set<node>::iterator it=setnewnode.begin();it!=setnewnode.end();++it) {
            node no(*it);
            stableForEach(n, used.getNodesEqualTo(false)) {
                layout_new_graph->setNodeValue(no, layout_old_graph->getNodeValue(n));
                size_new_graph->setNodeValue(no, size_old_graph->getNodeValue(n));
                used.setNodeValue(n, true);
                toremove.push_back(it);
                break;
            }
        }
        for(auto it: toremove)
                setnewnode.erase(it);
        if(!setnewnode.empty())
            cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << ": TODO!! " << endl;

        //rule algorithm
        //Applying the algorithm
        std::string err("");
        // Get the algorithm
        std::string text("");
        _rule->getAttribute(PorgyConstants::RULE_ALGORITHM, text);

        BooleanProperty *lhs_rule_selection = graph->getLocalProperty<BooleanProperty>(PorgyConstants::NEW);
        lhs_rule_selection->setAllNodeValue(true);
        lhs_rule_selection->setAllEdgeValue(true);

        tlp::DataSet algoDataSet;
        algoDataSet.set<string>("Instructions",text);
        algoDataSet.set<string>("Mode", "execute");
        algoDataSet.set<tlp::Graph*>("Lhs", graph);
        algoDataSet.set<string>("RhsMappingProperty", SAVEID);
        algoDataSet.set<string>("LhsMappingProperty", PorgyConstants::TAG);
        algoDataSet.set<string>("RhsSelectionProperty", PorgyConstants::NEW);
        algoDataSet.set<string>("LhsSelectionProperty", PorgyConstants::NEW);

        if (!new_graph->applyAlgorithm(PorgyConstants::APPLY_RULE_ALGORITHM, err, &algoDataSet))
            tlp::warning() << "Unable to parse the algorithm: " << err << " - step ignored" << std::endl;

        //removing remaining nodes from the LHS instance (edges will be removed automatically)
        Iterator<node> *itn = graph->getNodes();
        new_graph->delNodes(itn);
        delete itn;
        Observable::unholdObservers();
        //nettoyage
        //new_graph->delLocalProperty(SAVEID);
        _rule->delLocalProperty(SAVEID);
        if(new_graph->existLocalProperty(PorgyConstants::ARITY))
            new_graph->delLocalProperty(PorgyConstants::ARITY);
        if(new_graph->existLocalProperty(PorgyConstants::M))
            new_graph->delLocalProperty(PorgyConstants::M);
        if(new_graph->existLocalProperty(PorgyConstants::N))
            new_graph->delLocalProperty(PorgyConstants::N);
        if(new_graph->existLocalProperty(PorgyConstants::RULE_SIDE))
            new_graph->delLocalProperty(PorgyConstants::RULE_SIDE);

        vector<string> matchingPropertiesList;
        if (graph->getRoot()->attributeExist(PorgyConstants::MATCHING_PROPERTIES)) {
            graph->getRoot()->getAttribute<vector<string> >(PorgyConstants::MATCHING_PROPERTIES, matchingPropertiesList);
            for(const auto& propertyName: matchingPropertiesList) {
                if(new_graph->existLocalProperty(PorgyConstants::MATCHING_PROPERTIES_PREFIX + propertyName + PorgyConstants::MATCHING_PROPERTIES_SUFFIX))
                    new_graph->delLocalProperty(PorgyConstants::MATCHING_PROPERTIES_PREFIX + propertyName + PorgyConstants::MATCHING_PROPERTIES_SUFFIX);
            }
        }

        if(st!=PorgyConstants::state::IF) {
            node tgt = traceobj.createMetaNode(new_graph);
//            if(!isProba)
                traceobj.addSimpleTransformationEdge(traceobj.findNode(old_graph->getName()),tgt,Pos,Ban,graph,_rule->getId());
//            else
//                traceobj.addProbabilisticTransformationEdge(traceobj.findNode(old_graph->getName()),tgt,Pos,Ban,graph,_rule->getId(), proba);

            bool draw_model=true;
            if(dataSet!=nullptr)
                dataSet->get<bool>(PorgyConstants::layoutNewModel, draw_model);
            if(draw_model){
                string errMsg;
                DataSet ds;
                ds.set("Layout", PGMrule.getLayoutAlgorithm());
                if(!new_graph->applyPropertyAlgorithm(PorgyConstants::REDRAW_GRAPH, new_graph->getProperty<LayoutProperty>("viewLayout"),errMsg,pluginProgress,&ds))
                    cerr << "Echec du layout du nouveau modèle : " << errMsg << endl;
            }
        }
#ifndef NDEBUG
        else
            cerr << __PRETTY_FUNCTION__ << ": if or while test condition detected => no update to the trace graph" << endl;
#endif
        dataSet->set(PorgyConstants::newData, new_graph);
        return true;
    }
};

PLUGIN(ApplyRule)
