/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "RunStrategyElement.h"
#include "UtreeWalk.h"

#include <tulip/StringCollection.h>
#include <tulip/BooleanProperty.h>
#include <tulip/DoubleProperty.h>
#include <tulip/IntegerProperty.h>

#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/Trace.h>

#include <numeric>
#include <typeinfo>

using namespace tlp;
using namespace std;
using namespace boost;

/////////////
static bool run_rule(const string &rulename, Graph *g, string& P, string& Ban, const PorgyConstants::state state, PluginProgress *pp, bool isAll, StrategyElement::vectNewGraphs& newGraphs, bool isMatch) {
  DataSet ds;
  ds.set(PorgyConstants::POSITION, g->getProperty<BooleanProperty>(P));
  ds.set(PorgyConstants::BAN, g->getProperty<BooleanProperty>(Ban));
  ds.set<string>(PorgyConstants::RuleName, rulename);
  ds.set<unsigned int>(PorgyConstants::minimum_number_of_instances, 1);
  ds.set<bool>(PorgyConstants::applyConnectedComponentPacking, false);
  ds.set<PorgyConstants::state>(PorgyConstants::isCondition, state);
  ds.set<bool>(PorgyConstants::checkRule, true);
  ds.set<int>(PorgyConstants::maximum_number_of_instances, isAll?(-1):1);
  bool ret(false);
  string errMsg;
  if(!isMatch) {
#ifndef NDEBUG
      cerr << "Running " << rulename << " on " << g->getName() << endl;
#endif
      ret = g->applyAlgorithm(PorgyConstants::CHECK_APPLY_RULE, errMsg, &ds, pp);
      ds.get(PorgyConstants::newModelAll,newGraphs);
    }
  else {
#ifndef NDEBUG
      cerr << "Try yo match " << rulename << " on " << g->getName() << endl;
#endif
      g->push();
      ds.set<int>(PorgyConstants::applyRule, false);
      if(g->applyAlgorithm(PorgyConstants::CHECK_APPLY_RULE, errMsg, &ds, pp)) {
          g->pop();
          ret=true;
          newGraphs.push_back(g);
        }
    }
  return ret;
}

//////////////////////////////////////
static bool run_parallel_rule(const StrategyElement::parallel_datastruct &rule, string& P, string& Ban, PluginProgress *pp, Graph *g, const PorgyConstants::state state, bool isAll, StrategyElement::vectNewGraphs& newGraphs, bool isMatch) {
    //fabrication de la règle parallèle
    string errMsg;
    vector<string> rulesv(get<1>(rule));
    rulesv.insert(rulesv.begin(),get<0>(rule));
    StringCollection rulesc(rulesv);
    DataSet ds;
    ds.set<StringCollection>("Rule string", rulesc);
    if(!g->applyAlgorithm(PorgyConstants::CREATE_PARALLEL_RULE, errMsg, &ds, pp)) {
        cerr << "Failed to create the parallel rule: " << errMsg << endl;
        pp->setError("error");
        return false;
    }
    string parallel_rulename;
    ds.get(PorgyConstants::PARALLEL_RULENAME, parallel_rulename);
    //exécution de la règle parallèle
    return run_rule(parallel_rulename, g, P, Ban, state, pp, isAll, newGraphs, isMatch);
}

unsigned runStrategyElement::step = 1;

//to run a strategy
bool runStrategyElement::run_strat(StrategyElement::stackStrat &s, const PorgyConstants::state &state, PluginProgress *pp, boost::optional<StrategyElement::vectNewGraphs& > res) {
    bool ret(false);
    while(!s.empty()) {
        ret=true;
        pp->setError("");
        StrategyElement::stratElem elt = s.top();
        s.pop();
        StrategyElement::vector_strategyelts cond(get<3>(elt));
        Graph *g = get<2>(elt);
        string P(get<0>(elt));
        string Ban(get<1>(elt));
        if(cond.empty()&&res) {
            (*res).push_back(g);
            continue;
        }
        for(StrategyElement::vector_strategyelts::iterator it=cond.begin();((it!=cond.end())&&(pp->state()==TLP_CONTINUE));++it) {
            const StrategyElement &elt = *it;
            StrategyElement::vector_strategyelts remaining_op(it+1, cond.end());
            bool doBreak(false);
            ++step;
            pp->setComment(string("Running strategy step ").append(to_string(step)));
            StrategyElement::vectNewGraphs newGraphAll;
            ret = elt.run(pp, g, P, Ban, state, s, remaining_op, doBreak, newGraphAll)&&(pp->state()==TLP_CONTINUE);
            if(ret) {
                if(doBreak)
                    break;
                if(res&&(it+1==cond.end())) //when we reach the end of a strategy, keep a trace of its result for future branching (for repeat and orelse operators)
                    (*res).insert((*res).end(), newGraphAll.begin(), newGraphAll.end());
                if(newGraphAll.empty())
                    break;
                else if (newGraphAll.size()==1) {
                    g = newGraphAll[0];
                }
                else {  //a all(R) has been used
                    //associate for each graph the remaining operations to do
                    if(it+1==cond.end()) //if we reach the end of the strategy, there is no need to continue
                        break;
                    StrategyElement::vector_strategyelts v(it+1, cond.end()); //copy all remaining operations
                    for(auto& g: newGraphAll) {
                        s.push(StrategyElement::stratElem(P, Ban, g, v));
                    }
                    break;
                }
                if(pp->state()==TLP_STOP) {
                    return true;
                }
                if(pp->state()==TLP_CANCEL) {
                    //errorMsg = "Cancel button pressed";
                    return false;
                }

                if(pp->getError()=="error") {
                    cerr << "Error when running the strategy" << endl;
                    return false;
                }
            }
            else {
                break;
            }
        }
    }
    return ret;
}

/////////////
runStrategyElement::runStrategyElement(PluginProgress *pp, Graph *g, string& P, string& Ban, const PorgyConstants::state st,  StrategyElement::stackStrat &s, StrategyElement::vector_strategyelts &v, bool &d, StrategyElement::vectNewGraphs &newGraphAll):
    _pp(pp),_g(g),_P(P), _Ban(Ban), _state(st),_st(s), _r(v), _doBreak(d), _newGraphAll(newGraphAll){}
///////////////
class OneAllVisitor : public boost::static_visitor<bool> {
    PluginProgress *_pp;
    Graph *_g;
    string& _P;
    string& _Ban;
    const PorgyConstants::state &_state;
    StrategyElement::vectNewGraphs& _newGraphs;
    bool _isAll, _isMatch;

public:
    OneAllVisitor(PluginProgress *pp, Graph *g, string& P, string& Ban, const PorgyConstants::state& s, StrategyElement::rule_operators op, StrategyElement::vectNewGraphs& newGraphs):_pp(pp),_g(g),_P(P),
      _Ban(Ban),_state(s), _newGraphs(newGraphs), _isAll(false), _isMatch(false){
      switch(op) {
      case StrategyElement::rule_operators::ALL:
  #ifndef NDEBUG
          cerr << "__ALL__" << endl;
  #endif
          _isAll=true;
          _isMatch=false;
          break;
      case StrategyElement::rule_operators::ONE:
  #ifndef NDEBUG
          cerr << "__ONE__" << endl;
  #endif
          _isAll=false;
          _isMatch=false;
          break;
        case StrategyElement::rule_operators::MATCH:
    #ifndef NDEBUG
            cerr << "__MATCH__" << endl;
    #endif
            _isAll=false;
            _isMatch=true;
            break;
      }

    }

    //for running a simple rule
    bool operator() (const string &rulename) {
        return run_rule(rulename, _g, _P, _Ban, _state, _pp, _isAll, _newGraphs, _isMatch);
    }

    //for running rules in //
    bool operator() (const StrategyElement::parallel_datastruct &ds) {
        return run_parallel_rule(ds, _P, _Ban, _pp, _g, _state, _isAll, _newGraphs, _isMatch);
    }
};


bool runStrategyElement::operator()(const StrategyElement::one_all_match_datastruct &ds ) {

    OneAllVisitor visit(_pp, _g, _P, _Ban, _state, get<0>(ds), _newGraphAll);
    return get<1>(ds).apply_visitor(visit);
}

////////////
/// \brief runStrategyElement::operator ()
/// \param ifds
/// \return
bool runStrategyElement::operator()(const StrategyElement::if_datastruct &ifds) {
#ifndef NDEBUG
    cerr << "____IF____" << endl;
#endif
    //exécution de la condition du if
    _g->push(false);
    Graph *g=_g;
#ifndef NDEBUG
    cerr << "__IF__: test condition" << endl;
#endif
    StrategyElement::stackStrat s;
    StrategyElement::stratElem elt(_P, _Ban, g, get<0>(ifds));
    s.push(elt);
    bool cond_status=run_strat(s, PorgyConstants::state::IF, _pp);
    _pp->setError("");
    _g->pop(false);
    //la seule chose intéressante de la condition est le code de retour
    StrategyElement::vector_strategyelts v;
    if(cond_status) { //test de la condition
#ifndef NDEBUG
        cerr << "__IF__: then" << endl;
#endif
        v = get<1>(ifds);
    }
    else if(get<2>(ifds)) {  //Do we have an else?
#ifndef NDEBUG
        cerr << "__IF__: else" << endl;
#endif
        v = *get<2>(ifds);
    }
#ifndef NDEBUG
    else {
        cerr << "__IF__: no else" << endl;
    }
#endif
    v.insert(v.end(), _r.begin(), _r.end());
    elt = StrategyElement::stratElem(_P, _Ban, _g, v);
    _st.push(elt);
    _doBreak=true;
    return true;
}
//////////////////////////////////
bool runStrategyElement::operator ()(const StrategyElement::orelse_datastruct &ods) {
#ifndef NDEBUG
  cerr << "____ORELSE____" << endl;
#endif
  StrategyElement::vector_strategyelts v;
  StrategyElement::stratElem elt;
  //We try to execute the 1st strategy and keep a trace of the 2nd in case of a Fail on the 1st strategy
  _g->push(false);
  Graph *g = _g;
  elt = StrategyElement::stratElem(_P, _Ban, _g, get<0>(ods));
  StrategyElement::stackStrat s;
  s.push(elt);
  StrategyElement::vectNewGraphs newGraphAll;
  if(!run_strat(s, PorgyConstants::state::ORELSE, _pp, newGraphAll)) {
      _g->pop(false);
      _g = g;
      v = get<1>(ods);
      _doBreak=true;
      v.insert(v.end(), _r.begin(), _r.end());
      elt = StrategyElement::stratElem(_P, _Ban, _g, v);
      _st.push(elt);
    }
  else {
      v.insert(v.end(), _r.begin(), _r.end());
      if(newGraphAll.empty()) {
          elt = StrategyElement::stratElem(_P, _Ban, _g, v);
          _st.push(elt);
          _doBreak=true;
        }

      else if (newGraphAll.size()==1) {
          cerr << "graph : " << newGraphAll[0]->getName() << endl;
          elt = StrategyElement::stratElem(_P, _Ban, newGraphAll[0], v);
          _st.push(elt);
          _doBreak=true;
        }
      else {  //a all(R) has been used
          //associate for each graph the remaining operations to do
          for(auto& g: newGraphAll) {
              _st.push(StrategyElement::stratElem(_P, _Ban, g, v));
              _doBreak=true;
            }
        }
    }
  return true;
}

bool runStrategyElement::operator()(const StrategyElement::while_datastruct &wds) {
#ifndef NDEBUG
  cerr << "____WHILE____" << endl;  //while(S)[(max)]do(S1) => if(S)then(S1;while(S)[(max-1)]do(S1))else(Id)
#endif
  StrategyElement::stratElem elt;
  if(get<1>(wds)&&(*get<1>(wds)==0)) { //operation counter
      _doBreak=true;
      elt = StrategyElement::stratElem(_P, _Ban, _g, _r);
      _st.push(elt);
    }
  else {
      StrategyElement::vector_strategyelts v;
      StrategyElement::if_datastruct ifds;
      get<0>(ifds) = get<0>(wds);
      StrategyElement::vector_strategyelts v1(get<2>(wds));
      if(get<1>(wds)) {
          unsigned num = *get<1>(wds);
          --num;
          v1.push_back(StrategyElement(StrategyElement::while_datastruct(get<0>(wds), num, get<2>(wds))));
        }
      else {
          v1.push_back(StrategyElement(StrategyElement::while_datastruct(get<0>(wds), optional<unsigned>(), get<2>(wds))));
        }
      get<1>(ifds) = v1;
      v.push_back(StrategyElement(ifds));
      _doBreak=true;
      v.insert(v.end(), _r.begin(), _r.end());
      elt = StrategyElement::stratElem(_P, _Ban, _g, v);
      _st.push(elt);
    }
  return true;
}
//////////

unsigned runStrategyElement::random_number(const StrategyElement::ppick_datastruct &vect) const {
    vector<double> pr(vect.size());
    for(unsigned i=0;i<vect.size();++i) {
        pr[i] = vect[i].second;
    }
    discrete_distribution<unsigned> dist(pr.begin(), pr.end());
    return dist(PorgyTlpGraphStructure::gen);
}

bool runStrategyElement::operator()(const StrategyElement::ppick_datastruct &ds) {
#ifndef NDEBUG
    cerr << "__PPICK__" << endl;
#endif
    //compute a random number with the given probabilities then run the associated strategy
    unsigned rnd = random_number(ds);
//    _ds.set<double>(PorgyConstants::Proba, ds[rnd].second);
     StrategyElement::vector_strategyelts v(ds[rnd].first);
    _doBreak=true;
    v.insert(v.end(), _r.begin(), _r.end());
    StrategyElement::stratElem elt(_P, _Ban, _g, v);
    _st.push(elt);
    return true;
}

////////////
/// \brief runStrategyElement::operator () runs a repeat(S) which is equivalent to S;repeat(S)
/// execute S, then go back one step inside the remaining strategy operations (i.e. execute repeat(S) again.
/// \param ds
/// \return
///
bool runStrategyElement::operator()(const StrategyElement::repeat_ds &ds) {
#ifndef NDEBUG
    cerr << "____REPEAT____" << endl;
#endif
    if(_state!=PorgyConstants::state::OTHER)
        return true;
    //We try to execute the 1st strategy. If Id => do it again, else if Fail => remove the error and continue the strategy
    _g->push(false);
    Graph *g = _g;
    StrategyElement::stratElem elt(_P, _Ban, _g, get<0>(ds));
    StrategyElement::stackStrat s;
    s.push(elt);
    StrategyElement::vectNewGraphs newGraphAll;
    if(get<1>(ds)&&(*get<1>(ds)==0)) { //operation counter
        _g = g;
        _doBreak=true;
        elt = StrategyElement::stratElem(_P, _Ban, _g, _r);
        _st.push(elt);
    }
    else {
        if(!run_strat(s, PorgyConstants::state::REPEAT, _pp, newGraphAll)) {
            _g->pop(false);
            _g = g;
            _doBreak=true;
            elt = StrategyElement::stratElem(_P, _Ban, _g, _r);
            _st.push(elt);
        }
        else {
            StrategyElement::vector_strategyelts v;
            if(get<1>(ds)) {
              unsigned  num = *get<1>(ds);
                --num;
                v.push_back(StrategyElement(StrategyElement::repeat_ds(get<0>(ds), num)));
            }
            else {
                v.push_back(StrategyElement(StrategyElement::repeat_ds(get<0>(ds), boost::optional<unsigned>())));
            }
            v.insert(v.end(), _r.begin(), _r.end());
            if(newGraphAll.empty()) {
                elt = StrategyElement::stratElem(_P, _Ban, _g, v);
                _st.push(elt);
                _doBreak=true;
            }
            else if (newGraphAll.size()==1) {
                elt = StrategyElement::stratElem(_P, _Ban, newGraphAll[0], v);
                _st.push(elt);
                _doBreak=true;
            }
            else {  //a all(R) has been used
                //associate for each graph the remaining operations to do
                for(auto& g: newGraphAll) {
                    _st.push(StrategyElement::stratElem(_P, _Ban, g, v));
                }
                _doBreak=true;
            }
        }
    }
    return true;
}
/////////////
//ID/FAIL
bool runStrategyElement::operator()(const StrategyElement::idfail &pu) {
    bool ret=false;
    switch(pu) {
    case StrategyElement::idfail::ID:
#ifndef NDEBUG
        cerr << "____ID____" << endl;
#endif
        ret=true;
        break;
    case StrategyElement::idfail::FAIL:
    {
#ifndef NDEBUG
        cerr << "____FAIL____" << endl;
#endif
        Trace(PorgyTlpGraphStructure::getMainTrace(_g)).addFailureNode(_g->getName(),-1, nullptr,nullptr);
        ret=false;
        break;
    }
    }
    return ret;
}

static BooleanProperty* make_prop(Graph* new_model, const string &prop_name, booleanptr prop_new) {
    //resetting current value;
   BooleanProperty* b = new_model->getLocalProperty<BooleanProperty>(prop_name);
    node n;
    forEach(n, b->getNodesEqualTo(true)) {
        b->setNodeValue(n, false);
    }
    edge e;
    forEach(e, b->getEdgesEqualTo(true)) {
        b->setEdgeValue(e, false);
    }
    b->copy(prop_new.get());
    return b;
}

////////////
bool runStrategyElement::operator()(const StrategyElement::setPosBan_datastruct &fd) {  //setPos, setBan
    Graph *new_model=PorgyTlpGraphStructure::clone(_g, _P, _Ban);
    spirit::utree ut(get<1>(fd));
//#ifndef NDEBUG
//    cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << ": new graph clone => " << new_model->getName() << endl;
//    cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << ": property for Position: " << p->getName() << endl;
//    cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << ": property for Ban: " << ((ban!=nullptr)?ban->getName():"nullptr") << endl;
//    cerr << "utree: " << ut << endl;
//#endif

    PorgyConstants::strat_focus_op operation(get<0>(fd));
    vector<booleanptr> v;
    spirit::utree::visit(ut, utree_walk(v, _P, _Ban, new_model));
    assert(v.size()==1);    //should have only one result if the computation was successfull
    booleanptr prop_new(v[0]);
    v.clear();
    if(operation==PorgyConstants::strat_focus_op::SETPOS) {
#ifndef NDEBUG
        cerr << "____SETPOS____" << endl;
#endif
        make_prop(new_model, _P, prop_new);
    }
    else if(operation==PorgyConstants::strat_focus_op::SETBAN) {
#ifndef NDEBUG
        cerr << "____SETBAN____" << endl;
#endif
        if(_Ban.empty()) {
            _Ban="Ban";
            unsigned i=1;
            while(new_model->existProperty(_Ban)) {
                _Ban = "Ban"+to_string(i++);
            }
        }
        make_prop(new_model, _Ban, prop_new);

    }

    Trace traceobj(PorgyTlpGraphStructure::getMainTrace(new_model));
    node tgt = traceobj.createMetaNode(new_model);
    traceobj.addFocusTransformationEdge(traceobj.findNode(_g->getName()), tgt, prop_new.get(), _g, operation);
    _newGraphAll.push_back(new_model);
    return true;
}

////////////////////////
bool runStrategyElement::operator()(const StrategyElement::update_ds &fd) {
    string name(get<0>(fd).name);
#ifndef NDEBUG
    cerr << "__UPDATE__" << endl;
    cerr << "Lauching tulip plugin " << name << endl;
#endif
    DataSet parameters;
    Graph *new_model=PorgyTlpGraphStructure::clone(_g, _P, _Ban);
    // plugin parameters
    ParameterDescriptionList params = PluginLister::instance()->getPluginParameters(name);
    params.buildDefaultDataSet(parameters, new_model);
    if(get<1>(fd)) {
        for(const auto& p: *get<1>(fd)){
            ParameterDescription pa(get<0>(p));
            if(pa.getTypeName()==typeid(StringCollection).name()) {
                StringCollection scp;
                parameters.get(pa.getName(), scp);
                scp.setCurrent(get<1>(p));
                parameters.set(pa.getName(), scp);
            }
            else if(pa.getTypeName()==typeid(DoubleProperty).name()) {
                DoubleProperty *dp = new_model->getRoot()->getProperty<DoubleProperty>(get<1>(p));
                parameters.set(pa.getName(), dp);
            }
            else if(pa.getTypeName()==typeid(BooleanProperty).name()) {
                BooleanProperty *dp = new_model->getRoot()->getProperty<BooleanProperty>(get<1>(p));
                parameters.set(pa.getName(), dp);
            }
            else if(pa.getTypeName()==typeid(IntegerProperty).name()) {
                IntegerProperty *dp = new_model->getRoot()->getProperty<IntegerProperty>(get<1>(p));
                parameters.set(pa.getName(), dp);
            }
            else
                parameters.set(pa.getName(), get<1>(p));
        }
    }
    string err;
    PropertyInterface* prop=NULL;
    parameters.get("result", prop);
    bool ret =  new_model->applyPropertyAlgorithm(name, prop, err, _pp, &parameters);
    Trace traceobj(PorgyTlpGraphStructure::getMainTrace(new_model));
    node tgt = traceobj.createMetaNode(new_model);
    traceobj.addOtherEdge(traceobj.findNode(_g->getName()), tgt, Trace::Update_Transformation);
    _newGraphAll.push_back(new_model);
    return ret;
}

static bool check_equality(booleanptr prop1, booleanptr prop2) {
    bool ret(true);
    node n;
    edge e;
    //check if prop1 and prop2 are equals
    forEach(n, prop1->getNonDefaultValuatedNodes()) {
        if(prop2->getNodeValue(n)==prop1->getNodeDefaultValue()){
            ret=false;
            break;
        }
    }
    if(ret) {
        forEach(n, prop2->getNonDefaultValuatedNodes()) {
            if(prop2->getNodeValue(n)==prop1->getNodeDefaultValue()){
                ret=false;
                break;
            }
        }
        if(ret) {
            forEach(e, prop1->getNonDefaultValuatedEdges()) {
                if(prop2->getEdgeValue(e)==prop1->getEdgeDefaultValue()){
                    ret=false;
                    break;
                }
            }
            if(ret) {
                forEach(e, prop2->getNonDefaultValuatedEdges()) {
                    if(prop1->getEdgeValue(e)==prop1->getEdgeDefaultValue()){
                        ret=false;
                        break;
                    }
                }
            }
        }
    }
    return ret;
}


////////////////////////
bool runStrategyElement::operator()(const StrategyElement::comparison_datastruct &fd) {
#ifndef NDEBUG
    cerr << "__COMPARISON__" << endl;
#endif
    //get utrees
    spirit::utree ut1(get<0>(fd));
    spirit::utree ut2(get<2>(fd));
    vector<booleanptr> v1, v2;
    //execute both utrees and retrieve the resulting Boolean property
    spirit::utree::visit(ut1, utree_walk(v1, _P, _Ban, _g));
    spirit::utree::visit(ut2, utree_walk(v2, _P, _Ban, _g));
    assert((v1.size()==1)&&(v2.size()==1));
    booleanptr prop1=v1[0], prop2=v2[0];
    v1.clear();
    v2.clear();
    //prop1 and prop2 are the properties to compare.
    StrategyElement::comparison_operators op(get<1>(fd));
    bool correct(true);
    switch(op) {
    case StrategyElement::comparison_operators::EQUAL:
    {
#ifndef NDEBUG
        cerr << "__EQUAL__" << endl;
#endif
         correct = check_equality(prop1, prop2);
    }
        break;
    case StrategyElement::comparison_operators::SUBSET:
    {
#ifndef NDEBUG
        cerr << "__SUBSET__" << endl;
#endif
        node n;
        edge e;
        //check if prop1 is a subset of prop2
        forEach(n, prop1->getNonDefaultValuatedNodes()) {
            if(prop2->getNodeValue(n)==prop1->getNodeDefaultValue()){
                correct=false;
                break;
            }
        }
        if(correct) {
            forEach(e, prop1->getNonDefaultValuatedEdges()) {
                if(prop2->getEdgeValue(e)==prop1->getEdgeDefaultValue()){
                    correct=false;
                    break;
                }
            }
        }
    }
        break;
    case StrategyElement::comparison_operators::NEQ:
    {
#ifndef NDEBUG
        cerr << "__NOT_EQUAL__" << endl;
#endif
         correct = !check_equality(prop1, prop2);
    }
        break;
    }

    if(correct) {
        if(_state!=PorgyConstants::state::IF) {
            Graph *new_model=PorgyTlpGraphStructure::clone(_g, _P, _Ban);
            Trace traceobj(PorgyTlpGraphStructure::getMainTrace(new_model));
            node tgt = traceobj.createMetaNode(new_model);
            traceobj.addFocusTransformationEdge(traceobj.findNode(_g->getName()), tgt, nullptr, _g, PorgyConstants::strat_focus_op::COMPARISON);
            _newGraphAll.push_back(new_model);
        }
    }
    else {
        if(_state!=PorgyConstants::state::IF) {
            Trace(PorgyTlpGraphStructure::getMainTrace(_g)).addFailureNode(_g->getName(),-1, nullptr,nullptr);
        }
    }
    return correct;
}
