/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/Trace.h>

#include <tulip/Algorithm.h>
#include <tulip/BooleanProperty.h>

#define BOOST_SPIRIT_USE_PHOENIX_V3
#include "StrategyParser.h"
#include "RunStrategyElement.h"

using namespace std;
using namespace tlp;
using namespace boost;

class StrategySpirit:public tlp::Algorithm {

    std::string strategy;
    StrategyElement::vector_strategyelts strategy_elements;

    //find a local property in the while hierarchy
    bool findP(const string &property_name) const {
        Graph *models_root = PorgyTlpGraphStructure::getModelsRoot(graph);
        Graph *g;
        forEach(g, models_root->getSubGraphs()) {
            if(g->existLocalProperty(property_name))
                return true;
        }
        return false;
    }

    bool parse(string &errMsg) {

        vector<TulipProp> properties;
        vector<std::string> property_names;
        //built a vector with all properties and another one with only Boolean properties
          PropertyInterface *abstractp;
          forEach(abstractp, graph->getObjectProperties()) {
              TulipProp p;
              p.name=abstractp->getName();
              properties.push_back(p);
              property_names.push_back(abstractp->getName());
          }

          //built a vector with all rulenames
          vector<string> rules;
          Graph *rule;
          rules.reserve(PorgyTlpGraphStructure::getRulesRoot(graph)->numberOfSubGraphs());
          forEach(rule, PorgyTlpGraphStructure::getRulesRoot(graph)->getSubGraphs()) {
              rules.push_back(rule->getName());
          }

          //built a vector with all porgy plugins
          vector<string> porgy_plugins;
          vector<PluginName> porgyplugins;
          std::list<std::string> plugins = PluginLister::instance()->availablePlugins<PropertyAlgorithm>();
          for(const auto& name: plugins){
              Plugin* pl(PluginLister::getPluginObject(name, nullptr));
              if(pl->group()==PorgyConstants::CATEGORY_NAME) {
                  porgy_plugins.push_back(name);
                  PluginName p;
                  p.name = name;
                  porgyplugins.push_back(p);
              }
              delete pl;
          }

        typedef string::const_iterator iter;
        strategy_grammar<iter> strat_parser(errMsg, rules, properties, property_names, graph, porgy_plugins, porgyplugins); //le parser du langage de stratégie
        iter begin = strategy.begin();
        iter end = strategy.end();
        comments<iter> comments_;    //le parser pour les commentaires
        bool r = spirit::qi::phrase_parse(begin, end, strat_parser, comments_, strategy_elements);

        //parsing ok ?
        if(r && begin==end)
            return true;
        else if(errMsg.empty()){
            errMsg = " problem around "+string(begin, end);
        }
        return false;
    }

public:
    PLUGININFORMATION(PorgyConstants::APPLY_STRATEGY_ALGORITHM, "Bruno Pinaud", "19/07/10", "Comments", "1.0", PorgyConstants::CATEGORY_NAME)
    StrategySpirit(const PluginContext *context):Algorithm(context) {

        addDependency(PorgyConstants::CREATE_PARALLEL_RULE, "1.0.1");
        addDependency(PorgyConstants::CHECK_APPLY_RULE, "1.0");

        addInParameter<string>("Strategy", "", "ppick(rule_1;rule_1,0.5,rule_1||rule_1,0.5);rule_2");
        addInParameter<BooleanProperty>(PorgyConstants::POSITION, "", "", false);
        addInParameter<BooleanProperty>(PorgyConstants::BAN,"", "", false);
    }

    bool check(string &errorMsg) {
        //pas de preview
        pluginProgress->showPreview(false);

        if(!PorgyTlpGraphStructure::isModelGraph(graph)) {
            errorMsg = "A strategy must be applied on a model";
            return false;
        }

        if(dataSet!=nullptr)
            dataSet->get("Strategy", strategy);
        else {
            errorMsg = "dataSet problem";
            return false;
        }
        if(strategy.empty()) {
            errorMsg = "Strategy is empty";
            return false;
        }

        //parse the strategy
        string Pname;
        if(parse(errorMsg)) {
            BooleanProperty *P=nullptr, *Ban=nullptr;
            //position handling
            BooleanProperty *rest=graph->getProperty<BooleanProperty>("viewSelection");
            dataSet->get<BooleanProperty*>(PorgyConstants::POSITION, rest);
            dataSet->get<BooleanProperty*>(PorgyConstants::BAN, Ban);
            if(rest->getName()== "viewSelection") {
                Iterator<node>* it = rest->getNodesEqualTo(true);
                bool sel(it->hasNext()); //true if there is at least one node selected
                delete it;
                string property_name(PorgyConstants::P_PREFIX);
                int indice=0;
                while(1) {
                    if((!graph->existLocalProperty(property_name))&&(!findP(property_name))) {
                        P = graph->getLocalProperty<BooleanProperty>(property_name);
                        break;
                    }
                    ++indice;
                    property_name+=to_string(indice);
                }
                assert(P);
                if(!sel) { //nothing selected => select the whole graph
                    P->setAllNodeValue(true);
                }
                else //put into P property every selected elements
                    P->copy(rest);
                Pname = P->getName();
            }
            else
                Pname = rest->getName();

          //  graph->getRoot()->push(false);
            Trace traceobj(PorgyTlpGraphStructure::getMainTrace(graph));
            ObserverHolder hold;
            (void)hold;
            //running strategy
            unsigned oldnumberofNodes(traceobj.numberOfNodes());
            StrategyElement::stackStrat s;
            string BanName((Ban!=nullptr)?Ban->getName():"");
            StrategyElement::stratElem elt(Pname, BanName, graph, strategy_elements);
            s.push(elt);
            StrategyElement::vectNewGraphs newGraphAll;
            runStrategyElement::run_strat(s, PorgyConstants::state::OTHER, pluginProgress, newGraphAll);
            //the strategy has at least one rewriting step. Strategy edge only added when not using a all(R) operator
            if((traceobj.numberOfNodes()>oldnumberofNodes)&&((newGraphAll.size()==1)||((newGraphAll.size()==0)&&(traceobj.attributeExist(PorgyConstants::FAILURE_NODE_ADDED))))) {
                node last;
                if(!traceobj.getAttribute(PorgyConstants::FAILURE_NODE_ADDED, last)) {
                    Graph *new_graph = newGraphAll[0];
                    last = traceobj.findNode(new_graph->getName());
                }
                else {
                    traceobj.removeAttribute(PorgyConstants::FAILURE_NODE_ADDED);
                }

                node first = traceobj.findNode(graph->getName());
         //       BooleanProperty *ban=nullptr;
        //        dataSet->get<BooleanProperty *>(PorgyConstants::BAN, ban);
                traceobj.addStrategyTransformationEdge(first,last,P,Ban,strategy);
            }
            traceobj.redraw(errorMsg, pluginProgress);
            return true;
        }
        else
            return false;
    }

    bool run() {
        return true;
    }

};

PLUGIN(StrategySpirit)
