/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef STRATEGYELEMENT_H
#define STRATEGYELEMENT_H

#define BOOST_SPIRIT_DISABLE_UTREE_IO

#include <boost/variant.hpp>
#include <boost/spirit/include/qi_optional.hpp>
#include <boost/spirit/include/support_utree.hpp>

#include <portgraph/PorgyConstants.h>

#include <tulip/WithParameter.h>

#include <stack>
#include <tuple>

namespace tlp{
class Graph;
class PluginProgress;
class DataSet;
}

struct TulipProp {
    std::string name;
};

struct PluginName {
    std::string name;
};

class StrategyElement
{
public:
    enum class strat_operators{RULE, VECT, FOR, IDFAIL, PPICK};
    enum class rule_operators{ALL, ONE, MATCH};
    enum class syntactic_sugar{TRY, NOT};
    enum class idfail{ID, FAIL};
    enum class Graph_Element{NODE, PORT, EDGE};
    enum class relational_operators{EQ, NEQ, GT, GTE, LT, LTE, REGEXP};
    enum class comparison_operators{EQUAL, SUBSET, NEQ};
    enum class property_check{VALUE, PROP};
    enum class position_keywords{NGB, PROPERTY};
    enum class set_theory_op : unsigned {MINUS=0, UNION=1, INTER=2};
    enum class focus_operation : unsigned {CRTGRAPH=0, CRTPOS=1, CRTBAN=2, EMPTYSET=3};

    typedef std::vector<StrategyElement> vector_strategyelts;
    typedef std::tuple<std::string, std::string, tlp::Graph*, vector_strategyelts> stratElem;
    typedef std::stack<stratElem> stackStrat;
    typedef std::tuple<syntactic_sugar, vector_strategyelts> trynot_datastruct;
    typedef std::tuple<std::string, std::vector<std::string> > parallel_datastruct;
    typedef std::tuple<rule_operators, boost::variant<std::string, parallel_datastruct> >one_all_match_datastruct;
    typedef std::tuple<TulipProp, StrategyElement::relational_operators, boost::variant<std::string, TulipProp>> ngbprop_ds;
    typedef boost::variant<ngbprop_ds, std::string> property_ngb_expr;
    typedef std::tuple<position_keywords, boost::spirit::utree, StrategyElement::Graph_Element, property_ngb_expr> property_datastruct;
    typedef boost::variant<focus_operation, property_datastruct> focusop_datastruct;
    typedef std::tuple<PorgyConstants::strat_focus_op,  boost::spirit::utree> setPosBan_datastruct;
    typedef std::tuple<vector_strategyelts, vector_strategyelts, boost::optional<vector_strategyelts> > if_datastruct;
    typedef std::tuple<vector_strategyelts, vector_strategyelts> orelse_datastruct;
    typedef std::tuple<vector_strategyelts, boost::optional<unsigned>, vector_strategyelts> while_datastruct;
    typedef std::tuple<boost::spirit::utree, comparison_operators, boost::spirit::utree> comparison_datastruct;
    typedef std::pair<vector_strategyelts, double> ppick_elt;
    typedef std::vector<ppick_elt> ppick_datastruct;
    typedef std::tuple<vector_strategyelts, boost::optional<unsigned>> repeat_ds;
    typedef std::tuple<tlp::ParameterDescription, std::string> parameterdesc;
    typedef std::vector<StrategyElement::parameterdesc> paramvect;
    typedef std::tuple<PluginName,  boost::optional<paramvect> > update_ds;


    typedef std::vector<tlp::Graph*> vectNewGraphs;

    /**
     * @brief StrategyElement Constructor for one() and all() operators.
     * @brief by default, a rule R (or R||R) alone is considered as one(R)
     * @param rules
     * @param op Operation type
     */
    StrategyElement(const one_all_match_datastruct &rules);
    //Id/Fail
    StrategyElement(const idfail op=idfail::FAIL);
    //setPos, setBan
    StrategyElement(const setPosBan_datastruct &op);
    //ppick
    StrategyElement(const ppick_datastruct &vect);
    //orelse
    StrategyElement(const orelse_datastruct &v);
    //while
    StrategyElement(const while_datastruct &v);
    //if
    StrategyElement(const if_datastruct &v);
    //try, not
    StrategyElement(const trynot_datastruct &ds);
    //repeat
    StrategyElement(const repeat_ds &ds);
    //vector of elements
    StrategyElement(const vector_strategyelts &v);
    //equiv
    StrategyElement(const comparison_datastruct &v);
    //set
    StrategyElement(const update_ds &v);
    //isEmpty
    StrategyElement(const boost::spirit::utree &u);

    //run an element of the strategy
    bool run(tlp::PluginProgress *pluginProgress, tlp::Graph *g, std::string& P, std::string& Ban, const PorgyConstants::state state, stackStrat &s, vector_strategyelts &r, bool &doBreak, vectNewGraphs &newGraphAll) const;

private:
    //boost::variant to store all data structures
    boost::variant<idfail, one_all_match_datastruct, orelse_datastruct, while_datastruct, if_datastruct, repeat_ds, ppick_datastruct, setPosBan_datastruct, comparison_datastruct, update_ds> data;
};

#endif // STRATEGYELEMENT_H
