/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef STRATEGYPARSER_CXX
#define STRATEGYPARSER_CXX

#include <boost/spirit/include/qi.hpp>
#include <boost/spirit/include/phoenix.hpp>
#include <boost/spirit/include/phoenix_fusion.hpp>
#include <boost/fusion/adapted/std_pair.hpp>
#include <boost/fusion/adapted/std_tuple.hpp>

#include "StrategyElement.h"

#include <tulip/ForEach.h>

//pour parser les stratégies qui se réduise à ID ou Fail
struct id_fail : boost::spirit::qi::symbols<char, StrategyElement::idfail> {
    id_fail(const std::string& name = std::string("Id or Fail")):boost::spirit::qi::symbols<char, StrategyElement::idfail>(name) {
        add     ("id"    , StrategyElement::idfail::ID)
                ("fail"  , StrategyElement::idfail::FAIL)
                ;}
} id_fail;

struct comparison_op: boost::spirit::qi::symbols<char, StrategyElement::comparison_operators> {
    comparison_op(const std::string& name = std::string("Comparison operator")):boost::spirit::qi::symbols<char, StrategyElement::comparison_operators>(name) {
        add     ("="    , StrategyElement::comparison_operators::EQUAL)
                ("[subSet]"  , StrategyElement::comparison_operators::SUBSET)
                ("!="    , StrategyElement::comparison_operators::NEQ)
                ;}
} comparison_op;

//opérations pour le focus
struct focusing_op : boost::spirit::qi::symbols<char, boost::spirit::binary_string_type> {
    focusing_op(const std::string& name = std::string("CrtGraph/CrtPos/CrtBan/Emptyset")):boost::spirit::qi::symbols<char, boost::spirit::binary_string_type>(name) {
        add     ("crtGraph", boost::spirit::binary_string_type(std::to_string(static_cast<unsigned>(StrategyElement::focus_operation::CRTGRAPH))))
                ("crtPos", boost::spirit::binary_string_type(std::to_string(static_cast<unsigned>(StrategyElement::focus_operation::CRTPOS))))
                ("crtBan", boost::spirit::binary_string_type(std::to_string(static_cast<unsigned>(StrategyElement::focus_operation::CRTBAN))))
                ("[emptySet]", boost::spirit::binary_string_type(std::to_string(static_cast<unsigned>(StrategyElement::focus_operation::EMPTYSET))))
                ;}
} focusing_op;

//opérations sur les éléments
struct graph_element : boost::spirit::qi::symbols<char, StrategyElement::Graph_Element> {
    graph_element(const std::string& name = std::string("Graph Element")):boost::spirit::qi::symbols<char, StrategyElement::Graph_Element>(name) {
        add     ("node", StrategyElement::Graph_Element::NODE)
                ("port", StrategyElement::Graph_Element::PORT)
                ("edge", StrategyElement::Graph_Element::EDGE)
                ;}
} graph_element;

//relational operators
struct relop : boost::spirit::qi::symbols<char, StrategyElement::relational_operators> {
    relop(const std::string& name = std::string("Relational operators")):boost::spirit::qi::symbols<char, StrategyElement::relational_operators>(name) {
        add     ("==", StrategyElement::relational_operators::EQ)
                ("!=", StrategyElement::relational_operators::NEQ)
                (">", StrategyElement::relational_operators::GT)
                (">=", StrategyElement::relational_operators::GTE)
                ("<", StrategyElement::relational_operators::LT)
                ("<=", StrategyElement::relational_operators::LTE)
            ("=~", StrategyElement::relational_operators::REGEXP)
                ;}
} relop;

//position computation operators
struct position : boost::spirit::qi::symbols<char, StrategyElement::position_keywords> {
    position(const std::string& name = std::string("Property or Ngb keywords")):boost::spirit::qi::symbols<char, StrategyElement::position_keywords>(name) {
        add     ("property", StrategyElement::position_keywords::PROPERTY)
                ("ngb", StrategyElement::position_keywords::NGB)
                ;}
} position;

//position computation operators
struct positionUpdate : boost::spirit::qi::symbols<char, PorgyConstants::strat_focus_op> {
    positionUpdate(const std::string& name = std::string("setPos/setBan")):boost::spirit::qi::symbols<char, PorgyConstants::strat_focus_op>(name) {
        add     ("setPos", PorgyConstants::strat_focus_op::SETPOS)
                ("setBan", PorgyConstants::strat_focus_op::SETBAN)
                ;}
} positionUpdate;

//One / All for rules application
struct OneAllMatch : boost::spirit::qi::symbols<char, StrategyElement::rule_operators> {
    OneAllMatch(const std::string& name = std::string("one()/all()/match()")):boost::spirit::qi::symbols<char, StrategyElement::rule_operators>(name) {
        add     ("all", StrategyElement::rule_operators::ALL)
                ("one", StrategyElement::rule_operators::ONE)
                ("match", StrategyElement::rule_operators::MATCH)
                ;}
} OneAllMatch;

//try(), not(), repeat()
struct SyntacticSugar : boost::spirit::qi::symbols<char, StrategyElement::syntactic_sugar> {
    SyntacticSugar(const std::string& name = std::string("try(), not() or repeat()")):boost::spirit::qi::symbols<char, StrategyElement::syntactic_sugar>(name) {
        add     ("try", StrategyElement::syntactic_sugar::TRY)
                ("not", StrategyElement::syntactic_sugar::NOT)
                ;}
} SyntacticSugar;

struct expr_cup {
    template <typename T1, typename T2 = void>
    struct result { typedef void type; };

    void operator()(boost::spirit::utree& expr, boost::spirit::utree const& rhs) const {
        boost::spirit::utree lhs;
        lhs.swap(expr);
        expr.push_back(lhs);
        expr.push_back(rhs);
        expr.push_back(boost::spirit::utf8_symbol_type(std::to_string(static_cast<unsigned>(StrategyElement::set_theory_op::UNION))));
    }
};

struct expr_cap {
    template <typename T1, typename T2 = void>
    struct result { typedef void type; };

    void operator()(boost::spirit::utree& expr, boost::spirit::utree const& rhs) const {
        boost::spirit::utree lhs;
        lhs.swap(expr);
        expr.push_back(lhs);
        expr.push_back(rhs);
        expr.push_back(boost::spirit::utf8_symbol_type(std::to_string(static_cast<unsigned>(StrategyElement::set_theory_op::INTER))));
    }
};

struct expr_minus {
    template <typename T1, typename T2 = void>
    struct result { typedef void type; };

    void operator()(boost::spirit::utree& expr, boost::spirit::utree const& rhs) const {
        boost::spirit::utree lhs;
        lhs.swap(expr);
        expr.push_back(lhs);
        expr.push_back(rhs);
        expr.push_back(boost::spirit::utf8_symbol_type(std::to_string(static_cast<unsigned>(StrategyElement::set_theory_op::MINUS))));
    }
};

struct one_expr {
    template <typename T1>
    struct result { typedef boost::spirit::utree type; };

    boost::spirit::utree operator()(boost::spirit::utree const& rhs) const {
        boost::spirit::utree lhs;
        lhs.push_back(rhs);
        lhs.push_back(boost::spirit::utf8_string_type("one"));
        return lhs;
    }
};

struct all_expr {
    template <typename T1>
    struct result { typedef boost::spirit::utree type; };

    boost::spirit::utree operator()(boost::spirit::utree const& rhs) const {
        boost::spirit::utree lhs;
        lhs.push_back(rhs);
        lhs.push_back(boost::spirit::utf8_string_type("all"));
        return lhs;
    }
};

//parser for 0.?? style number
template <typename T>
struct proba : boost::spirit::qi::strict_ureal_policies<T> {
    static bool const expect_dot = true;    //obligation d'avoir une virgule
    static bool const allow_leading_dot=false;
    static bool const allow_trailing_dot=false;
    //no exponent
    template <typename It>
    static bool parse_exp(It&, It const&){return false;}
    //  No exponent
    template <typename It, typename Attribute>
    static bool parse_exp_n(It&, It const&, Attribute&){return false;}
    //forcément 0 avant la virgule
    template <typename It, typename Attribute>
    static bool parse_n(It& first, It const& last, Attribute& ) {
        return boost::spirit::qi::parse(first, last, boost::spirit::qi::uint_(0)); //needs at least boost 1.46.1
    }
};

//grammar to handle comments, tab or space
template <typename Iterator>
struct comments : boost::spirit::qi::grammar<Iterator> {

    comments():comments::base_type(start) {
        using boost::spirit::ascii::char_;
        using boost::spirit::qi::lit;

        start = boost::spirit::ascii::space  // tab/space/cr/lf
                | lit("/*") >> *(char_ - lit("*/")) >> lit("*/")                 // C++-style comments
                | lit("//") >> *(char_ - boost::spirit::qi::eol);
    }
    boost::spirit::qi::rule<Iterator> start;
};

//phoenix functor for checking ppick probabilities
struct checkproba {

    template<typename CONTAINER>
    struct result
    {
        typedef bool type;
    };

    template<typename CONTAINER>
    bool operator()(CONTAINER &v) const {
        double sum=0;
        for(const auto &elt: v) {
            sum += elt.second;
        }
        return (sum == 1.0);
    }
};

//graph plugin parameters
struct CheckParams {
  template<typename CONTAINER>
    struct result {
        typedef void type;
    };

    template<typename CONTAINER>
    void operator()(CONTAINER &plugin_name, boost::spirit::qi::symbols<char, tlp::ParameterDescription> &pluginParam) const {
     const tlp::ParameterDescriptionList &param =  tlp::PluginLister::getPluginParameters(plugin_name.name);
     tlp::ParameterDescription p;
     forEach(p, param.getParameters()) {
      //   if(p.getDirection()==tlp::IN_PARAM||tlp::INOUT_PARAM)
             pluginParam.add(p.getName(), p);
     }
}
};

//phoenix functor for checking that we have two tulip property of the same type
struct checkprop {
    template<typename CONTAINER>
    struct result
    {
        typedef bool type;
    };

    template<typename CONTAINER>
    bool operator()(tlp::Graph *graph, CONTAINER &p1, CONTAINER &p2) const {
        const TulipProp prop1(p1), prop2(p2);
        return (graph->getProperty(prop1.name)->getTypename()==graph->getProperty(prop2.name)->getTypename());
    }
};

boost::phoenix::function<expr_cup> const cup_e = expr_cup();
boost::phoenix::function<expr_cap> const cap_e = expr_cap();
boost::phoenix::function<expr_minus> const minus_e = expr_minus();
boost::phoenix::function<one_expr> const onef = one_expr();
boost::phoenix::function<all_expr> const allf = all_expr();
boost::phoenix::function<checkproba> const _checkProbaPpick_ = checkproba();
boost::phoenix::function<checkprop> const _CheckProp_ = checkprop();
boost::phoenix::function<CheckParams> const CheckParam = CheckParams();

BOOST_FUSION_ADAPT_STRUCT(TulipProp,(std::string, name))

BOOST_FUSION_ADAPT_STRUCT(PluginName,(std::string, name))

template <typename Iterator>
struct strategy_grammar : boost::spirit::qi::grammar<Iterator, StrategyElement::vector_strategyelts(), comments<Iterator> > {

    typedef comments<Iterator> comments_;

strategy_grammar(std::string &errorMsg, const std::vector<std::string> &rules, const std::vector<TulipProp> &propertynames, const std::vector<std::string> &property_names, tlp::Graph* g, const std::vector<std::string> &porgy_plugin_names, const std::vector<PluginName> &porgyplugins):
    strategy_grammar::base_type(start, "Porgy Grammar"),errMsg(errorMsg),sym_rules(rules, rules, "valid rule name"),
    sym_properties(property_names, propertynames, "valid property name"),graph(g), sym_porgy_plugins(porgy_plugin_names, porgyplugins, "Existing porgy tulip plugin"){
    using boost::spirit::qi::lit;
    using boost::spirit::qi::lexeme;
    using boost::spirit::qi::char_;
    using boost::spirit::qi::uint_;
    using boost::phoenix::construct;
    using boost::phoenix::val;
    using boost::spirit::qi::labels::_1;
    using boost::spirit::qi::labels::_2;
    using boost::spirit::qi::labels::_3;
    using boost::spirit::qi::labels::_4;
    using boost::spirit::qi::labels::_a;
    using boost::spirit::qi::labels::_val;
    using boost::spirit::qi::labels::_pass;

    //parser for float number style 0.xx
    boost::spirit::qi::real_parser<double, proba<double> > doublep_;

    //top level structure of a strategy => s;s;....;s
    start %= seq % ';';

    //all possible elements
    seq =  parallel_only[_val = construct<StrategyElement>(_1)]|
            one_all_match_T[_val = construct<StrategyElement>(_1)]|
            ppick[_pass=_checkProbaPpick_(_1)][_val = construct<StrategyElement>(_1)]|
            isEmpty[_val = construct<StrategyElement>(_1)]|
            pwhile[_val = construct<StrategyElement>(_1)]|
            pif[_val = construct<StrategyElement>(_1)]|
            porelse[_val = construct<StrategyElement>(_1)]|
            id_fail[_val = construct<StrategyElement>(_1)]|
            ptry_not[_val = construct<StrategyElement>(_1)]|
            prepeat[_val = construct<StrategyElement>(_1)]|
            setpos_ban[_val = construct<StrategyElement>(_1)]|
            comparison[_val = construct<StrategyElement>(_1)]|
            rulename_only[_val = construct<StrategyElement>(_1)]|
            update[_val = construct<StrategyElement>(_1)];

            //ppick operator
            ppick %= lexeme[lit("ppick")]
            > '('
            > ((start > lit(',') > doublep_) % ',')
            > lit(')');

                //while loop
                pwhile %= lexeme[lit("while")]
                > lit('(')
                > start
                > lit(')')
                > -('(' > uint_ > ')') //optional parameter to indicate the maximum number of loops
                > lexeme[lit("do")]
                > lit('(')
                > start
                > lit(')');

                //orelse operator
                porelse %= lit('(')
                > start
                > lit(')')
                >lexeme[lit("orelse")]
                > lit('(')
                > start
                > lit(')');

                //if operator
                pif %= lexeme[lit("if")]
                > lit('(')
                > start
                > lit(')')
                > lexeme[lit("then")]
                > lit('(')
                > start
                > lit(')')
                > -(lexeme[lit("else")]> lit('(')> start > lit(')'));

                //try(), not()
                ptry_not %= lexeme[SyntacticSugar]
                > lit('(')
                > start
                > lit(')');

            // repeat()
            prepeat %= lexeme["repeat"]
            > lit('(')
            > start
            > lit(')')
            > -('(' > uint_ > ')'); //optional parameter to indicate the maximum number of loops


                //règle en parallèle A||A' (minimun 2, pas de maxi)
                parallel %= sym_rules >> +(lexeme[lit("||")] > sym_rules);
                parallel_only = parallel[_val = construct<StrategyElement::one_all_match_datastruct>(StrategyElement::rule_operators::ONE,_1)];

                //a simple rule. By default it is considered as one()
                rulename_only = sym_rules[_val = construct<StrategyElement::one_all_match_datastruct>(StrategyElement::rule_operators::ONE,_1)];

                //one(T) or all(T) operator
                one_all_match_T %= lexeme[OneAllMatch]
                > lit('(')
                > (parallel | sym_rules)
                >lit(')');

//////////////////////////
                expression = term [_val = _1]
                >> *((cup >> term  [cup_e(_val, _1)])
                | (cap >> term [cap_e(_val,_1)])
                        );

                term = focusing [_val = _1]
                >> *(minus >> focusing [minus_e(_val,_1)]);

                focusing =  focusing_op [_val = _1]
                            | property_ngb [_val = construct<boost::spirit::any_ptr>(boost::phoenix::new_<StrategyElement::property_datastruct>(_1))] //should be a utree
                            |   '(' >> expression [_val = _1] >> ')';

                scope =  one_F [_val = _1]
                         | all_F  [_val = _1]
                         | expression [_val = allf(_1)]; //by default, it is all(F)

                cup = lit("[CUP]");
                cap = lit("[CAP]");
                minus = lit("\\");

                //One(F) => returns one element of F chosen uniformely at random
                one_F %= lexeme[lit("one")]
                > '('
                > expression [_val = onef(_1)]
                > ')';

                //All(F) => returns all element of F
                all_F %= lexeme[lit("all")]
                > '('
                > expression [_val = allf(_1)]
                > ')';

            property_ngb_expr %= (sym_properties[_a = _1] > relop > ((L'"' > +((char_) - L'"') > lit('"'))|sym_properties[_pass = _CheckProp_(graph, _a, _1)]))
                                 | (boost::spirit::ascii::string("True"));

                //Property or Ngb
            property_ngb %= lexeme[position]
                > L'('
                > expression
                > ','
                > graph_element > ',' > property_ngb_expr
                > ')';

                //position update (setPos, setBan)
                setpos_ban %= lexeme[positionUpdate]
                > L'('
                > scope
                > L')';

                //comparison
                comparison %= scope
                > comparison_op
                > scope;

                //isEmpty
                isEmpty %=  lexeme[lit("isEmpty")]
                > L'('
                > expression
                > L')';

                //set
                param %= pluginParam >> lit(':') >> *(char_("a-zA-Z_0-9 "));
                paramlist %= param % ',';
                update %= lexeme[lit("update")]
                > L'('
                > L'"' > sym_porgy_plugins[CheckParam(_1, boost::phoenix::ref(pluginParam))] > lit('"')
                > -(L'{' >> paramlist >> L'}')
                > L')';

//////////////////////////////////////
            start.name("start");
            seq.name("seq");
            ppick.name("ppick");
            rulename_only.name("valid rulename");
            pwhile.name("while");
            porelse.name("orelse");
            ptry_not.name("try(), not()");
            prepeat.name("repeat() operator");
            pif.name("if");
            parallel.name("parallel");
            parallel_only.name("parallel");
            setpos_ban.name("setPos()/setBan()");
            property_ngb.name("Property() or Ngb()");
            one_all_match_T.name("one() or all()");
            focusing.name("focusing expression");
            scope.name("scope expression");
            term.name("scope term");
            expression.name("expression");
            cup.name("union operator");
            cap.name("inter operator");
            minus.name("minus operator");
            comparison.name("Equiv operator");
            one_F.name("One operator");
            update.name("update operator");
            pluginParam.name("Valid plugin parameter");
            paramlist.name("List of plugin's parameters");
            param.name("Plugin parameter");
            isEmpty.name("IsEmpty operator");
            property_ngb_expr.name("Property or Ngb expression");
        all_F.name("All Focusing");

        boost::spirit::qi::on_error<boost::spirit::qi::fail>(start, boost::phoenix::ref(errMsg) = (construct<std::string>(val("Error! Expecting "))
        + boost::phoenix::bind(&boost::spirit::info::tag, _4)      //what failed
        + construct<std::string>(val(" here: \"") )
        + construct<std::string>(_3, _2)
        + construct<std::string>( val("\"") )
        )
        );
}

        std::string &errMsg;
        boost::spirit::qi::symbols<char, std::string> sym_rules;
        boost::spirit::qi::symbols<char, tlp::ParameterDescription> pluginParam;
        boost::spirit::qi::symbols<char, TulipProp> sym_properties;
        tlp::Graph *graph;
        boost::spirit::qi::symbols<char, PluginName> sym_porgy_plugins;
        boost::spirit::qi::rule<Iterator, StrategyElement::vector_strategyelts(), comments_> start;
        boost::spirit::qi::rule<Iterator, StrategyElement::trynot_datastruct(), comments_> ptry_not;
        boost::spirit::qi::rule<Iterator, StrategyElement::repeat_ds(), comments_> prepeat;
        boost::spirit::qi::rule<Iterator, StrategyElement::ppick_datastruct(), comments_> ppick;
        boost::spirit::qi::rule<Iterator, StrategyElement::one_all_match_datastruct(), comments_> one_all_match_T,rulename_only,parallel_only;
        boost::spirit::qi::rule<Iterator, StrategyElement::parallel_datastruct(), comments_> parallel;
        boost::spirit::qi::rule<Iterator, StrategyElement::setPosBan_datastruct(), comments_> setpos_ban;
        boost::spirit::qi::rule<Iterator, StrategyElement::orelse_datastruct(), comments_> porelse;
        boost::spirit::qi::rule<Iterator, StrategyElement::while_datastruct(), comments_> pwhile;
        boost::spirit::qi::rule<Iterator, StrategyElement(), comments_> seq;
        boost::spirit::qi::rule<Iterator, StrategyElement::if_datastruct(), comments_> pif;
        boost::spirit::qi::rule<Iterator, StrategyElement::comparison_datastruct(), comments_> comparison;
        boost::spirit::qi::rule<Iterator, StrategyElement::property_datastruct(), comments_> property_ngb; //should be a utree
        boost::spirit::qi::rule<Iterator, StrategyElement::property_ngb_expr(), comments_, boost::spirit::qi::locals<TulipProp> > property_ngb_expr;
        boost::spirit::qi::rule<Iterator, boost::spirit::utree(), comments_> expression, focusing, term, one_F, all_F, scope, isEmpty;
        boost::spirit::qi::rule<Iterator, boost::spirit::utf8_symbol_type(), comments_> cup, cap, minus;
        boost::spirit::qi::rule<Iterator, StrategyElement::parameterdesc(), comments_> param;
        boost::spirit::qi::rule<Iterator, StrategyElement::paramvect(), comments_> paramlist;
        boost::spirit::qi::rule<Iterator, StrategyElement::update_ds(), comments_> update;

    };
#endif // STRATEGYPARSER_CXX
