/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <tulip/ImportModule.h>
#include <tulip/Graph.h>
#include <tulip/ConnectedTest.h>
#include <tulip/StringCollection.h>
#include <tulip/LayoutProperty.h>
#include <tulip/TlpQtTools.h>
#include <tulip/TulipProject.h>

#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/PortGraphModel.h>

using namespace tlp;
using namespace std;

namespace {

const char * paramHelp[] = {
    // filename
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "pathname" ) \
    HTML_HELP_BODY() \
    "The file to import." \
    HTML_HELP_CLOSE(),

    // port
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "boolean" ) \
    HTML_HELP_DEF( "default", "true" ) \
    HTML_HELP_BODY() \
    "Level the number of ports per node." \
    HTML_HELP_CLOSE(),

    // differenciate in/out port
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "boolean" ) \
    HTML_HELP_DEF( "default", "true" ) \
    HTML_HELP_BODY() \
    "Used to specify whether or not in- and outcoming edges will use same ports (if false, specific ports will be created to differenciate the edges directions)." \
    HTML_HELP_CLOSE(),

    // apply drawing algorithm
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "boolean" ) \
    HTML_HELP_DEF( "default", "false" ) \
    HTML_HELP_BODY() \
    "Apply the default drawing algorithm (FM^3)." \
    HTML_HELP_CLOSE(),
};
}

class PorgyTlpImport : public ImportModule {
public:
    PLUGININFORMATION("Porgy Tlp Import","Name","2014/06/31","Description","1.1",PorgyConstants::CATEGORY_NAME)
    PorgyTlpImport(tlp::PluginContext* context):ImportModule(context) {
        addInParameter<std::string>("file::filename",paramHelp[0],"");
        addInParameter<bool>("Additionnal ports",paramHelp[1],"true");
        addInParameter<bool>("Same In/Out ports",paramHelp[2],"true");
        addInParameter<bool>("Redraw the graph",paramHelp[3],"false");

        addDependency("TLP Import", "1.0");
        addDependency(PorgyConstants::REDRAW_GRAPH, "1.0");
    }

     std::list<std::string> fileExtensions() const {
      list<string> l;
      l.push_back(".porgy.tlp");
//      l.push_back(".tlpx");
      l.push_back(".porgy.tlp.gz");
      return l;
    }

    bool importGraph() {

        string filen;
        dataSet->get("file::filename", filen);
        QString filename(tlp::tlpStringToQString(filen));
        list<string> ext(fileExtensions());
        bool found=false;
        for(auto &st : ext) {
            if(filename.endsWith(tlp::tlpStringToQString(st))) {
                found=true;
                break;
            }
        }
        if(!found) {
            pluginProgress->setError("file format not supported. You have to use the old tlp format with the PORGY extension (.porgy.tlp, .porgy.tlp.gz)");
            return false;
        }
        Graph *g = tlp::importGraph("TLP Import", *dataSet, pluginProgress, graph);

        if(g!=nullptr) {
            Observable::holdObservers();

            bool multiPort=true, mergeInOut=true, redraw=false;
            dataSet->get<bool>("Additionnal ports", multiPort);
            dataSet->get<bool>("Same In/Out ports", mergeInOut);
            dataSet->get<bool>("Redraw the graph", redraw);
            PorgyTlpGraphStructure::Graph2PortGraph(graph, pluginProgress, multiPort, mergeInOut);

            //layout new graph
            if (redraw) {
                DataSet dSet;
                string errMsg;
                StringCollection algo(LAYOUT_ALGO);
                algo.setCurrent("FM^3 (OGDF)");
                dSet.set("Layout", algo);
                LayoutProperty *layout = graph->getProperty<LayoutProperty>("viewLayout");
                if (!graph->applyPropertyAlgorithm(PorgyConstants::REDRAW_GRAPH, layout, errMsg, pluginProgress, &dSet)) {
                    cerr << "Problem when applying layout algorithm: " << errMsg << endl;
                    Observable::unholdObservers();
                    return false;
                }
            }
            Observable::unholdObservers();
            return true;
        }
        else {
            cerr << "problem when importing a new graph" << endl;
            return false;
        }
    }
};

PLUGIN(PorgyTlpImport)
