#include <tulip/ImportModule.h>
#include <tulip/Graph.h>
#include <tulip/ConnectedTest.h>
#include <tulip/StringCollection.h>
#include <tulip/LayoutProperty.h>
#include <tulip/TlpQtTools.h>

#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/PortGraphModel.h>
#include <portgraph/PorgyConstants.h>

#include <set>

using namespace tlp;
using namespace std;

namespace {

const char * paramHelp[] = {
    // filename
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "pathname" ) \
    HTML_HELP_BODY() \
    "The file to import." \
    HTML_HELP_CLOSE(),
};
}

class PorgyRuleImport : public ImportModule {
private:
    std::map<unsigned int, unsigned int> nodeMap;
    std::map<unsigned int, unsigned int> edgeMap;

    void copyNodes(Graph *sourceGraph, Graph *targetGraph){
        tlp::node currentNode;
        tlp::node newNode;
        forEach(currentNode, sourceGraph->getNodes()){
            newNode = targetGraph->addNode();
            nodeMap.insert(std::pair<unsigned int, unsigned int>(currentNode.id, newNode.id));
        }
    }

    void copyEdges(Graph *sourceGraph, Graph *targetGraph){
        tlp::edge currentEdge;
        tlp::edge newEdge;
        tlp::node sourceNode;
        tlp::node targetNode;
        forEach(currentEdge, sourceGraph->getEdges()){
            sourceNode = node(nodeMap[sourceGraph->source(currentEdge).id]);
            assert(sourceNode.isValid());
            targetNode = node(nodeMap[sourceGraph->target(currentEdge).id]);
            assert(targetNode.isValid());
            newEdge = targetGraph->addEdge(sourceNode, targetNode);
            edgeMap.insert(std::pair<unsigned int, unsigned int>(currentEdge.id, newEdge.id));
        }
    }

    void copyProperties(Graph *sourceGraph, Graph *targetGraph) {
        std::string currentPropertyName;
        tlp::PropertyInterface *currentProperty;
        tlp::PropertyInterface *newProperty;
        tlp::node sourceNode;
        tlp::node targetNode;
        tlp::edge sourceEdge;
        tlp::edge targetEdge;
        //all properties are inherited
        forEach(currentPropertyName, sourceGraph->getProperties()) {
            currentProperty = sourceGraph->getProperty(currentPropertyName);
            if (currentPropertyName==PorgyConstants::ARITY ||
                    currentPropertyName == PorgyConstants::N ||
                    currentPropertyName == PorgyConstants::M)
                newProperty = targetGraph->getProperty(currentPropertyName, currentProperty->getTypename());
            else if(currentPropertyName==PorgyConstants::RULE_SIDE)
                newProperty = PorgyTlpGraphStructure::getRulesRoot(targetGraph)->getProperty(currentPropertyName, currentProperty->getTypename());
            else
                newProperty = targetGraph->getRoot()->getProperty(currentPropertyName, currentProperty->getTypename());
            //forEach node
            forEach(sourceNode, sourceGraph->getNodes()) {
                targetNode = node(nodeMap[sourceNode.id]);
                assert(targetNode.isValid());
                newProperty->setNodeStringValue(targetNode, currentProperty->getNodeStringValue(sourceNode));
            }
            //forEach edge
            forEach(sourceEdge, sourceGraph->getEdges()) {
                targetEdge = edge(edgeMap[sourceEdge.id]);
                assert(targetEdge.isValid());
                newProperty->setEdgeStringValue(targetEdge, currentProperty->getEdgeStringValue(sourceEdge));
            }
        }
    }

    void copyRuleAttributes(Graph *sourceGraph, Graph *targetGraph) {
        tlp::DataSet dSet = sourceGraph->getAttributes();
        tlp::Iterator<std::pair<std::string, tlp::DataType*> > *it;
        std::pair<std::string, tlp::DataType*> currentData;
        std::string temp;
        std::set<std::string> allowed_property (PorgyConstants::RULE_IMPORT_ALLOWED, PorgyConstants::RULE_IMPORT_ALLOWED + PorgyConstants::RULE_IMPORT_ALLOWED_SIZE);
        it=dSet.getValues();
        while (it->hasNext()) {
            currentData = it->next();
            if (allowed_property.find(currentData.first) != allowed_property.end()) {
                sourceGraph->getAttribute<std::string>(currentData.first, temp);
                targetGraph->setAttribute<std::string>(currentData.first, temp);
            }
        }
    }

    void transformAlgorithm(Graph *targetGraph) {
        std::string algorithm;
        targetGraph->getAttribute<std::string>(PorgyConstants::RULE_ALGORITHM, algorithm);
        std::string ancient, recent, strIdentifier="#?*";
        std::size_t index;
        for (auto& it : nodeMap) {
            ancient = "node("+ to_string(it.first) +")";
            recent = "node("+strIdentifier+ to_string(it.second) +")";
            index = algorithm.find(ancient);
            while(index!=std::string::npos) {
                algorithm.replace(index, ancient.length(), recent);
                index = algorithm.find(ancient);
            }
            ancient = "n("+ to_string(it.first) +")";
            recent = "n("+strIdentifier+ to_string(it.second) +")";
            index = algorithm.find(ancient);
            while(index!=std::string::npos) {
                algorithm.replace(index, ancient.length(), recent);
                index = algorithm.find(ancient);
            }
        }
        for (auto& it : edgeMap) {
            ancient = "edge("+ to_string(it.first) +")";
            recent = "edge("+strIdentifier+ to_string(it.second) +")";
            index = algorithm.find(ancient);
            while(index!=std::string::npos) {
                algorithm.replace(index, ancient.length(),recent);
                index = algorithm.find(ancient);
            }
            ancient = "e("+ to_string(it.first) +")";
            recent = "e("+strIdentifier+ to_string(it.second) +")";
            index = algorithm.find(ancient);
            while(index!=std::string::npos) {
                algorithm.replace(index, ancient.length(), recent);
                index = algorithm.find(ancient);
            }
        }
        ancient = "("+strIdentifier;
        recent = "(";
        index = algorithm.find(ancient);
        while(index!=std::string::npos) {
            algorithm.replace(index, ancient.length(), recent);
            index = algorithm.find(ancient);
        }
        targetGraph->setAttribute<std::string>(PorgyConstants::RULE_ALGORITHM, algorithm);
    }

public:
    PLUGININFORMATION("Rule Import","Jason Vallet","2014/06/31","Description","1.1",PorgyConstants::CATEGORY_NAME)
    PorgyRuleImport(tlp::PluginContext* context):ImportModule(context) {
        addInParameter<std::string>("file::filename",paramHelp[0],"");

        addDependency("TLP Import", "1.0");
        addDependency(PorgyConstants::REDRAW_GRAPH, "1.0");
    }

    std::list<std::string> fileExtensions() const {
     list<string> l;
     l.push_back(".porgy.rule");
     return l;
   }

    bool importGraph() {
        string filen;
        dataSet->get("file::filename", filen);
        QString filename(tlp::tlpStringToQString(filen));
        list<string> ext(fileExtensions());
        bool found=false;
        for(list<string>::const_iterator it=ext.begin();it!=ext.end();++it) {
            if(filename.endsWith(tlp::tlpStringToQString(*it))) {
                found=true;
                break;
            }
        }
        if(!found) {
            pluginProgress->setError("file format not supported. You have to use the PORGY rule format (.porgy.rule)");
            return false;
        }

        Graph *rule = tlp::importGraph("TLP Import", *dataSet, pluginProgress);

        if(rule!=nullptr) {
            Observable::holdObservers();
            Graph *newRule = graph->addSubGraph();

            copyNodes(rule, newRule);
            copyEdges(rule, newRule);
            copyProperties(rule, newRule);
            copyRuleAttributes(rule, newRule);
            transformAlgorithm(newRule);

            Observable::unholdObservers();
            delete rule;
            return graph;
        }
        else {
            std::cerr << "problem when importing a new rule" << endl;
            return false;
        }
    }
};

PLUGIN(PorgyRuleImport)
