/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <tulip/ImportModule.h>
#include <tulip/Graph.h>
#include <tulip/StringCollection.h>
#include <tulip/LayoutProperty.h>
#include <tulip/StringProperty.h>

#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/PortGraphModel.h>

using namespace tlp;
using namespace std;

namespace {

const char * paramHelp[] = {
    // minsize
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "int" ) \
    HTML_HELP_DEF( "default", "10" ) \
    HTML_HELP_BODY() \
    "Minimal number of nodes in the tree." \
    HTML_HELP_CLOSE(),
    // maxsize
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "int" ) \
    HTML_HELP_DEF( "default", "100" ) \
    HTML_HELP_BODY() \
    "Maximal number of nodes in the tree." \
    HTML_HELP_CLOSE(),
    // maxdegree
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "int" ) \
    HTML_HELP_DEF( "default", "5" ) \
    HTML_HELP_BODY() \
    "Maximal degree of the nodes." \
    HTML_HELP_CLOSE(),
    // tree layout
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "bool" ) \
    HTML_HELP_DEF( "default", "false" ) \
    HTML_HELP_BODY() \
    "If true, the generated tree is drawn with a tree layout algorithm." \
    HTML_HELP_CLOSE(),
  // Port labels
  HTML_HELP_OPEN() \
  HTML_HELP_DEF( "type", "bool" ) \
  HTML_HELP_DEF( "default", "true" ) \
  HTML_HELP_BODY() \
  "If true, the source port of each edge will be called \"Parent\", the target port will be called \"Child\"." \
  HTML_HELP_CLOSE(),
};
}

class random_tree : public ImportModule {
public:
    PLUGININFORMATION("Random Tree","Pinaud","21/08/2013","Imports a new randomly generated tree (portgraph)","1.1",PorgyConstants::CATEGORY_NAME)
    random_tree(tlp::PluginContext* context):ImportModule(context) {
        addInParameter<unsigned>("Minimum size",paramHelp[0],"10");
        addInParameter<unsigned>("Maximum size",paramHelp[1],"100");
        addInParameter<unsigned>("Maximal node's degree",paramHelp[2],"5");
        addInParameter<bool>("Apply a tree layout",paramHelp[3],"true");
        addInParameter<bool>("Add port labels", paramHelp[4],"true");
        addDependency("Random General Tree", "1.1");
        addDependency(PorgyConstants::REDRAW_GRAPH, "1.0");
    }

    bool importGraph() {
        Graph *g = tlp::importGraph("Random General Tree", *dataSet, pluginProgress, graph);
        if(g!=nullptr) {
            Observable::holdObservers();

            PorgyTlpGraphStructure::Graph2PortGraph(graph, pluginProgress);

            bool addLabels=true;
            dataSet->get("Add port labels", addLabels);
            StringProperty *viewLabel = graph->getProperty<StringProperty>("viewLabel");
            if(addLabels) {
                edge e;
                PortGraphModel pg(graph);
                forEach(e, pg.getEdges()) {
                  const pair<node, node>&ends = graph->ends(e);
                  viewLabel->setNodeValue(ends.first, "Child");
                  viewLabel->setNodeValue(ends.second, "Parent");
                }
              }

            //layout new graph
            bool layout=false;
            if(dataSet!=nullptr) {
                dataSet->get("Apply a tree layout", layout);
            }
            if(layout) {
                // apply Tree Leaf
                DataSet dSet;
                string errMsg;
                StringCollection algo(LAYOUT_ALGO);
                algo.setCurrent("Tree Leaf");
                dSet.set("Layout", algo);
                LayoutProperty *layout = graph->getProperty<LayoutProperty>("viewLayout");
                if(!graph->applyPropertyAlgorithm(PorgyConstants::REDRAW_GRAPH, layout, errMsg,pluginProgress, &dSet)) {
                    cerr << "Problem when applying layout algorithm: " << errMsg << endl;
                    Observable::unholdObservers();
                    return false;
                }
            }
            Observable::unholdObservers();
            return true;
        }
        else {
            cerr << "problem when importing a new graph" << endl;
            return false;
        }
    }
};

PLUGIN(random_tree)
