/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef STRATEGYMANAGER_H
#define STRATEGYMANAGER_H

#include <tulip/DataSet.h>

#include <porgy/Strategy.h>

#include <QString>
#include <QVector>
#include <QList>
#include <QAbstractListModel>
#include <QTextDocument>

namespace tlp{
class Graph;
}

/**
  * @brief Class model managing a list of strategy.
  **/
class StrategyManager : public QAbstractListModel
{

public:

    enum StrategyErrors{STRATEGY_NO_ERROR,STRATEGY_CIRCULAR_CALL,STRATEGY_MISSING_STRATEGY,STRATEGY_EMPTY_NAME,STRATEGY_EXISTING_NAME};

    StrategyManager(QObject* parent=nullptr);

    unsigned int count()const;
    bool empty() const;

    void setGraph(tlp::Graph *g);

    tlp::Graph* graph()const{
        return _graph;
    }
    unsigned int addStrategy(const QString& name,StrategyErrors& error);
    unsigned int addStrategy(const QString& name, const QString& code,StrategyErrors& error);
    void removeStrategy( int index);


    QString getStrategyName( int index)const;
    /**
      * @brief Return the last commited values of the document.
      **/
    QString getStrategyCode(int index, StrategyErrors&error)const;
    QString getStrategy(int index) const;

    Strategy getStrategy(int index,StrategyErrors& error) const;

    /**
      * @brief Return the working document for the index.
      **/
    QTextDocument& getStrategyDocument( int index);

    /**
      * @brief Update the strategy name.
      **/
    StrategyErrors setStrategyName( int index,const QString& newName);
    /**
      * @brief Update the strategy code. If there is a buffer for this strategy it will be erased.
      **/
    void setStrategyCode( int index,const QString& newCode);

    /**
      * @brief Store the new code in the buffer of the strategy. Use setStrategyCode if you want to modify the real strategy.
      **/
    void updateStrategyCode( int index,const QString& newCode);

    /**
      * @brief Search if a strategy with the given name exists and return it's index else return -1.
      **/
    int indexOfStrategy(const QString& strategyName)const;


    tlp::DataSet getData()const;
    void setData(const tlp::DataSet& data);

    //QabstractListModel interface
    int rowCount ( const QModelIndex & parent = QModelIndex() ) const;
    QVariant data ( const QModelIndex & index, int role = Qt::DisplayRole ) const;
    QVariant headerData ( int section, Qt::Orientation orientation, int role = Qt::DisplayRole ) const;
    bool setData ( const QModelIndex & index, const QVariant & value, int role = Qt::EditRole );
    Qt::ItemFlags flags ( const QModelIndex & index ) const;
    QMimeData * mimeData ( const QModelIndexList & indexes ) const;

    /**
      * @brief Return the label corresponding to the error code.
      **/
    static QString errorMessage(StrategyErrors& error);
private:

    /**
      * @brief Replace all the strategy reference in the code by the code of its strategies. Set the error variable with the rigth error code.
      **/
    QString expandStrategyCode(const QString& metaStrategy,QList<QString>& alreadyExpendedStrategies,StrategyErrors& error)const;



    QVector< std::pair<QString,QTextDocument*> > strategies;
    tlp::Graph *_graph;
};

#endif // STRATEGYMANAGER_H
