/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "PorgyPropertyManager.h"
#include "ui_PorgyPropertyManager.h"

#include <portgraph/PorgyConstants.h>
#include <portgraph/PorgyTlpGraphStructure.h>

#include <tulip/Graph.h>
#include <tulip/TlpQtTools.h>
#include <tulip/DataSet.h>
#include <tulip/PropertyCreationDialog.h>
#include <tulip/BooleanProperty.h>
#include <tulip/ColorProperty.h>
#include <tulip/DoubleProperty.h>
#include <tulip/IntegerProperty.h>
#include <tulip/LayoutProperty.h>
#include <tulip/SizeProperty.h>
#include <tulip/StringProperty.h>
#include <tulip/StringCollection.h>

#include <utility>
#include <QMessageBox>
#include <QComboBox>
#include <QTableWidgetItem>

using namespace tlp;
using namespace std;


void PorgyPropertyManager::propertyManager(tlp::Graph* parentGraph, QWidget *parent) {
    PorgyPropertyManager property(parent, parentGraph);
    property.exec();
}

PorgyPropertyManager::PorgyPropertyManager(QWidget *parent, Graph *g) :QDialog(parent), graph(g), ui(new Ui::PorgyPropertyManager) {
    ui->setupUi(this);
    connect(ui->createPropertyButton, SIGNAL(released()), SLOT(create()));
    connect(ui->deletePropertyButton, SIGNAL(released()), SLOT(remove()));
    connect(ui->donePropertyButton, SIGNAL(released()), SLOT(done()));
    fillPropertyTable();
}

PorgyPropertyManager::~PorgyPropertyManager() {
    delete ui;
}

std::string PorgyPropertyManager::propertyAt(int row) {
    return QStringToTlpString(ui->propertyTable->item(row,0)->text());
}

tlp::Graph* PorgyPropertyManager::getGraph() {
    return graph;
}

void PorgyPropertyManager::setMatchingStatus(std::string propertyName, bool match) {
    std::unordered_set<std::string> matchingPropertiesList;
    PorgyTlpGraphStructure::getMatchingPropertyNames(graph, matchingPropertiesList);

    if (match && matchingPropertiesList.find(propertyName) == matchingPropertiesList.end()) {
        matchingPropertiesList.insert(propertyName);
    } else if (!match) {
        std::unordered_set<std::string>::const_iterator itP = matchingPropertiesList.find(propertyName);
        if (itP != matchingPropertiesList.end()) {
            matchingPropertiesList.erase(itP);
        }
    }
    PorgyTlpGraphStructure::setMatchingPropertyNames(graph, matchingPropertiesList);
}

void PorgyPropertyManager::create() {
    // Use the Tulip creation dialog to generate the property.
    // Propose to add a new boolean var to enable the list or impose a choice
    tlp::PropertyInterface *newProperty = PropertyCreationDialog::createNewProperty(graph->getRoot(), this, "string");
    if (newProperty != nullptr)
    {
        std::unordered_set<std::string> matchingPropertiesList;
        PorgyTlpGraphStructure::getMatchingPropertyNames(graph, matchingPropertiesList);
        matchingPropertiesList.insert(newProperty->getName());
        PorgyTlpGraphStructure::setMatchingPropertyNames(graph, matchingPropertiesList);

        tlp::BooleanProperty *stateProperty = PorgyTlpGraphStructure::getRulesRoot(graph)->getLocalProperty<BooleanProperty>(PorgyConstants::MATCHING_PROPERTIES_PREFIX + newProperty->getName() + PorgyConstants::MATCHING_PROPERTIES_SUFFIX);
        stateProperty->setAllNodeValue(true);
        PorgyTlpGraphStructure::getMatchingPropertyNames(graph, matchingPropertiesList);
    }
    fillPropertyTable();
    return;
}

void PorgyPropertyManager::fillPropertyTable(){
    vector<string> vec;
    PorgyTlpGraphStructure::getStandardPropertyNames(graph, vec);

    unordered_set<string> matchingPropertiesList;
    PorgyTlpGraphStructure::getMatchingPropertyNames(graph, matchingPropertiesList);

    bool forbidden_property = false;
    unsigned int table_indice = 0;

    for (unsigned int row = 0; row < vec.size(); ++row) {
        for(std::list<std::string>::const_iterator it_pmf = PorgyConstants::PROPERTY_MANAGER_FORBIDDEN.begin(); it_pmf != PorgyConstants::PROPERTY_MANAGER_FORBIDDEN.end(); ++it_pmf) {
            std::string tmp = *it_pmf;
            if (tmp.compare(vec[row]) == 0) {
                forbidden_property = true;
                break;
            }
        }
        if(!forbidden_property) {
            table_indice++;
        }
        forbidden_property = false;
    }

    ui->propertyTable->clearContents();
    ui->propertyTable->setColumnCount(2);
    ui->propertyTable->setRowCount(table_indice);
    ui->propertyTable->setItemDelegate(new PorgyPropertyManagerDelegate(this));
    ui->propertyTable->setHorizontalHeaderLabels(QStringList() << tr("Property") << tr("Test for match"));

    table_indice = 0;

    for (unsigned int row = 0; row < vec.size(); ++row) {
        for(const auto& tmp: PorgyConstants::PROPERTY_MANAGER_FORBIDDEN) {
            if (tmp == vec[row]) {
                forbidden_property = true;
                break;
            }
        }
        if(forbidden_property) {
            forbidden_property = false;
            continue;
        }
        QTableWidgetItem *item0 = new QTableWidgetItem(tlpStringToQString(vec[row]));
        item0->setFlags(item0->flags() ^ Qt::ItemIsEditable);
        ui->propertyTable->setItem(table_indice, 0, item0);

        QTableWidgetItem *item1;
        if (matchingPropertiesList.find(vec[row]) != matchingPropertiesList.end()) {
            item1 = new QTableWidgetItem(QString("true"));
        } else {
            item1 = new QTableWidgetItem(QString("false"));
        }
        item1->setTextAlignment(Qt::AlignRight);
        ui->propertyTable->setItem(table_indice, 1, item1);

        table_indice++;
    }
}

void PorgyPropertyManager::done() {
    QDialog::accept();
    return;
}

void PorgyPropertyManager::remove() {
    // Allows any suppression except for the forbidden properties (the standard tulip "view*" and those specified in PROPERTY_MANAGER_FORBIDDEN) whom are only reseted
    QList<QTableWidgetItem*> list = ui->propertyTable->selectedItems();
    bool forbidden_property = false;
    foreach(const QTableWidgetItem* qti, list) {
        std::string delPropertyName = propertyAt(qti->row());
        setMatchingStatus(delPropertyName, false);
        forbidden_property = false;
        for(std::list<std::string>::const_iterator it_pmf = PorgyConstants::PROPERTY_MANAGER_FORBIDDEN.begin(); it_pmf != PorgyConstants::PROPERTY_MANAGER_FORBIDDEN.end(); ++it_pmf) {
            std::string tmp = *it_pmf;
            if (tmp.compare(delPropertyName) == 0) {
                forbidden_property = true;
                break;
            }
        }
        std::size_t found = delPropertyName.find("view");
        if (found!=std::string::npos && found == 0)
            forbidden_property = true;
        if(!forbidden_property) {
            graph->getRoot()->delLocalProperty(delPropertyName);
        }

        if (PorgyTlpGraphStructure::getRulesRoot(graph)->existLocalProperty(PorgyConstants::MATCHING_PROPERTIES_PREFIX + delPropertyName + PorgyConstants::MATCHING_PROPERTIES_SUFFIX))
            PorgyTlpGraphStructure::getRulesRoot(graph)->delLocalProperty(PorgyConstants::MATCHING_PROPERTIES_PREFIX + delPropertyName + PorgyConstants::MATCHING_PROPERTIES_SUFFIX);
    }
    fillPropertyTable();
    return;
}

