/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <ui_strategymanagementwidget.h>

#include "strategymanagementwidget.h"

#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/Trace.h>

#include <porgy/porgymimedata.h>

#include <tulip/StringCollection.h>
#include <tulip/TlpQtTools.h>
#include <tulip/TulipMimes.h>
#include <tulip/BooleanProperty.h>
#include <tulip/StringProperty.h>

#include <QMimeData>
#include <QFileDialog>
#include <QMessageBox>
#include <QMenu>
#include <QDragEnterEvent>
#include <QFile>

using namespace tlp;
using namespace std;

StrategyManagementWidget::StrategyManagementWidget(QWidget *parent):QWidget(parent),ui(new Ui::StrategyManagementWidget),strategies(nullptr)
{
    ui->setupUi(this);

    ui->createStrategyPushButton->setIcon(QIcon::fromTheme("list-add",QIcon(":/porgyicons/list-add.png")));
    ui->removeStrategyPushButton->setIcon(QIcon::fromTheme("list-remove",QIcon(":/porgyicons/list-remove.png")));
    ui->exportStrategyPushButton->setIcon(QIcon::fromTheme("document-save",QIcon(":/porgyicons/document-save.png")));

    QMenu* menu = new QMenu(ui->createStrategyPushButton);
    menu->addAction(tr("Empty strategy"),this,SLOT(addNewEmptyStrategy()));
    menu->addAction(tr("Strategy from derivation tree"),this,SLOT(addStrategyFromSelectedElementsInTraceRoot()));
    menu->addAction(tr("Strategy from a file"),this,SLOT(importStrategy()));
    ui->createStrategyPushButton->setMenu(menu);

    connect(ui->removeStrategyPushButton,SIGNAL(clicked( bool)),this,SLOT(removeCurrentStrategy()));
    connect(ui->exportStrategyPushButton,SIGNAL(clicked( bool)),this,SLOT(exportCurrentStrategy()));

    ui->codeTextEdit->installEventFilter(this);
    //Update list view
    ui->strategiesListView->setModel(&strategies);
    connect(ui->strategiesListView->selectionModel(),SIGNAL(currentChanged(const QModelIndex& ,const QModelIndex &)),this,SLOT(selectedStrategyChanged(QModelIndex,QModelIndex)));
    selectStrategy(UINT_MAX);
}

//bool StrategyManagementWidget::getStepbyStepUndo() {
//    return ui->pushstepbystep->checked();
//}

StrategyManagementWidget::~StrategyManagementWidget(){
    delete ui;

}

void StrategyManagementWidget::setGraph(Graph *g){
    strategies.setGraph(g);
}

void StrategyManagementWidget::changeEvent(QEvent *e)
{
    QWidget::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}
QString StrategyManagementWidget::getCurrentStrategyCode()const{
    int index = currentStrategy();
    if(index != -1){
        StrategyManager::StrategyErrors error;
        QString strategy = strategies.getStrategyCode(index,error);
        if(error == StrategyManager::STRATEGY_NO_ERROR){
            return strategy;
        }else{
            tlp::warning()<<StrategyManager::errorMessage(error).toStdString() << endl;
            return QString();
        }
    }else{
        tlp::warning()<<"No strategy selected" << endl;
        return QString();
    }
}

void StrategyManagementWidget::addNewStrategy(QString name,QString strategy){
    if(name.isEmpty()){
        name = QString("new strategy");
    }
    QString origName = name;
    int cpt = 0;
    while(strategies.indexOfStrategy(name) != -1){
        ++cpt;
        name = origName +QString( "_")+QString::number(cpt);
    }
    StrategyManager::StrategyErrors error;
    unsigned int index =strategies.addStrategy(name,strategy,error);
    ui->strategiesListView->selectionModel()->setCurrentIndex(strategies.index(index),QItemSelectionModel::ClearAndSelect);
}

void StrategyManagementWidget::addNewEmptyStrategy(){
    addNewStrategy(QString("new strategy"),QString(""));

}

void StrategyManagementWidget::importStrategy(){
    QString selectedFile = QFileDialog::getOpenFileName(this,QString("Open a strategy"));
    if(QFile::exists(selectedFile)){
        importStrategy(selectedFile);
    }
}

void StrategyManagementWidget::importStrategy(const QString& filename){
    if(QFile::exists(filename)){
        QFile strategy(filename);
        if(!strategy.open(QIODevice::ReadOnly | QIODevice::Text)){
            tlp::warning()<<"Cannot open file \""<<filename.toStdString()<<"\"" << endl;
            return;
        }

        QString code = QString::fromUtf8(strategy.readAll().data());
        QFileInfo info(strategy);
        int index = strategies.indexOfStrategy(filename);
        if(index == -1){
            StrategyManager::StrategyErrors error;
            index = strategies.addStrategy(info.completeBaseName(),code,error);
            if(error == StrategyManager::STRATEGY_NO_ERROR){
                tlp::warning()<<"Strategy \""<<filename.toStdString() <<"\" imported" << endl;
            }else{
                tlp::error()<<"Error during import process of strategy \""<<filename.toStdString()<<"\"" << endl;
            }
        }else{
            int overwrite = QMessageBox::question(this,tr("Overwrite strategy?"),tr("A strategy with the same name already exists overwrite it?"),QMessageBox::Yes,QMessageBox::No);
            if(overwrite == QMessageBox::Yes){
                strategies.setStrategyCode(index,code);
            }
        }
        ui->strategiesListView->selectionModel()->setCurrentIndex(strategies.index(index),QItemSelectionModel::ClearAndSelect);
    }
}

void StrategyManagementWidget::removeCurrentStrategy(){
    int currentStrat = currentStrategy();
    if(currentStrat != -1){
        strategies.removeStrategy(currentStrat);
        if(strategies.count() == 0){
            displayStrategyCode(UINT_MAX);
        }
    }
}

void StrategyManagementWidget::selectedStrategyChanged(const QModelIndex & current, const QModelIndex & ){    
    displayStrategyCode(current.row());
}

void StrategyManagementWidget::selectStrategy(unsigned int index){
    if( index !=UINT_MAX && strategies.count() > index){
        ui->strategiesListView->selectionModel()->select(strategies.index(index),QItemSelectionModel::ClearAndSelect);
    }else{
        ui->strategiesListView->selectionModel()->clearSelection();
    }
    displayStrategyCode(index);
}

void StrategyManagementWidget::displayStrategyCode(unsigned int index){
    if(index !=UINT_MAX && strategies.count()>index){
        ui->codeTextEdit->setDocument(&strategies.getStrategyDocument(index));
        ui->codeTextEdit->setEnabled(true);
        ui->removeStrategyPushButton->setEnabled(true);
    }else{
        //Set a new text document to avoid invalid state if the previous text document is deleted.
        //The new QTextDocument must be a child of the QTextEdit to be automatically deleted when setDocument is called again.
        ui->codeTextEdit->setDocument(new QTextDocument(ui->codeTextEdit));
        ui->codeTextEdit->setEnabled(false);
        ui->removeStrategyPushButton->setEnabled(false);
    }
}

int StrategyManagementWidget::currentStrategy()const{
    QModelIndexList rows = ui->strategiesListView->selectionModel()->selectedRows();
    assert(rows.size() <2);
    if(rows.empty()){
        return -1;
    }else{
        return rows.front().row();
    }
}


void StrategyManagementWidget::exportCurrentStrategy(){
    QString fileName = QFileDialog::getSaveFileName(this,QString("Export a strategy"));
    if(!fileName.isEmpty()){
        exportCurrentStrategy(fileName);
    }
}

void StrategyManagementWidget::exportCurrentStrategy(const QString& filename){
    int currentStrat = currentStrategy();
    if(currentStrat!=-1){
        exportStrategy(currentStrat,filename);
    }
}
void StrategyManagementWidget::exportStrategy(int strategyIndex,const QString& filename){
    QFile file(filename);
    if (!file.open(QIODevice::WriteOnly | QIODevice::Text))
        return;

    QString strategy = strategies.getStrategy(strategyIndex);
    if(!strategy.isEmpty()){
        if(file.write(strategy.toUtf8()) != -1){
            tlp::warning() <<"Strategy saved in "<<filename.toStdString() << endl;
        }else{
            tlp::error()<<"Error during strategy export process" << endl;
        }
    }
    file.close();
}

void StrategyManagementWidget::importStrategies(const QString& dirName){
    QDir directory(dirName);
    if(directory.exists()){
        foreach(QString file,directory.entryList(QDir::Files | QDir::NoDot | QDir::NoDotDot, QDir::Name)){
            importStrategy(directory.absoluteFilePath(file));
        }
    }
}

void StrategyManagementWidget::exportStrategies(const QString& dirName){
    QDir directory(dirName);
    if(directory.exists()){
        for(int i = 0 ; i < strategies.rowCount() ; ++ i){
            exportStrategy(i,directory.absolutePath()+QDir::separator()+strategies.getStrategyName(i));
        }
    }
}

//DataSet StrategyManagementWidget::getData()const{
//    return strategies.getData();
//}

void StrategyManagementWidget::setData(const DataSet& data){
    strategies.setData(data);
    //If there is strategies select the first one
    selectStrategy(strategies.empty() ? -1 : 0);
}

bool StrategyManagementWidget::eventFilter(QObject *obj, QEvent *event){
    bool acceptEvent = false;
    if(obj == ui->codeTextEdit && ui->codeTextEdit->isEnabled()){
        if(event->type() == QEvent::DragEnter){
            QDragEnterEvent* dragEnterEvent = static_cast<QDragEnterEvent*>(event);
            const QMimeData* mimeData = dragEnterEvent->mimeData();
            if(mimeData->hasFormat(GRAPH_MIME_TYPE)){
                const GraphMimeType* gMimeType = static_cast<const GraphMimeType*>(mimeData);
                Graph* graphPointer = gMimeType->graph();
                //Not a rule graph ignore the drag enter event.
                if(!PorgyTlpGraphStructure::isRuleGraph(graphPointer)){
                    dragEnterEvent->setDropAction(Qt::IgnoreAction);
                    dragEnterEvent->accept();
                    return true;
                }
                acceptEvent = true;
                dragEnterEvent->acceptProposedAction();
            }
        }else if(event->type() == QEvent::Drop){
            QDropEvent *dropEvent = static_cast<QDropEvent*>(event);
            const QMimeData* mimeData = dropEvent->mimeData();
            QString codeToAppend;
            if(mimeData->hasFormat(GRAPH_MIME_TYPE)){
                const GraphMimeType* gMimeType = static_cast<const GraphMimeType*>(mimeData);
                Graph* graphPointer = gMimeType->graph();
                codeToAppend.append(tlpStringToQString(graphPointer->getName()));
            }
            //If we have code to append.
            if(!codeToAppend.isEmpty()){
                QString currentText = ui->codeTextEdit->toPlainText();
                if(!currentText.isEmpty()){
                    int currentChar = currentText.size()-1;
                    //Search the last non space char.
                    while(currentText[currentChar]==QChar(' ') && currentChar > 0){
                        --currentChar;
                    }
                    //If the last char is not a ; add it before inserting rule name.
                    if(currentChar!=0 && currentText[currentChar]!=QChar(';') && currentText[currentChar]!=QChar('(')&& currentText[currentChar]!=QChar(',')){
                        currentText.append(QChar(';'));
                    }
                }
                currentText.append(codeToAppend);
                strategies.updateStrategyCode(currentStrategy(),currentText);
            }
        }
        else if(event->type()==QEvent::Wheel) {
             QWheelEvent *ev =static_cast<QWheelEvent*>(event);
             if (ev->modifiers() &
     #if defined(__APPLE__)
                 Qt::AltModifier
     #else
                 Qt::ControlModifier
     #endif
                ) {
              int numDegrees = ev->delta()/8;
              int numSteps = numDegrees / 15;
             QFont f = ui->codeTextEdit->font();
             f.setPointSize(f.pointSize()+numSteps);
             ui->codeTextEdit->setFont(f);
             ui->strategiesListView->setFont(f);
             acceptEvent=true;
             }
         }
    }
    if(acceptEvent){
        event->accept();
        return true;
    }else{
        return QWidget::eventFilter(obj,event);
    }
}

//void StrategyManagementWidget::executeButtonPressed(){
//    int index = currentStrategy();
//    if(index != -1){
//        StrategyManager::StrategyErrors error;
//        QString strategy = strategies.getStrategyCode(index,error);
//        if(error == StrategyManager::STRATEGY_NO_ERROR){
//            if(!strategy.isEmpty()){
//                emit executeStrategy(strategy);
//            }else{
//                tlp::warning()<<"Cannot execute empty strategy. Input instructions before executing it." << endl;
//            }
//        }else{
//            tlp::warning()<<StrategyManager::errorMessage(error).toStdString() << endl;
//        }
//    }else{
//        tlp::warning()<<"No strategy selected" << endl;
//    }
//}

void StrategyManagementWidget::addStrategyFromSelectedElementsInTraceRoot(){
    Observable::holdObservers();
    Graph* traceRoot = PorgyTlpGraphStructure::getMainTrace(strategies.graph());
    string errMsg;
    DataSet data;
    data.set("Nodes",traceRoot->getProperty<BooleanProperty>("viewSelection"));
    BooleanProperty currentSelection(traceRoot);
    traceRoot->applyPropertyAlgorithm("Induced Sub-Graph",&currentSelection,errMsg,nullptr,&data);
    Trace trace(traceRoot);
    node firstNode,lastNode;
    string strategy = trace.computeStrategy(&currentSelection,&firstNode,&lastNode);
    if(strategy.empty()){
        QMessageBox::critical(this,QString("Error when computing strategy"),QString("An error occured when computing the stratgy"));
    }else{
        StringProperty *label = traceRoot->getProperty<StringProperty>("viewLabel");
        addNewStrategy(QString("Code from ")+tlpStringToQString(label->getNodeValue(firstNode))+QString(" to ")+tlpStringToQString(label->getNodeValue(lastNode)),tlpStringToQString(strategy));
    }
    Observable::unholdObservers();
}
