/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "strategymanager.h"

#include <typeinfo>

#include <tulip/ForEach.h>
#include <tulip/TlpQtTools.h>

#include <porgy/porgymimedata.h>

#include <QMimeData>

using namespace tlp;
using namespace std;

StrategyManager::StrategyManager(QObject* parent):QAbstractListModel(parent), _graph(nullptr)
{
}

void StrategyManager::setGraph(Graph *g){
    _graph=g;
}


unsigned int StrategyManager::count()const{
    return strategies.size();
}

bool StrategyManager::empty() const{
    return strategies.empty();
}


unsigned int StrategyManager::addStrategy(const QString& name,StrategyErrors& error){
    return addStrategy(name,"",error);
}
unsigned int StrategyManager::addStrategy(const QString& name, const QString& code,StrategyErrors& error){
    if(indexOfStrategy(name) == -1){
        if(name.isEmpty()){
            error = STRATEGY_EMPTY_NAME;
            return -1;
        }else{
            unsigned int newIndex = strategies.size();
            beginInsertRows(QModelIndex(),newIndex,newIndex);
            QTextDocument *document=new QTextDocument(this);
            document->setPlainText(code);
            document->setModified(false);
            strategies.push_back(make_pair(name,document));
            endInsertRows();
            error = STRATEGY_NO_ERROR;
            return newIndex;
        }
    }else{
        error = STRATEGY_EXISTING_NAME;
        return -1;
    }
}
void StrategyManager::removeStrategy(int index){
    assert(strategies.size()>index);
    beginRemoveRows(QModelIndex(),index,index);
    strategies[index].second->deleteLater();
    strategies.remove(index);
    endRemoveRows();
}

QString StrategyManager::getStrategyCode(int index,StrategyErrors& error) const{
    assert(strategies.size()>index);    
    QList<QString> expendedStrategies;
    QString strategyCode = expandStrategyCode(strategies[index].second->toPlainText(),expendedStrategies,error);
    if(error != STRATEGY_NO_ERROR){
        return QString();
    }else{
        return strategyCode;
    }
}

QString StrategyManager::getStrategy(int index) const{
    assert(strategies.size()>index);
    return strategies[index].second->toPlainText();
}

QTextDocument& StrategyManager::getStrategyDocument(int index){
    assert(strategies.size()>index);
    return *strategies[index].second;
}

QString StrategyManager::getStrategyName(int index)const{
    assert(strategies.size()>index);
    return strategies[index].first;
}

Strategy StrategyManager::getStrategy(int index,StrategyErrors& error) const{
    assert(strategies.size()>index);
    return Strategy(getStrategyName(index),getStrategyCode(index,error));
}

StrategyManager::StrategyErrors StrategyManager::setStrategyName(int currentIndex,const QString& newName){
    assert(strategies.size()>currentIndex);
    if(!newName.isEmpty()){
        int existingStrategyIndex = indexOfStrategy(newName);
        if(existingStrategyIndex == -1 || existingStrategyIndex == currentIndex ){
            strategies[currentIndex].first=newName;
            emit dataChanged(index(currentIndex),index(currentIndex));
            return STRATEGY_NO_ERROR;
        }else{
            return STRATEGY_EXISTING_NAME;
        }
    }else{
        return STRATEGY_EMPTY_NAME;
    }
}
void StrategyManager::setStrategyCode(int currentIndex,const QString& newCode){
    assert(strategies.size()>currentIndex);
    strategies[currentIndex].second->setPlainText(newCode);
    emit dataChanged(index(currentIndex),index(currentIndex));
}

DataSet StrategyManager::getData()const{
    DataSet data;
    unsigned int currentIndex=0;
    pair<QString,QTextDocument*> p;
    foreach (p, strategies) {
        DataSet temp;
        temp.set<string>("name",QStringToTlpString(p.first));
        temp.set<string>("code",QStringToTlpString(p.second->toPlainText()));
        data.set(QStringToTlpString(QString::number(currentIndex)),temp);
        ++currentIndex;
    }
    return data;
}
void StrategyManager::setData(const DataSet& data){
    beginResetModel();
    pair<QString,QTextDocument*> p;
    foreach (p, strategies) {
        p.second->deleteLater();
    }

    strategies.clear();
    pair<string,DataType*> value;
    forEach(value,data.getValues()){                
        if(value.second->getTypeName().compare(typeid(DataSet).name())==0){
            DataSet *temp = (DataSet*)value.second->value;
            if(temp->exist("name") && temp->exist("code")){
                string name;
                temp->get("name",name);
                string code;
                temp->get("code",code);
                StrategyErrors error;
                addStrategy(tlpStringToQString(name),tlpStringToQString(code),error);
                if(error!=STRATEGY_NO_ERROR){
                    tlp::warning() << "Error when loading strategy \""<<name<<"\" "<<errorMessage(error).toStdString() << endl;
                }
            }
        }
    }
    endResetModel();
}


void StrategyManager::updateStrategyCode(int currentIndex,const QString& newCode){
    assert(strategies.size()>currentIndex);
    strategies[currentIndex].second->setPlainText(newCode);
    emit dataChanged(index(currentIndex),index(currentIndex));
}


int StrategyManager::rowCount ( const QModelIndex & ) const{
    return count();
}
QVariant StrategyManager::data ( const QModelIndex & index, int role ) const{
    if(index.isValid()){
        if(role == Qt::DisplayRole){
            return getStrategyName(index.row());
        }
    }
    return QVariant();
}
QVariant StrategyManager::headerData ( int , Qt::Orientation , int role ) const{
    if(role == Qt::DisplayRole){
        return QVariant(QString("Name"));
    }
    return QVariant();
}
bool StrategyManager::setData ( const QModelIndex & curindex, const QVariant & value, int role){
    if(curindex.isValid()){
        if(role == Qt::EditRole){
            QString strategyName = value.toString();
            StrategyErrors errorCode = setStrategyName(curindex.row(),strategyName);
            if(errorCode == STRATEGY_NO_ERROR){
                return true;
            }else{
                tlp::warning()<<errorMessage(errorCode).toStdString() << endl;
                return false;
            }
        }
    }
    return false;
}

Qt::ItemFlags StrategyManager::flags ( const QModelIndex & index) const{
    return QAbstractListModel::flags(index)|Qt::ItemIsDragEnabled | Qt::ItemIsEditable;
}

QMimeData * StrategyManager::mimeData( const QModelIndexList & indexes ) const{
    PorgyMimeData *data= nullptr;
    foreach(const QModelIndex m, indexes) {
        if(m.isValid()){
            StrategyErrors error;
            Strategy strategy = getStrategy(m.row(),error);
            if(error ==STRATEGY_NO_ERROR && strategy.isValid()){
                if(data ==nullptr){
                    data = new PorgyMimeData();
                }
                data->setStrategy(strategy);
            }else{
                if(error != StrategyManager::STRATEGY_NO_ERROR){
                    tlp::warning() <<errorMessage(error).toStdString() << endl;
                }else{
                    tlp::warning() << "Cannot drag empty strategy. Input instructions before using it" << endl;
                }
            }
        }
    }
    return data;
}

QString StrategyManager::expandStrategyCode(const QString& metaStrategy,QList<QString>& alreadyExpendedStrategies,StrategyErrors& error)const{
    QRegExp rx("#[^#]*#");
    QString strategy = metaStrategy;
    int pos = 0;
    while ((pos = rx.indexIn(strategy, pos)) != -1) {
        //Get tokens
        QString strategyName = rx.cap(0);
        int length = strategyName.size();
        //Remove #
        strategyName = strategyName.mid(1,strategyName.size()-2);
        int strategyIndex = indexOfStrategy(strategyName);
        QString strategyCode;
        if(strategyIndex != -1){
            //Check if we have already expended this strategy to avoid cycle
            if(alreadyExpendedStrategies.indexOf(strategies[strategyIndex].first) != -1){
                error = STRATEGY_CIRCULAR_CALL;
                return "";
            }else{
                //Store the name of the properties we will expand to avoid circular inclusion
                alreadyExpendedStrategies.push_back(strategies[strategyIndex].first);
                strategyCode = expandStrategyCode(strategies[strategyIndex].second->toPlainText(),alreadyExpendedStrategies,error);
                alreadyExpendedStrategies.pop_back();
                if(error != STRATEGY_NO_ERROR){
                    return "";
                }
            }
        }else{
            error = STRATEGY_MISSING_STRATEGY;
            return "";
        }
        //Replace the # token by the strategy code.
        strategy.replace(pos,length,strategyCode);
        pos += strategyCode.size();
    }
    error = STRATEGY_NO_ERROR;
    return strategy;
}

int StrategyManager::indexOfStrategy(const QString& strategyName)const{
    for( int i  = 0 ; i<  strategies.size() ; ++i){
        if(strategies[i].first.compare(strategyName)==0){
            return i;
        }
    }
    return -1;
}
QString StrategyManager::errorMessage(StrategyErrors& error){
    switch(error){
    case STRATEGY_CIRCULAR_CALL:
        return QString("Cannot generate strategy with circular call");
        break;
    case STRATEGY_MISSING_STRATEGY:
        return QString("The strategy uses non existing strategy");
        break;
    case STRATEGY_EMPTY_NAME:
        return QString("Strategies cannot have an empty name");
        break;
    case STRATEGY_EXISTING_NAME:
        return QString("Strategies must have unique names");
        break;
    default:
        return QString();
    }

}
