/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "traceswidget.h"
#include "ui_traceswidget.h"
#include "NewTrace.h"
#include "qtimagetooltip.h"
#include "sortfilterproxymodel.h"

#include <QAbstractItemModel>
#include <QMenu>
#include <QHelpEvent>

#include <tulip/GlGraphRenderingParameters.h>
#include <tulip/BooleanProperty.h>
#include <tulip/ConnectedTest.h>

#include <portgraph/Trace.h>
#include <portgraph/PorgyTlpGraphStructure.h>

using namespace tlp;
using namespace std;

TracesWidget::TracesWidget(QWidget *parent) :QWidget(parent),ui(new Ui::TracesWidget),graph(nullptr),_model(nullptr)
{
    ui->setupUi(this);
    connect(ui->traceTreeView, SIGNAL(customContextMenuRequested(const QPoint&)),this, SLOT(ShowTreeContextMenu(const QPoint&)));
    connect(ui->traceTreeView, SIGNAL(doubleClicked(QModelIndex)),this, SLOT(indexDoubleClicked(QModelIndex)));

    _proxyModel = new SortFilterProxyModel(this);
    _proxyModel->setDynamicSortFilter(true);

    _model = new TracesItemModel(this);    
    GlGraphRenderingParameters parameters;
    parameters.setViewArrow(true);
    parameters.setEdgeColorInterpolate(false);
    parameters.setEdgeSizeInterpolate(true);
    parameters.setViewEdgeLabel(false);
    parameters.setViewNodeLabel(true);
    parameters.setViewMetaLabel(true);
    parameters.setLabelsDensity(0);
    parameters.setAntialiasing(true);
    parameters.setLabelScaled(true);
    parameters.setViewArrow(true);
    _model->snapshotManager().setRenderingParameters(parameters);
    _model->snapshotManager().setMetanodeRenderer(new GlMetaNodeRenderer(nullptr));
    _model->snapshotManager().setPreviewSize(QSize(32,32));

    _proxyModel->setSourceModel(_model);
    ui->traceTreeView->setModel(_proxyModel);
    ui->traceTreeView->installEventFilter(this);
    connect(ui->lineEdit,SIGNAL(textEdited(QString)),this,SLOT(filterTraces(QString)));
}

TracesWidget::~TracesWidget()
{
    delete ui;
}

QList<Graph*> TracesWidget::selectedTraces()const{
    QModelIndexList indexes = ui->traceTreeView->selectionModel()->selectedRows();
    QModelIndexList sourcesIndexes;
    foreach(QModelIndex m, indexes) {
        sourcesIndexes.push_back(_proxyModel->mapToSource(m));
    }
    return _model->indexesGraphs(sourcesIndexes);
}

void TracesWidget::addNewTrace() {
    Graph* newTrace = NewTrace::createNewTrace(graph,this);
    if(newTrace != nullptr){
        update();
        emit showTrace(newTrace,true);
    }
}

void TracesWidget::computeOnTrace() {
    QList<Graph*> graphs = selectedTraces();
    if(!graphs.empty()){
        Graph* newTrace = NewTrace::computeOnTrace(graphs.front(),this);
        if(newTrace != nullptr){
            update();
            //emit showTrace(newTrace,true);
        }
    }
}

void TracesWidget::setGraph(Graph* graph){
    this->graph = graph;    
    _model->setRootGraph(graph);
}

void TracesWidget::ShowTreeContextMenu(const QPoint& point)
{
    QMenu menu(this);

    Graph *selectedGraph = _model->indexGraph(_proxyModel->mapToSource(ui->traceTreeView->indexAt(point)));
    BooleanProperty* selectionProperty = graph->getProperty<BooleanProperty>("viewSelection");
    Iterator<node>* nodes  = selectionProperty->getNodesEqualTo(true,graph);

    QAction *showTrace = menu.addAction(tr("&Show"), this,SLOT(showTrace()));
    showTrace->setEnabled(selectedGraph != nullptr);

    QAction* createNewAction = menu.addAction(tr("&Create new"), this,SLOT(addNewTrace()));
    //If nodes are selected in the trace
    createNewAction->setEnabled(nodes->hasNext());

    QAction *deltrace = menu.addAction(tr("&Delete"), this,SLOT(deleteTrace()));
    //If there is a trace selected and if the selected trace is not the main trace graph.
    deltrace->setEnabled( selectedGraph != nullptr && PorgyTlpGraphStructure::getMainTrace(selectedGraph) != selectedGraph);

    QAction* updateMetricAction = menu.addAction(tr("&Update Metric"), this,SLOT(computeOnTrace()));
    //If nodes are selected in the trace
    updateMetricAction->setEnabled(nodes->hasNext());

    delete nodes;

    menu.exec(ui->traceTreeView->viewport()->mapToGlobal(point));
}

void TracesWidget::deleteTrace(){
    QList<Graph*> graphs = selectedTraces();
    if(!graphs.empty()){
        emit deleteTrace(graphs.front());
    }
}

void TracesWidget::showTrace(){    
    QList<Graph*> graphs = selectedTraces();
    if(!graphs.empty()){
        emit showTrace(graphs.front(),false);
    }
}

bool TracesWidget::eventFilter(QObject *obj, QEvent *evt){
    switch(evt->type()){
    case QEvent::ToolTip:
    {
        QHelpEvent *he = static_cast<QHelpEvent *>(evt);        
        if(obj == ui->traceTreeView){
            //Need to use viewport <idget to compute correct coordinate tranformation.
            //Return wrong QModelIndex if we use he->Pos();
            Graph *graph = nullptr;
            graph = _model->indexGraph(_proxyModel->mapToSource(ui->traceTreeView->indexAt(ui->traceTreeView->viewport()->mapFromGlobal(he->globalPos()))));
            if(graph){
                //Check if the trace graph have been already drawn.
                //If not init it's layout.
                if(!graph->existLocalProperty("viewLayout")){
                    string errMsg;
                    Trace trace(graph);
                    trace.redraw(errMsg,nullptr);
                }
                QtImageToolTip::showGraph(he->globalPos(), graph,QSize(512,512),Color(255,255,255),_model->snapshotManager().renderingParameters(),_model->snapshotManager().metanodeRenderer(),this,QRect());
                return true;
            }
        }
    }
        break;
    default:
        break;
    }

    return QWidget::eventFilter(obj,evt);
}

void TracesWidget::indexDoubleClicked(const QModelIndex& index){
    Graph* graph = _model->indexGraph(_proxyModel->mapToSource(index));
    if(graph){
        emit showTrace(graph,false);
    }
}

void TracesWidget::filterTraces(const QString& pattern){
    _proxyModel->setFilterWildcard(pattern);
}
