/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "AbstractPorgyGraphView.h"

#include <QApplication>
#include <QPixmap>

#include <tulip/GlScene.h>
#include <tulip/GlMainWidget.h>
#include <tulip/GlGraphComposite.h>
#include <tulip/GlMetaNodeRenderer.h>

#include <tulip/Gl2DRect.h>
#include <tulip/GlVertexArrayManager.h>
#include <tulip/DrawingTools.h>
#include <tulip/GlOverviewGraphicsItem.h>
#include <tulip/QtGlSceneZoomAndPanAnimator.h>
#include <tulip/SceneConfigWidget.h>
#include <tulip/TlpTools.h>
#include <tulip/OpenGlConfigManager.h>

using namespace tlp;
using namespace std;

AbstractPorgyGraphView::AbstractPorgyGraphView():GlMainView() {
}

AbstractPorgyGraphView::~AbstractPorgyGraphView() {
}

QList<QWidget*> AbstractPorgyGraphView::configurationWidgets() const{
    return QList<QWidget*>() << _sceneConfigurationWidget;
}

tlp::DataSet AbstractPorgyGraphView::state() const{
    GlScene *scene=getGlMainWidget()->getScene();
    DataSet outDataSet;
    outDataSet.set<DataSet>("Display",scene->getGlGraphComposite()->getRenderingParameters().getParameters());
    std::string out;
    scene->getXML(out);
    size_t pos=out.find(TulipBitmapDir);

    while(pos!=std::string::npos) {
        out.replace(pos,TulipBitmapDir.size(),"TulipBitmapDir/");
        pos=out.find(TulipBitmapDir);
    }

    outDataSet.set<std::string>("scene",out);
    //Overview
    outDataSet.set("overviewVisible",overviewVisible());
    return outDataSet;
}

void AbstractPorgyGraphView::setState(const tlp::DataSet& parameters){

    //Overview
    bool overviewVisible=false;
    if(parameters.exist("overviewVisible")) {
        parameters.get<bool>("overviewVisible",overviewVisible);
    }
    setOverviewVisible(overviewVisible);
    initScene(graph(), parameters);
    initObservers();
}

void AbstractPorgyGraphView::graphChanged(tlp::Graph* graph){
    clearObservers();
    GlScene *scene=getGlMainWidget()->getScene();

    if(!scene->getLayer("Main")) {//No main layer initialize the scene
        initScene(graph,DataSet());
    }else{
        GlGraphComposite* oldGraphComposite=(GlGraphComposite *)(scene->getLayer("Main")->findGlEntity("graph"));
        if(!oldGraphComposite) {//No composite initialize the scene
            initScene(graph,DataSet());
        }else{

            GlGraphRenderingParameters param=oldGraphComposite->getRenderingParameters();
            GlMetaNodeRenderer *metaNodeRenderer=oldGraphComposite->getInputData()->getMetaNodeRenderer();
            // prevent deletion of MetaNodeRenderer when deleting oldGraphComposite
            oldGraphComposite->getInputData()->setMetaNodeRenderer(nullptr,false);
            GlGraphComposite* graphComposite=new GlGraphComposite(graph);
            graphComposite->setRenderingParameters(param);

            metaNodeRenderer->setInputData(graphComposite->getInputData());

            graphComposite->getInputData()->setMetaNodeRenderer(metaNodeRenderer);

            if(oldGraphComposite->getInputData()->graph==graph) {
                delete graphComposite->getInputData()->getGlVertexArrayManager();
                graphComposite->getInputData()->setGlVertexArrayManager(oldGraphComposite->getInputData()->getGlVertexArrayManager());
                oldGraphComposite->getInputData()->setGlVertexArrayManager(nullptr);
                graphComposite->getInputData()->getGlVertexArrayManager()->setInputData(graphComposite->getInputData());
            }

            scene->getLayer("Main")->addGlEntity(graphComposite,"graph");

            delete oldGraphComposite;
            getGlMainWidget()->emitGraphChanged();

            GlGraphRenderingParameters parameters;
            initGlParameters(parameters);//Initialize parameters
            getGlMainWidget()->getScene()->getGlGraphComposite()->setRenderingParameters(parameters);
        }
    }
    initObservers();
    emit drawNeeded();
}

void AbstractPorgyGraphView::initScene(Graph *graph,const DataSet dataSet){

    GlScene *scene=getGlMainWidget()->getScene();
    scene->clearLayersList();

    std::string sceneInput="";

    if(dataSet.exist("scene")) {
        dataSet.get("scene",sceneInput);
    }

    if(sceneInput.empty()) {
        //Default scene
        GlLayer* layer=new GlLayer("Main");
        GlLayer *backgroundLayer=new GlLayer("Background");
        backgroundLayer->setVisible(false);
        GlLayer *foregroundLayer=new GlLayer("Foreground");
        foregroundLayer->setVisible(true);

        backgroundLayer->set2DMode();
        foregroundLayer->set2DMode();
        std::string dir=TulipBitmapDir;
        Gl2DRect *labri=new Gl2DRect(5.,5.,50.,50.,dir + "logolabri.jpg",true,false);
        labri->setVisible(false);
        foregroundLayer->addGlEntity(labri,"labrilogo");

        scene->addExistingLayer(backgroundLayer);
        scene->addExistingLayer(layer);
        scene->addExistingLayer(foregroundLayer);
        GlGraphComposite* graphComposite=new GlGraphComposite(graph,scene);
        scene->getLayer("Main")->addGlEntity(graphComposite,"graph");
        graphComposite->getRenderingParametersPointer()->setViewNodeLabel(true);
        graphComposite->getRenderingParametersPointer()->setEdgeColorInterpolate(false);
        graphComposite->getRenderingParametersPointer()->setNodesStencil(0x0002);
        graphComposite->getRenderingParametersPointer()->setNodesLabelStencil(0x0001);
        scene->centerScene();
    }
    else {
        size_t pos=sceneInput.find("TulipBitmapDir/");

        while(pos!=std::string::npos) {
            sceneInput.replace(pos,15,TulipBitmapDir);
            pos=sceneInput.find("TulipBitmapDir/");
        }

        pos=sceneInput.find("TulipLibDir/");

        while(pos!=std::string::npos) {
            sceneInput.replace(pos,12,TulipLibDir);
            pos=sceneInput.find("TulipLibDir/");
        }

        scene->setWithXML(sceneInput,graph);
    }

    if(dataSet.exist("Display")) {
        DataSet renderingParameters;
        dataSet.get("Display",renderingParameters);
        GlGraphRenderingParameters rp=scene->getGlGraphComposite()->getRenderingParameters();
        rp.setParameters(renderingParameters);
        scene->getGlGraphComposite()->setRenderingParameters(rp);
    }

    tlp::GlGraphRenderingParameters renderingParameters;
    initGlParameters(renderingParameters);
    scene->getGlGraphComposite()->setRenderingParameters(renderingParameters);
    scene->getGlGraphComposite()->getInputData()->setMetaNodeRenderer(new GlMetaNodeRenderer(scene->getGlGraphComposite()->getInputData()));
    OpenGlConfigManager::getInst().setAntiAliasing(true);
    getGlMainWidget()->emitGraphChanged();

}

void AbstractPorgyGraphView::centerWithAnimation(){
    if(overviewVisible())
        overviewItem()->draw(false);
    GlGraphInputData* inputData = getGlMainWidget()->getScene()->getGlGraphComposite()->getInputData();
    BoundingBox bbox = tlp::computeBoundingBox(inputData->getGraph(),inputData->getElementLayout(),inputData->getElementSize(),inputData->getElementRotation());
    QtGlSceneZoomAndPanAnimator zoomAndPan(getGlMainWidget(), bbox);
    zoomAndPan.animateZoomAndPan();
    getGlMainWidget()->getScene()->centerScene();
    emit drawNeeded();
}

void AbstractPorgyGraphView::initObservers(){
    if(getGlMainWidget()->getScene()->getGlGraphComposite() != nullptr){
        GlGraphInputData* inputData = getGlMainWidget()->getScene()->getGlGraphComposite()->getInputData();
        std::set<tlp::PropertyInterface*> properties = inputData->properties();
        for(auto& it : properties){
            addRedrawTrigger(it);
        }
        addRedrawTrigger(inputData->graph);
    }
}

void AbstractPorgyGraphView::clearObservers(){
    if(getGlMainWidget()->getScene()->getGlGraphComposite() != nullptr){
        GlGraphInputData* inputData = getGlMainWidget()->getScene()->getGlGraphComposite()->getInputData();
        std::set<tlp::PropertyInterface*> properties = inputData->properties();
        for(auto&  it : properties){
            removeRedrawTrigger(it);
        }
        removeRedrawTrigger(inputData->graph);
    }
}
