/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "glitem.h"

#include <QGraphicsScene>
#include <QGraphicsView>
#include <QPainter>
#include <QGLPixelBuffer>
#include <QGLWidget>

#include <tulip/GlTools.h>
#include <tulip/QGlPixelBufferManager.h>

GlItem::GlItem(QDeclarativeItem *parent):QDeclarativeItem(parent),_redrawNeeded(true),_renderingStore(nullptr)
{
    setFlag(QGraphicsItem::ItemHasNoContents ,false);
}

GlItem::~GlItem(){
    delete [] _renderingStore;
}

void setRasterPosition(unsigned int x, unsigned int y) {
    float val[4];
    unsigned char tmp[10];
    glGetFloatv(GL_CURRENT_RASTER_POSITION, (float*)&val);
    glBitmap(0,0,0,0,-val[0] + x, -val[1] + y, tmp);
    glGetFloatv(GL_CURRENT_RASTER_POSITION, (float*)&val);
    tlp::glTest(__PRETTY_FUNCTION__);
}

void GlItem::paint(QPainter *painter, const QStyleOptionGraphicsItem *, QWidget *){
    if(QDeclarativeItem::scene() != nullptr && _size.width()!=0 && _size.height()  != 0){        
        painter->beginNativePainting();
        glPushAttrib(GL_ALL_ATTRIB_BITS);

        glMatrixMode(GL_PROJECTION);
        glPushMatrix();

        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();

        //Render the graph in the back buffer.
        QPointF p = mapToScene(pos());//Translate in scene position

        float vpX = p.x();
        float vpY = QDeclarativeItem::scene()->height() - (p.y() + _size.height());
        float vpW = _size.width();
        float vpH = _size.height();
        _scene.setClearBufferAtDraw(false);
        _scene.setViewport(vpX,vpY,vpW,vpH);

        if(_redrawNeeded) {
            _scene.draw();
            glDisable(GL_TEXTURE_2D);
            glDisable(GL_DEPTH_TEST);
            glDisable(GL_STENCIL_TEST);
            glDisable(GL_BLEND);
            glDisable(GL_LIGHTING);

            glReadBuffer(GL_BACK);
            glReadPixels(vpX,vpY,vpW,vpH,GL_RGBA,GL_UNSIGNED_BYTE,_renderingStore);
            glFlush();

            _redrawNeeded=false;
        }
        else {

            glDisable(GL_TEXTURE_2D);
            glDisable(GL_DEPTH_TEST);
            glDisable(GL_STENCIL_TEST);
            glDisable(GL_BLEND);
            glDisable(GL_LIGHTING);

            glDrawBuffer(GL_BACK);
            setRasterPosition(vpX,vpY);
            glDrawPixels(vpW,vpH,GL_RGBA,GL_UNSIGNED_BYTE,_renderingStore);
            glFlush();
        }
        _scene.setClearBufferAtDraw(true);
        glFlush();


        glMatrixMode(GL_MODELVIEW);
        glPopMatrix();

        glMatrixMode(GL_PROJECTION);
        glPopMatrix();

        glPopAttrib();
        painter->endNativePainting();
    }
}

void GlItem::geometryChanged(const QRectF &newGeometry, const QRectF &oldGeometry){
    _size = newGeometry.size();
    QDeclarativeItem::geometryChanged(newGeometry,oldGeometry);
    updateGlWidgetGeometry(newGeometry);
}

void GlItem::updateGlWidgetGeometry(const QRectF &newGeometry){
    delete [] _renderingStore;
    _renderingStore = new unsigned char[(int)(newGeometry.width()*newGeometry.height()*4)];
    _redrawNeeded = true;
}

bool GlItem::pickGlEntities(const int x, const int y,
                            const int width, const int height,
                            std::vector<tlp::SelectedEntity> &pickedEntities,
                            tlp::GlLayer* layer) {
    makeCurrent();
    return _scene.selectEntities((tlp::RenderingEntitiesFlag)(tlp::RenderingSimpleEntities | tlp::RenderingWithoutRemove),x, y,
                                 width, height,
                                 layer,
                                 pickedEntities);
}
//==================================================
bool GlItem::pickGlEntities(const int x, const int y,
                            std::vector <tlp::SelectedEntity> &pickedEntities,
                            tlp::GlLayer* layer) {
    return pickGlEntities(x,y,2,2,pickedEntities,layer);
}
//==================================================
void GlItem::pickNodesEdges(const int x, const int y,
                            const int width ,const int height,
                            std::vector<tlp::SelectedEntity> &selectedNodes, std::vector<tlp::SelectedEntity> &selectedEdges,
                            tlp::GlLayer* layer) {
    makeCurrent();
    _scene.selectEntities((tlp::RenderingEntitiesFlag)(tlp::RenderingNodes | tlp::RenderingWithoutRemove), x, y, width, height, layer, selectedNodes);
    _scene.selectEntities((tlp::RenderingEntitiesFlag)(tlp::RenderingEdges | tlp::RenderingWithoutRemove), x, y, width, height, layer, selectedEdges);
}
//=====================================================
bool GlItem::pickNodesEdges(const int x, const int y,tlp::SelectedEntity &selectedEntity, tlp::GlLayer* layer) {
    makeCurrent();
    std::vector<tlp::SelectedEntity> selectedEntities;

    if(_scene.selectEntities((tlp::RenderingEntitiesFlag)(tlp::RenderingNodes | tlp::RenderingWithoutRemove), x-1, y-1, 3, 3, layer, selectedEntities)) {
        selectedEntity=selectedEntities[0];
        return true;
    }

    if(_scene.selectEntities((tlp::RenderingEntitiesFlag)(tlp::RenderingEdges | tlp::RenderingWithoutRemove), x-1, y-1, 3, 3, layer, selectedEntities)) {
        selectedEntity=selectedEntities[0];
        return true;
    }

    return false;
}

void GlItem::makeCurrent(){
    if(!QDeclarativeItem::scene()->views().empty()){
        QGLWidget* viewPort = dynamic_cast<QGLWidget*>(QDeclarativeItem::scene()->views().first()->viewport());
        if(viewPort){
            viewPort->makeCurrent();
            _scene.setViewport(0,0,_size.width(),_size.height());
        }
    }
}

QPixmap GlItem::createPixmap(const QSize& size){  

    QImage resultImage;

    QPointF p = mapToScene(pos());//Translate in scene position



    _scene.setViewport(0,0,size.width(),size.height());

#ifndef WITHOUT_QT_PICTURE_OUTPUT

    QGLPixelBuffer *glFrameBuf=tlp::QGlPixelBufferManager::getInst().getPixelBuffer(size.width(),size.height());

    glFrameBuf->makeCurrent();
    _scene.draw();
    resultImage=glFrameBuf->toImage();
#else
    makeCurrent();

    if(center)
        scene.ajustSceneToSize(width,height);

    _scene.prerenderMetaNodes();

    unsigned char *image = (unsigned char *)malloc(width*height*3*sizeof(unsigned char));
    computeInteractors();
    scene.draw();
    drawInteractors();
    glFlush();
    glFinish();
    glPixelStorei(GL_PACK_ALIGNMENT,1);
    glReadPixels(0,0,width,height,GL_RGB,GL_UNSIGNED_BYTE,image);

    QPixmap pm(width,height);
    QPainter painter;
    painter.begin(&pm);

    for (int y=0; y<height; y++) {
        for (int x=0; x<width; x++) {
            painter.setPen(QColor(image[(height-y-1)*width*3+(x)*3],
                                  image[(height-y-1)*width*3+(x)*3+1],
                                  image[(height-y-1)*width*3+(x)*3+2]));
            painter.drawPoint(x,y);
        }
    }

    painter.end();
    free(image);
    resultImage= pm.toImage();
#endif
    float vpX = p.x();
    float vpY = QDeclarativeItem::scene()->height() - (p.y() + _size.height());
    float vpW = _size.width();
    float vpH = _size.height();
     _scene.setViewport(vpX,vpY,vpW,vpH);


    //The QGlPixelBuffer returns the wrong image format QImage::Format_ARGB32_Premultiplied. We need to create an image from original data with the right format QImage::Format_ARGB32.
    //We need to clone the data as when the image var will be destroy at the end of the function it's data will be destroyed too and the newly created image object will have invalid data pointer.
    return QPixmap::fromImage(QImage(resultImage.bits(),resultImage.width(),resultImage.height(),QImage::Format_ARGB32));
}
