/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "glitemnavigationinteractor.h"
#include "glitem.h"

#include <QEvent>
#include <QGraphicsSceneMouseEvent>
#include <QKeyEvent>
#include <QGraphicsScene>

GlItemNavigationController::GlItemNavigationController(QObject* parent):QObject(parent),x(0), y(0),inRotation(false),inZoom(false),_interactions(All_Interaction)
{
}

GlItemNavigationController::GlItemNavigationController(MouseInteractions interactions,QObject* parent):QObject(parent),_interactions(interactions){

}

void GlItemNavigationController::install(GlItem* item){
    installEventFilter(item);    
}

#define WHEEL_DELTA 120

bool GlItemNavigationController::eventFilter(QObject *obj, QEvent *e){
    GlItem* glItem = qobject_cast<GlItem*>(obj);
    if(glItem == nullptr){
        return false;
    }

    bool update =false;

    if(e->type() == QEvent::GraphicsSceneMousePress){        
        QGraphicsSceneMouseEvent* me = static_cast<QGraphicsSceneMouseEvent*>(e);
        x = me->pos().x();
        y = me->pos().y();
        inZoom = false;
        inRotation = false;
        return true;
    }else if (e->type() == QEvent::GraphicsSceneMouseMove) {
        QGraphicsSceneMouseEvent* me = static_cast<QGraphicsSceneMouseEvent*>(e);
        if(me->modifiers() == Qt::KeyboardModifiers(Qt::ShiftModifier)){
            if(_interactions.testFlag(Rotation)){
                int deltaX,deltaY;
                deltaX=me->pos().x()-x;
                deltaY=me->pos().y()-y;

                if (abs(deltaX)>abs(deltaY))
                    deltaY=0;
                else
                    deltaX=0;

                if (deltaY!=0){
                    glItem->scene().rotateScene(deltaY,0,0);
                    update = true;
                }

                if (deltaX!=0) {
                    glItem->scene().rotateScene(0,deltaX,0);
                    update = true;
                }
            }
        }else if(me->modifiers() == Qt::ControlModifier){
            if(_interactions.testFlag(Rotation)){
                int deltaX,deltaY;

                if(!inRotation && !inZoom) {
                    deltaX = me->pos().x() - x;
                    deltaY= me->pos().y() - y;

                    if (deltaY && abs(deltaX) >= 3 * abs(deltaY)) {
                        inRotation=true;
                        inZoom=false;
                    }
                    else if (deltaX && abs(deltaY) >=  3 * abs(deltaX)) {
                        inZoom=true;
                        inRotation=false;
                    }
                    else {

                    }

                    x = me->pos().x();
                    y = me->pos().y();
                }

                if (inZoom) {
                    // Zoom
                    deltaY = me->pos().y() - y;
                    glItem->scene().zoom(-deltaY/2);
                    y = me->pos().y();
                    update = true;
                }

                if(inRotation) {
                    // Rotation
                    deltaX = me->pos().x() - x;
                    glItem->scene().rotateScene(0,0,deltaX);
                    x = me->pos().x();
                    update = true;
                }
            }
        }else if(me->modifiers() == Qt::KeyboardModifier()){//No modifiers
            if(_interactions.testFlag(Translation)){
                if (me->pos().x() != x){
                    glItem->scene().translateCamera(me->pos().x()-x,0,0);
                    update = true;
                }

                if (me->pos().y() != y){
                    glItem->scene().translateCamera(0,y-me->pos().y(),0);
                    update = true;
                }
            }

        }

        x=me->pos().x();
        y=me->pos().y();

    }else if(e->type() == QEvent::GraphicsSceneWheel){
        if(_interactions.testFlag(Zoom)){
            QGraphicsSceneWheelEvent* we = static_cast<QGraphicsSceneWheelEvent*>(e);
            if (we->orientation() == Qt::Vertical) {
                glItem->scene().zoomXY(we->delta() / WHEEL_DELTA,we->pos().x(), we->pos().y());
                update = true;
            }
        }
    }else if (e->type() == QEvent::KeyPress) {
        int delta = (static_cast<QKeyEvent *>(e)->isAutoRepeat() ? 3 : 1);

        switch(((QKeyEvent *) e)->key()) {
        case Qt::Key_Left:
            if(_interactions.testFlag(Translation)){
                glItem->scene().translateCamera(delta * 2,0,0);
                update = true;
            }
            break;

        case Qt::Key_Right:
            if(_interactions.testFlag(Translation)){
                glItem->scene().translateCamera(-1 * delta * 2,0,0);
                update = true;
            }
            break;

        case Qt::Key_Up:
            if(_interactions.testFlag(Translation)){
                glItem->scene().translateCamera(0,-1 * delta * 2,0);
                update = true;
            }
            break;

        case Qt::Key_Down:
            if(_interactions.testFlag(Translation)){
                glItem->scene().translateCamera(0,delta * 2,0);
                update = true;
            }
            break;

        case Qt::Key_PageUp:
            if(_interactions.testFlag(Zoom)){
                glItem->scene().zoom(delta);
                update = true;
            }
            break;

        case Qt::Key_PageDown:
            if(_interactions.testFlag(Zoom)){
                glItem->scene().zoom(-1 * delta);
                update = true;
            }
            break;

        case Qt::Key_Home:
            if(_interactions.testFlag(Translation)){
                glItem->scene().translateCamera(0,0,-1 * delta * 2);
                update = true;
            }
            break;

        case Qt::Key_End:
            if(_interactions.testFlag(Translation)){
                glItem->scene().translateCamera(0,0,delta * 2);
                update = true;
            }
            break;

        case Qt::Key_Insert:
            if(_interactions.testFlag(Rotation)){
                glItem->scene().rotateScene(0,0,-1 * delta * 2);
                update = true;
            }
            break;

        case Qt::Key_Delete :
            if(_interactions.testFlag(Rotation)){
                glItem->scene().rotateScene(0,0,delta * 2);
                update = true;
            }
            break;

        default:
            break;
        }        
    }    
    if(update){//Ugly hack
        glItem->draw();
        glItem->QGraphicsItem::scene()->update();
    }
    return update;
}
