/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef PORGYSMALLMULTIPLESTRACEVIEW_H
#define PORGYSMALLMULTIPLESTRACEVIEW_H

#include "graphstatemanager.h"

#include <tulip/View.h>
#include <tulip/GlGraphRenderingParameters.h>

#include <QAbstractAnimation>
#include <QDeclarativeView>

namespace tlp{
class Graph;
class GlMainWidget;
class GlScene;
}

class QAction;

class PorgySmallMultiplesConfigurationWidget;
class AnimationConfigurationWidget;
class SmallMultiplesConfigurationWidget;
class GlItem;

enum SmallMultiplesLayout{Automatic,ForceLineNumber};

/**
 * @brief The RulePreviewGeometry enum describe the geometry of the rule preview widget.
 */
enum RulePreviewGeometry{No_Rule_Preview /** The preview widget is hidden */
                         ,Small_Rule_Preview
                         ,Medium_Rule_Preview
                         ,Large_Rule_Preview};

/**
 * @brief The DeclarativeGraphicsView class is a specialization of the QDeclarative view dedicated to TraceAnimationView
 */
class DeclarativeGraphicsView : public QDeclarativeView{

    GlItem* _animationItem;//The gl item displaying the current state and the morphing between two states.
    GlItem* _smallMultipleItem;//Display the small multiples.
    GlItem* _rulePreviewItem;//Display the rule preview widget.
    QObject* _stateName;//Display the name of the current state
public:

    enum RemoteMode{Playing,Paused};

    DeclarativeGraphicsView(QWidget* parent=nullptr);
    void resizeEvent(QResizeEvent *event);
    GlItem* animationItem();
    GlItem* smallMultipleItem();
    GlItem* rulePreviewItem();
    QObject* stateName();

    void setRemoteMode(RemoteMode mode);

    void setRulePreviewVisible(bool visible);
    void setAnimationVisible(bool visible);
    void setSmallMultiplesVisible(bool visible);
    bool getAnimationVisible() const;
    bool getSmallMultiplesVisible() const;
    bool getRulePreviewVisible() const;

    void setStateName(const QString& name);
    void setRulePreviewName(const QString& name);

    void setRulePreviewGeometry(RulePreviewGeometry geometry);
};

/**
  * @brief View to visualize trace graph evolution with animation and small multiples. This view can only display Trace graphs with only one single path. If a node in the trace graph have more than one child the view don't display the graph.
  **/
class PorgySmallMultiplesTraceView : public tlp::View
{
    Q_OBJECT

    friend class GlSceneMorphing;

    QAction *_forceRedrawAction;

    DeclarativeGraphicsView* _graphicsView;//The declarative graphics view. A special implementation is used as the GlItem object need to have the whole scene displayed in the view to work correctly.

    //Configuration widgets
    PorgySmallMultiplesConfigurationWidget *_configurationWidget;//The generic configuration widget.
    AnimationConfigurationWidget *_animationConfigurationWidget;
    SmallMultiplesConfigurationWidget *_smallMultiplesConfigurationWidget;

    unsigned _currentState;//The index of the current state in the GraphStateManager
     QAbstractAnimation* _currentAnimation;//The current animation.
     bool _goNextStep;//Flag to control if at the end of an animation we had to go to the next step.
    GraphStateManager _states;//The states
    GraphVisualProperties _visualProperties;//Contain the overloaded visual properties to perform animations without having to modify original graph properties.

    tlp::Graph* _smallMultipleGraph;//The graph to handle small multiples elements
    std::vector<std::string> _textures;//The texture createed for the smallmultiples

    enum SmallMultipleUpdateType{
        Update_Name=0x1,
        Update_texture=0x2, // Update the preview of the element use it after a graph changed
        Update_Layout=0x4, //Update the layout of the element. Call after changing the spacing or the number of row
        Reset=0xF//Reset all elements update the number of nodes and regenerate all the visual properties. Call after updating states list
    };
    Q_DECLARE_FLAGS(SmallMultipleUpdateTypeFlags,SmallMultipleUpdateType)


public:
    PLUGININFORMATION(PorgyConstants::SMALL_MULTIPLE_VIEW_NAME,"Jonathan Dubois","17/01/2011","","1.0",PorgyConstants::CATEGORY_NAME)
    PorgySmallMultiplesTraceView(const tlp::PluginContext*);
    virtual ~PorgySmallMultiplesTraceView();

    QGraphicsView* graphicsView() const;

    QList<QWidget*> configurationWidgets() const;
    QString configurationWidgetsStyleSheet() const;
    bool eventFilter(QObject* obj, QEvent* event);
    QGraphicsItem* centralItem() const; //mandatory to have the configuration widgets visible

    GlItem* smallMultipleItem()const{
        return _graphicsView->smallMultipleItem();
    }

    tlp::DataSet state() const;

    QPixmap snapshot(const QSize&) const;
    void setState(const tlp::DataSet&);

    void draw();
    void setupUi();

protected slots:
    void graphChanged(tlp::Graph*);
    void graphDeleted(tlp::Graph*){

    }


private:    
    tlp::GlGraphRenderingParameters renderingParameters()const;        

    void fillContextMenu(QMenu* menu,const QPointF &);

    /**
     * @brief setAnimationData updates the data displayed by the animation item. This function is used to switch the graph during the animation process.
     * @param graph
     * @param properties
     */
    void setAnimationData(tlp::Graph* graph,GraphVisualProperties& properties);

    /**
     * @brief loadGraphInScene intialise the scene if there is no existing graph layer or change the graph in the scene.
     * @param scene the scene to update
     * @param graph the graph to display
     */
    void loadGraphInScene(tlp::GlScene& scene,tlp::Graph* graph);
    /**
     * @brief removeRedrawTriggers removes all the redraw triggers for the graph in this scene
     * @param scene
     */
    void removeRedrawTriggers(tlp::GlScene& scene);

    /**
     * @brief removeRedrawTriggers add the camera, the graph and all the graph properties as redraw triggers.
     * @param scene
     */
    void addRedrawTriggers(tlp::GlScene& scene);
     std::string getTextureNameForId(int stateId);
     std::string generateTexture(unsigned int stateId);
     void clearTextures();


    private slots:
      void openSnapshotDialog();
        //Animation control functions.
        /**
         * @brief start Begin to animate the states from the current state to the final state or until the user click on pause or stop
         */
        void start();
        /**
         * @brief pause Pause the current animation. Click on play to launch it again.
         */
        void pause();
        /**
         * @brief stop Stop the animation
         */
        void stop();
        /**
         * @brief seekBackward go one step backwark from the current state if possible
         */
        void seekBackward();
        /**
         * @brief seekForward go one step forward from the current state if possible
         */
        void seekForward();
        /**
         * @brief seekFirst go to the first state
         */
        void seekFirst();
        /**
         * @brief seekFirst go to the last state
         */
        void seekLast();
        void scalelabels(bool state);
        void updateNodeLabesVisibility(bool visible);
        void scaleSmallMultiplesTitles(bool scale);
        void spacechanged(double space);

        /**
          * @brief Handle the animation states.
          **/
        void currentAnimationStateChanged(QAbstractAnimation::State newState, QAbstractAnimation::State oldState);

        /**
          * @brief Force to redraw the layout of the small multiples.
          **/
        void updateSmallMultipleLayout();
        void displayAnimationWidget(bool displayWidget);
        void displaySmallMultipleWidget(bool displayWidget);

        /**
          * @brief Force to regenerate intermediary states
          **/
        void updateIntermediateStates();

        void rulePreviewGeometryChanged();

        /**
         * @brief updateSmallMultiples updates the small multiple graph and texture with the content of the _state attribute.
         */
        void updateSmallMultiples(SmallMultipleUpdateTypeFlags updateType);

    private:
        /**
          * @brief Animate from the current state to the new state.
          **/
        void animate(unsigned int to);
        /**
          * @brief Build a QabstractAnimation between the current state and the to state. If an animation is already running stop them. The current state is built from the displayed graph and visual properties.
          * @param to The final state
          **/
       QAbstractAnimation* buildMorphingBetweenStates( unsigned int to);        

};

#endif // PORGYSMALLMULTIPLESTRACEVIEW_H

