/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef PROPERTYVALUETOQVARIANTCONVERTER_H
#define PROPERTYVALUETOQVARIANTCONVERTER_H

#include <tulip/PropertyInterface.h>
#include <tulip/BooleanProperty.h>
#include <tulip/TulipMetaTypes.h>

#include <QVariant>

Q_DECLARE_METATYPE(std::vector<bool>)

/**
 * @brief The PropertyValueToQVariantConverter class convert properties values to QVariant object
 */
class PropertyValueToQVariantConverter
{
public:
    virtual ~PropertyValueToQVariantConverter();
    /**
     * @brief value Returns the QVariant object for the given type
     * @param id
     * @param property
     * @return
     */
    virtual QVariant value(unsigned int id,tlp::PropertyInterface* property,int role=Qt::EditRole) const=0;
    /**
     * @brief defaultValue
     * @param property
     * @return
     */
    virtual QVariant defaultValue(tlp::PropertyInterface* property,int role=Qt::EditRole) const=0;
    /**
     * @brief setValue Set the value in the qvariant
     * @param id
     * @param property
     * @param data
     * @return
     */
    virtual bool setValue(unsigned int id,tlp::PropertyInterface* property,const QVariant& data,int role=Qt::EditRole) const=0;
    /**
     * @brief setAllValue
     * @param property
     * @param data
     * @return
     */
    virtual bool setAllValue(tlp::PropertyInterface* property,const QVariant& data,int role=Qt::EditRole) const=0;


    virtual Qt::ItemFlags defaultFlags(tlp::PropertyInterface*){
        return Qt::ItemFlags(Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsEditable);
    }

    virtual Qt::ItemFlags flags(unsigned int,tlp::PropertyInterface*p){
        return defaultFlags(p);
    }
};

/**
 * @brief The EmptyPropertyValueToQVariantConverter class is an implementation of the PropertyValueToQVariantConverter that return invalid values.
 */
class EmptyPropertyValueToQVariantConverter : public PropertyValueToQVariantConverter{
public:
    QVariant value(unsigned int ,tlp::PropertyInterface* ,int ) const{
        return QVariant();
    }
    QVariant defaultValue(tlp::PropertyInterface* ,int ) const{
        return QVariant();
    }
    bool setValue(unsigned int ,tlp::PropertyInterface* ,const QVariant& ,int ) const{
        return false;
    }
    bool setAllValue(tlp::PropertyInterface* ,const QVariant& ,int ) const{
        return false;
    }
    virtual Qt::ItemFlags defaultFlags(tlp::PropertyInterface*){
        return Qt::ItemFlags();
    }
    virtual Qt::ItemFlags flags(unsigned int,tlp::PropertyInterface*){
        return Qt::ItemFlags();
    }
};

/**
 * @brief Template converter for nodes.
 */
template <typename PROPERTYTYPE,typename T>
class NodeValueToQVariantConverter : public PropertyValueToQVariantConverter{
public:
    QVariant value(unsigned int id, tlp::PropertyInterface *property,int role) const{
        if(role == Qt::DisplayRole || role == Qt::EditRole){
            PROPERTYTYPE* p = dynamic_cast<PROPERTYTYPE*>(property);
            return p == nullptr? QVariant():QVariant::fromValue<T>(p->getNodeValue(tlp::node(id)));
        }else{
            return QVariant();
        }
    }

    bool setValue(unsigned int id, tlp::PropertyInterface *property, const QVariant &data,int role) const{
        if(role == Qt::EditRole){
            PROPERTYTYPE* p = dynamic_cast<PROPERTYTYPE*>(property);
            if(p == nullptr){
                return false;
            }
            p->setNodeValue(tlp::node(id),data.value<T>());
            return true;
        }else{
            return false;
        }
    }

    QVariant defaultValue(tlp::PropertyInterface* property,int role) const{
        if( role == Qt::DisplayRole || role == Qt::EditRole){
            PROPERTYTYPE* p = dynamic_cast<PROPERTYTYPE*>(property);
            return p == nullptr? QVariant():QVariant::fromValue<T>(p->getNodeDefaultValue());
        }else{
            return QVariant();
        }
    }
    bool setAllValue(tlp::PropertyInterface* property,const QVariant& data,int role) const{
        if(role == Qt::EditRole){
            PROPERTYTYPE* p = dynamic_cast<PROPERTYTYPE*>(property);
            if(p == nullptr){
                return false;
            }
            p->setAllNodeValue(data.value<T>());
            return true;
        }else{
            return false;
        }
    }
};

/**
 * @brief Uneditable converter for nodes.
 */
template <typename PROPERTYTYPE,typename T>
class UneditableNodeValueToQVariantConverter : public NodeValueToQVariantConverter<PROPERTYTYPE,T>{
public:
    virtual Qt::ItemFlags defaultFlags(tlp::PropertyInterface*){
        return Qt::ItemFlags(Qt::ItemIsEnabled | Qt::ItemIsSelectable );
    }
};

/**
 * @brief Template converter for nodes.
 */
template <typename PROPERTYTYPE,typename T>
class EdgeValueToQVariantConverter : public PropertyValueToQVariantConverter{
public:
    QVariant value(unsigned int id, tlp::PropertyInterface *property,int role) const{
        if(role == Qt::DisplayRole || role == Qt::EditRole  ){
            PROPERTYTYPE* p = dynamic_cast<PROPERTYTYPE*>(property);
            return p == nullptr? QVariant():QVariant::fromValue<T>(p->getEdgeValue(tlp::edge(id)));
        }else{
            return QVariant();
        }
    }

    bool setValue(unsigned int id, tlp::PropertyInterface *property, const QVariant &data,int role) const{
        if(role == Qt::EditRole  ){
            PROPERTYTYPE* p = dynamic_cast<PROPERTYTYPE*>(property);
            if(p == nullptr){
                return false;
            }
            p->setEdgeValue(tlp::edge(id),data.value<T>());
            return true;
        }
        return false;
    }

    QVariant defaultValue(tlp::PropertyInterface* property,int role) const{
        if(role == Qt::EditRole  ){
            PROPERTYTYPE* p = dynamic_cast<PROPERTYTYPE*>(property);
            return p == nullptr? QVariant():QVariant::fromValue<T>(p->getEdgeDefaultValue());
        }else{
            return QVariant();
        }
    }
    bool setAllValue(tlp::PropertyInterface* property,const QVariant& data,int role) const{
        if(role == Qt::EditRole){
            PROPERTYTYPE* p = dynamic_cast<PROPERTYTYPE*>(property);
            if(p == nullptr){
                return false;
            }
            p->setAllEdgeValue(data.value<T>());
            return true;
        }else{
            return false;
        }
    }
};

/**
 * Uneditable value converter.
 */
template <typename PROPERTYTYPE,typename T>
class UneditableEdgeValueToQVariantConverter : public EdgeValueToQVariantConverter<PROPERTYTYPE,T>{
public:
    virtual Qt::ItemFlags defaultFlags(tlp::PropertyInterface*){
        return Qt::ItemFlags(Qt::ItemIsEnabled | Qt::ItemIsSelectable );
    }
};

/**
 * @brief The AbstractEnumPropertyEditor class
 */
class AbstractEnumPropertyEditor : public PropertyValueToQVariantConverter{

    bool _editable;
public:
    AbstractEnumPropertyEditor(bool isEditable =true):_editable(isEditable){

    }

    bool setValue(unsigned int ,tlp::PropertyInterface* ,const QVariant& ,int ) const{
        return false;
    }

    bool setAllValue(tlp::PropertyInterface* ,const QVariant& ,int ) const{
        return false;
    }
   virtual Qt::ItemFlags defaultFlags(tlp::PropertyInterface*){
        return _editable?Qt::ItemFlags(Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsEditable) : Qt::ItemFlags(Qt::ItemIsEnabled | Qt::ItemIsSelectable );
    }

    virtual QString label(const unsigned int& value)const=0;
};

/**
 * @brief The AbstractEnumNodePropertyEditor class handles enum types fot nodes.
 */
class AbstractEnumNodePropertyEditor : public AbstractEnumPropertyEditor{
public:
    QVariant value(unsigned int id,tlp::PropertyInterface* property,int role) const;
    QVariant defaultValue(tlp::PropertyInterface* property,int role=Qt::EditRole) const;
};

/**
 * @brief The AbstractEnumNodePropertyEditor class handles enum types fot Edge.
 */
class AbstractEnumEdgePropertyEditor : public AbstractEnumPropertyEditor{
public:
    QVariant value(unsigned int id,tlp::PropertyInterface* property,int role) const;
    QVariant defaultValue(tlp::PropertyInterface* property,int role=Qt::EditRole) const;
};

/**
 * @brief The AbstractGraphIdPropertyConverter class is an abstract class to transform graph id stored in an integer property to the graph name.
 */
class AbstractGraphIdPropertyConverter : public PropertyValueToQVariantConverter{
public:


    bool setValue(unsigned int ,tlp::PropertyInterface* ,const QVariant& ,int ) const{
        return false;
    }

    bool setAllValue(tlp::PropertyInterface* ,const QVariant& ,int ) const{
        return false;
    }
    Qt::ItemFlags defaultFlags(tlp::PropertyInterface*){
        return Qt::ItemFlags(Qt::ItemIsEnabled | Qt::ItemIsSelectable );
    }

    QString idToGraph(tlp::Graph* graph,const int& value)const;

};

/**
 * @brief The NodeGraphIdPropertyConverter class Transform id stored in nodes in graph names
 */
class NodeGraphIdPropertyConverter : public AbstractGraphIdPropertyConverter{
public:

    QVariant value(unsigned int id,tlp::PropertyInterface* property,int role) const;
    QVariant defaultValue(tlp::PropertyInterface* property,int role=Qt::EditRole) const;

};
/**
 * @brief The NodeGraphIdPropertyConverter class Transform id stored in edges in graph names
 */
class EdgeGraphIdPropertyConverter : public AbstractGraphIdPropertyConverter{
public:

    QVariant value(unsigned int id,tlp::PropertyInterface* property,int role) const;
    QVariant defaultValue(tlp::PropertyInterface* property,int role=Qt::EditRole) const;

};

/**
 * @brief The NodeShapePropertyEditor class is an editor for nodes shape.
 */
class NodeShapePropertyEditor : public PropertyValueToQVariantConverter{
public:
    QVariant value(unsigned int id,tlp::PropertyInterface* property,int role=Qt::EditRole) const;
    QVariant defaultValue(tlp::PropertyInterface* property,int role=Qt::EditRole) const;
    bool setValue(unsigned int id,tlp::PropertyInterface* property,const QVariant& data,int role=Qt::EditRole) const;
    bool setAllValue(tlp::PropertyInterface* property,const QVariant& data,int role=Qt::EditRole) const;
};

/**
 * @brief The NodeShapePropertyEditor class is an editor for edges shape.
 */
class EdgeShapePropertyEditor : public PropertyValueToQVariantConverter{
public:
    QVariant value(unsigned int id,tlp::PropertyInterface* property,int role=Qt::EditRole) const;
    QVariant defaultValue(tlp::PropertyInterface* property,int role=Qt::EditRole) const;
    bool setValue(unsigned int id,tlp::PropertyInterface* property,const QVariant& data,int role=Qt::EditRole) const;
    bool setAllValue(tlp::PropertyInterface* property,const QVariant& data,int role=Qt::EditRole) const;
};

/**
 * @brief The PortNodeBooleanPropertyConverter class handle selection for port nodes.
 */
class PortNodeBooleanPropertyConverter : public NodeValueToQVariantConverter<tlp::BooleanProperty,bool>{
    tlp::Graph* _graph;
public:
    PortNodeBooleanPropertyConverter(tlp::Graph* graph);
    bool setValue(unsigned int id,tlp::PropertyInterface* property,const QVariant& data,int role) const;
};

#endif // PROPERTYVALUETOQVARIANTCONVERTER_H
