/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/Trace.h>
#include <portgraph/PortGraphRule.h>
#include <portgraph/PortGraphModel.h>

#include <tulip/Graph.h>
#include <tulip/ForEach.h>
#include <tulip/StringProperty.h>
#include <tulip/IntegerProperty.h>
#include <tulip/ColorProperty.h>
#include <tulip/LayoutProperty.h>
#include <tulip/SizeProperty.h>
#include <tulip/GraphMeasure.h>
#include <tulip/BooleanProperty.h>
#include <tulip/DoubleProperty.h>

#include <iostream>
#include <chrono>

using namespace std;
using namespace tlp;

bool PorgyTlpGraphStructure::hide(Graph *g) {
    bool hide = false;
    g->getAttribute("PORGY, do not display",hide);
    return hide;
}

mt19937 PorgyTlpGraphStructure::gen(chrono::high_resolution_clock::now().time_since_epoch().count());

bool PorgyTlpGraphStructure::isRuleUsed(Graph *rule) {
    //check if rule has been used for rewriting
    Graph *models_root = PorgyTlpGraphStructure::getModelsRoot(rule);
    Graph* g;
    forEach(g, models_root->getDescendantGraphs()) {
        if(rule->getName()==getRuleNameForSubGraph(g))
            return true;
    }
    return false;
}

bool PorgyTlpGraphStructure::isGraphUsed(Graph *g){

    //LHS morphisms must not be modified
    if(g->existLocalProperty(PorgyConstants::TAG))
        return true;
    //check if the graph has been used for rewriting (degree > 0 in the derivation tree)
    Trace t(PorgyTlpGraphStructure::getMainTrace(g));
    node n;
    forEach(n, t.getModelsMetaNodes()) {
        if((t.getNodeMetaInfo(n)->getName()==g->getName())&&(t.deg(n)>0))
            return true;
    }
    return false;
}

tlp::Graph* PorgyTlpGraphStructure::setPorgyStructure(tlp::Graph *root){
    if (root == nullptr)
        root = tlp::newGraph();

    //Put the whole simple graph in G0 (no subgraph/no metadata)
    root->setName("root");
    root->getProperty<IntegerProperty>("viewTgtAnchorShape")->setAllEdgeValue(EdgeExtremityShape::None);
    root->getProperty<IntegerProperty>("viewSrcAnchorShape")->setAllEdgeValue(EdgeExtremityShape::None);
    root->getProperty<DoubleProperty>("viewBorderWidth")->setAllNodeValue(1.0);
    PortNodeBase::getTypeProperty(root);
    BooleanProperty selectRoot(root);
    selectRoot.setAllNodeValue(true);
    selectRoot.setAllEdgeValue(true);

    //graphs
    Graph *models_root = root->addSubGraph(&selectRoot, PorgyConstants::GRAPHS_GRAPH_NAME);
    BooleanProperty* selectModel = models_root->getLocalProperty<BooleanProperty>("viewSelection");
    selectModel->setAllNodeValue(true);
    selectModel->setAllEdgeValue(true);
    Graph *G0 = models_root->addSubGraph(selectModel, PorgyConstants::INITIAL_GRAPH);

    models_root->getProperty<IntegerProperty>("viewShape")->setAllNodeValue(PorgyConstants::CENTER_SHAPE);
    LayoutProperty *layout_root = root->getProperty<LayoutProperty>("viewLayout");
    LayoutProperty *layout = G0->getLocalProperty<LayoutProperty>("viewLayout");    //must be local to each graph
    layout->copy(layout_root); //restore value to the local property
    SizeProperty *size_root = root->getProperty<SizeProperty>("viewSize");
    SizeProperty *size = G0->getLocalProperty<SizeProperty>("viewSize");    //must be local to each graph
    size->copy(size_root); //restore value to the local property
    models_root->setAttribute<unsigned>(PorgyConstants::NUMBER, 0);

    //rules
    Graph *rules_root = root->addSubGraph(PorgyConstants::RULES_GRAPH_NAME);
    rules_root->getLocalProperty<BooleanProperty>("viewSelection");
    rules_root->getProperty<IntegerProperty>("viewShape")->setAllNodeValue(PorgyConstants::CENTER_SHAPE);
    PortGraphRuleDecorator dec(rules_root);
    dec.getSideProperty();

    //derivation trees
    Graph *traceroot = root->addSubGraph(PorgyConstants::TRACE_GRAPH_NAME);
    traceroot->getLocalProperty<BooleanProperty>("viewSelection");
    Graph *tracemain = traceroot->addSubGraph(PorgyConstants::TRACEMAIN);
    root->setAttribute(PorgyConstants::TRACEMAIN, tracemain->getId());
    Trace t(tracemain);
    t.createMetaNode(G0);

    selectModel->setAllNodeValue(false);
    selectModel->setAllEdgeValue(false);

    return root;
}

void PorgyTlpGraphStructure::Graph2PortGraph(Graph *graph, PluginProgress *pluginProgress, bool multiPort, bool mergeInOut) {
    //add ports
    node n;
    IntegerProperty *type = graph->getProperty<IntegerProperty>(PorgyConstants::PORTNODE_TYPE);
    StringProperty *viewLabel = graph->getProperty<StringProperty>("viewLabel");
    forEach(n, graph->getNodes()) {
        type->setNodeValue(n, PorgyConstants::CENTER);
    }

    //if we want the same number of port everywhere
    if (multiPort) {
        //find the node with the higher degree => max number of ports
        DoubleProperty deg_prop(graph);
        string err;
        graph->applyPropertyAlgorithm("Degree", &deg_prop, err, pluginProgress);
        int nb_ports=deg_prop.getNodeDoubleMax();
        edge e;
        node p;
        stableForEach(n, graph->getNodes()) {
            if (type->getNodeValue(n) == PorgyConstants::CENTER) {
                StableIterator<edge> it(graph->getInEdges(n));
               while(it.hasNext()) {
                   e = it.next();
                   p = graph->addNode();
                   type->setNodeValue(p, PorgyConstants::PORT);
                   graph->setTarget(e, p);
                   graph->addEdge(p, n);
               }
               StableIterator<edge> it2(graph->getOutEdges(n));
               while(it2.hasNext()) {
                   e = it2.next();
                   p = graph->addNode();
                   type->setNodeValue(p, PorgyConstants::PORT);
                   graph->setSource(e, p);
                   graph->addEdge(n, p);
               }
               for(int i=deg_prop.getNodeValue(n);i<nb_ports;i++) {
                   p = graph->addNode();
                   type->setNodeValue(p, PorgyConstants::PORT);
                   graph->addEdge(n, p);
               }
            }
        }
    }
    else {
        //if we use only one port to get all the connections
        // mergeInOut is used to specify whether or not specific ports should be created to differenciate in- and outcoming edges
        edge e;
        node p;
        stableForEach(n, graph->getNodes()) {
            if (type->getNodeValue(n) == PorgyConstants::CENTER) {
                p = graph->addNode();
                type->setNodeValue(p, PorgyConstants::PORT);
                StableIterator<edge> it(graph->getInEdges(n));
                while(it.hasNext()) {
                    e = it.next();
                    graph->setTarget(e, p);
                }
                graph->addEdge(p, n);
                if (!mergeInOut) {
                    viewLabel->setNodeValue(p, "In");
                    p = graph->addNode();
                    type->setNodeValue(p, PorgyConstants::PORT);
                }
                StableIterator<edge> it2(graph->getOutEdges(n));
                while(it2.hasNext()) {
                    e = it2.next();
                    graph->setSource(e, p);
                }
                if (!mergeInOut) {
                    graph->addEdge(n, p);
                    viewLabel->setNodeValue(p, "Out");
                }
            }
        }
    }
}

Graph *PorgyTlpGraphStructure::getRootGraph(const Graph* graph) {
    Graph *root=nullptr;
    if(PorgyTlpGraphStructure::isTraceGraph(graph))
        root = getTraceRoot(graph);
    else if(PorgyTlpGraphStructure::isModelGraph(graph))
        root = getModelsRoot(graph);
    else if(PorgyTlpGraphStructure::isRuleGraph(graph))
        root = getRulesRoot(graph);
    assert(root!=nullptr);
    return root;
}

bool PorgyTlpGraphStructure::checkEdgeState(const Graph *g, const edge &rule_e, const edge &graph_e) {
    std::vector<std::string> matchingPropertiesList;
    if (g->getRoot()->attributeExist(PorgyConstants::MATCHING_PROPERTIES)) {
        g->getRoot()->getAttribute<std::vector<std::string> >(PorgyConstants::MATCHING_PROPERTIES, matchingPropertiesList);

        for(unsigned int propertyIndex = 0; propertyIndex < matchingPropertiesList.size(); propertyIndex++) {
            std::string propertyName = matchingPropertiesList.at(propertyIndex);
            tlp::PropertyInterface* currentModelProperty = getModelsRoot(g)->getProperty(propertyName);
            tlp::PropertyInterface* currentRuleProperty = getRulesRoot(g)->getProperty(propertyName);
            tlp::BooleanProperty* testCurrentProperty = getRulesRoot(g)->getProperty<BooleanProperty>(PorgyConstants::MATCHING_PROPERTIES_PREFIX + propertyName + PorgyConstants::MATCHING_PROPERTIES_SUFFIX);
            if(testCurrentProperty->getEdgeValue(rule_e) == true) {
                if(currentRuleProperty->getEdgeStringValue(rule_e) != currentModelProperty->getEdgeStringValue(graph_e)) {
#ifndef NDEBUG
                    cerr << __PRETTY_FUNCTION__ << ": edge matching problem: property " << propertyName << " not compatible" << endl;
#endif
                    return false;
                }
            }
#ifndef NDEBUG
            else {
                cerr << __PRETTY_FUNCTION__ << ": edge : test on property " << propertyName << " not performed" << endl;
            }
#endif
        }
    } //else {
//#ifndef NDEBUG
//        cerr << __PRETTY_FUNCTION__ << ": no known matching properties, using default value." << endl;
//#endif
//        tlp::PropertyInterface *default_p_rule = getRulesRoot(g)->getProperty(PorgyConstants::MATCHING_PROPERTY_DEFAULT);
//        tlp::PropertyInterface *default_p_model = getModelsRoot(g)->getProperty(PorgyConstants::MATCHING_PROPERTY_DEFAULT);
//        if(default_p_rule->getEdgeStringValue(rule_e) != default_p_model->getEdgeStringValue(graph_e))
//            return false;
//    }

    return true;

}

//Remarque : prend en compte les multi-aretes
bool PorgyTlpGraphStructure::checkEdgeState(const Graph *g, const vector<edge> &rule_e, const vector<edge> &modele_e, bool exact) {
    bool b = true, default_p=false;
    set<edge> edge_set;
    if((exact && rule_e.size() != modele_e.size()) || rule_e.size() > modele_e.size() )
        return false;

    // load the matching properties if they exist, otherwise, use MATCHING_PROPERTY_DEFAULT as a default matching value
    std::vector<std::string> matchingPropertiesList;
    if (g->getRoot()->attributeExist(PorgyConstants::MATCHING_PROPERTIES)) {
        g->getRoot()->getAttribute<std::vector<std::string> >(PorgyConstants::MATCHING_PROPERTIES, matchingPropertiesList);
    } else {
//#ifndef NDEBUG
//        cerr << __PRETTY_FUNCTION__ << ": no known matching properties, using default value." << endl;
//#endif
//        matchingPropertiesList.push_back(PorgyConstants::MATCHING_PROPERTY_DEFAULT);
//        default_p = true;
        return true;
    }

    for(unsigned int propertyIndex = 0; propertyIndex < matchingPropertiesList.size(); propertyIndex++) {
        //foreach matching property
        std::string propertyName = matchingPropertiesList.at(propertyIndex);
        tlp::PropertyInterface* currentModelProperty = getModelsRoot(g)->getProperty(propertyName);
        tlp::PropertyInterface* currentRuleProperty = getRulesRoot(g)->getProperty(propertyName);
        tlp::BooleanProperty* testCurrentProperty = getRulesRoot(g)->getProperty<BooleanProperty>(PorgyConstants::MATCHING_PROPERTIES_PREFIX + propertyName + PorgyConstants::MATCHING_PROPERTIES_SUFFIX);

        bool edge_correct = false;

        for(vector<edge>::const_iterator ite_rule = rule_e.begin(); ite_rule != rule_e.end() ; ++ite_rule)
        {
            edge e_rule = *ite_rule;
            edge_correct = false;
            for(vector<edge>::const_iterator ite_modele = modele_e.begin(); ite_modele != modele_e.end(); ++ite_modele)
            {
                edge e_modele = *ite_modele;

                if(testCurrentProperty->getEdgeValue(e_rule) == true || default_p) {
                    if((currentRuleProperty->getEdgeStringValue(e_rule) == currentModelProperty->getEdgeStringValue(e_modele))) {// && (edge_set.find(e_modele) == edge_set.end())) {
                        // if both the values are equal and the edge cannot be find in the set
                        if (edge_set.find(e_modele) == edge_set.end())
                            edge_set.insert(e_modele);
                        edge_correct = true;
                        break;
                    }
#ifndef NDEBUG
                    else {
                        cerr << __PRETTY_FUNCTION__ << ": edge matching problem: property " << propertyName << " not compatible" << endl;
                    }
#endif
                }
                else {
                    edge_set.insert(e_modele);
                    edge_correct = true;
                    break;
  #ifndef NDEBUG
                    cerr << __PRETTY_FUNCTION__ << ": edge matching: test on property " << propertyName << " not performed" << endl;
                    cerr << testCurrentProperty->getEdgeStringValue(e_rule) << " " << currentRuleProperty->getEdgeStringValue(e_rule) << " " << currentModelProperty->getEdgeStringValue(e_rule) << std::endl;
 #endif
                }

            }
            if(!edge_correct)
            {
                b = false;
                break;
            }
        }
        if(!edge_correct)
        {
            b = false;
            break;
        }
    }

    return b;
}


void PorgyTlpGraphStructure::setPPropertyNameAttribute(Graph *g, const std::string &name) {
    g->setAttribute(PorgyConstants::P_NAME, name);
}

string PorgyTlpGraphStructure::getPPropertyNameAttribute(Graph *g) {
    string s;
    g->getAttribute(PorgyConstants::P_NAME, s);
    return s;
}

void PorgyTlpGraphStructure::setBanPropertyNameAttribute(Graph *g, const std::string &name) {
    g->setAttribute(PorgyConstants::BAN_NAME, name);
}

string PorgyTlpGraphStructure::getBanPropertyNameAttribute(Graph *g) {
    string s;
    g->getAttribute(PorgyConstants::BAN_NAME, s);
    return s;
}


//rulename set and get
void PorgyTlpGraphStructure::setRuleNameForSubGraph(tlp::Graph *g, const std::string &rulename) {
    g->setAttribute(PorgyConstants::RULENAME, rulename);
}

std::string PorgyTlpGraphStructure::getRuleNameForSubGraph(tlp::Graph *g) {
    string s;
    g->getAttribute(PorgyConstants::RULENAME, s);
    return s;
}


bool PorgyTlpGraphStructure::isPorgyGraph(const Graph* graph, string &errMsg){
    if(graph->numberOfNodes()==0)
        return true;
    //Check if the top-level graph structure is valid.
    Graph *rules_root=getRulesRoot(graph), *models_root=getModelsRoot(graph), *traces_root=getTraceRoot(graph), *traceMain=getMainTrace(graph);
    if(rules_root == nullptr) {
        errMsg = string("A required graph was not found (")+PorgyConstants::RULES_GRAPH_NAME+string(")");
        return false;
    }
    else if(models_root == nullptr) {
        errMsg = string("A required graph was not found (")+PorgyConstants::GRAPHS_GRAPH_NAME+string(")");
        return false;
    }
    else if(traces_root == nullptr) {
        errMsg = string("A required graph was not found (")+PorgyConstants::TRACE_GRAPH_NAME+string(")");
        return false;
    }
    else if(traceMain == nullptr) {
        errMsg = string("A required graph was not found (")+PorgyConstants::TRACEMAIN+string("). Is the attribute "+PorgyConstants::TRACEMAIN+" defined for the root graph?");
        return false;
    }

    //node types
    IntegerProperty *type = PortNodeBase::getTypeProperty(graph->getRoot());
    if(type==nullptr) {
        errMsg = "Cannot find the " + PorgyConstants::PORTNODE_TYPE + " Property";
        return false;
    }

    //verif que les propriétés sont correctement définies pour les sommets
    node n;
    forEach(n, rules_root->getNodes()) {
        if((type->getNodeValue(n)!=PorgyConstants::CENTER)&&(type->getNodeValue(n)!=PorgyConstants::PORT)) {
            errMsg = "Property "+PorgyConstants::PORTNODE_TYPE+" not defined for every nodes (at least for node " + to_string(n.id) + ") in " + rules_root->getName();
#ifndef NDEBUG
            errMsg += " (value: " + to_string(type->getNodeValue(n))+")";
#endif
            return false;
        }
    }

    PortGraphRuleDecorator dec(rules_root);
    IntegerProperty *side = dec.getSideProperty();
    if(side==nullptr) {
        errMsg = "Cannot find the " + PorgyConstants::RULE_SIDE + " Property";
        return false;
    }

    edge e;
    forEach(e, rules_root->getEdges()) {
        if(side->getEdgeValue(e)==PorgyConstants::NO_SIDE) {
            errMsg = "Property "+PorgyConstants::RULE_SIDE+" not defined for every edges (at least for edge " + to_string(e.id) + ") in " + rules_root->getName();
            return false;
        }
    }
    forEach(n, rules_root->getNodes()) {
        if(side->getNodeValue(n)==PorgyConstants::NO_SIDE) {
            errMsg = "Property "+PorgyConstants::RULE_SIDE+" not defined for every nodes (at least for node " + to_string(n.id) + ") in " + rules_root->getName();
            return false;
        }
    }

    return true;
}

bool PorgyTlpGraphStructure::isTraceGraph(const Graph* g){
    assert(g!=nullptr);
    Graph *traceRoot = getTraceRoot(g);
    return traceRoot == nullptr ? false : traceRoot->isDescendantGraph(g);
}

Graph *PorgyTlpGraphStructure::getMainTrace(const Graph* graph){
    Graph *traceRoot = getTraceRoot(graph);
    if(traceRoot==nullptr)
        return nullptr;
    unsigned index=0;
    graph->getRoot()->getAttribute<unsigned>(PorgyConstants::TRACEMAIN, index);
    assert(index!=0);
    return traceRoot->getDescendantGraph(index);
}

//clone du graphe courant vers le nouveau graphe
Graph *PorgyTlpGraphStructure::clone(Graph *old_graph, string pName, string BanName) {
    Graph *root_models=PorgyTlpGraphStructure::getModelsRoot(old_graph);

    //calcul du nom du nouveau modèle
    unsigned num=0;
    root_models->getAttribute<unsigned>(PorgyConstants::NUMBER, num);     //recherche numéro du dernier modèle créé
    ++num;
    Graph *new_graph = old_graph->addCloneSubGraph(PorgyConstants::GRAPH_PREFIX+to_string(num), true);
    root_models->setAttribute<unsigned>(PorgyConstants::NUMBER, num);

    //recopie de viewLayout et viewSize          
    LayoutProperty *layout_graph = old_graph->getProperty<LayoutProperty>("viewLayout");
    SizeProperty *size_graph = old_graph->getProperty<SizeProperty>("viewSize");
    LayoutProperty *layout_new_graph = new_graph->getLocalProperty<LayoutProperty>("viewLayout");
    SizeProperty *size_new_graph = new_graph->getLocalProperty<SizeProperty>("viewSize");
    layout_new_graph->copy(layout_graph);
    size_new_graph->copy(size_graph);

    BooleanProperty *new_element = new_graph->getLocalProperty<BooleanProperty>(PorgyConstants::NEW);
    new_element->setAllNodeValue(false);
    new_element->setAllEdgeValue(false);

    //je recopie la propriété P courante vers le nouveau modèle si les positions sont utilisées
    if(!pName.empty()) {
        BooleanProperty *oldP = old_graph->getProperty<BooleanProperty>(pName);
        BooleanProperty *newP = new_graph->getLocalProperty<BooleanProperty>(pName);
        newP->copy(oldP);
    }
    if(!BanName.empty()) {
        BooleanProperty *oldBan = old_graph->getProperty<BooleanProperty>(BanName);
        BooleanProperty *newBan = new_graph->getLocalProperty<BooleanProperty>(BanName);
        newBan->copy(oldBan);
    }

    return new_graph;
}

bool PorgyTlpGraphStructure::graphIsASimplePath(Graph* graph){
        map<unsigned int,node> depthMap;
        MutableContainer<unsigned int> level;
        //Compute the level for each node.
        dagLevel(graph,level);
        node n;
        forEach(n,graph->getNodes()){
                unsigned int currentDepth = level.get(n);
                map<unsigned int,node>::iterator it = depthMap.find(currentDepth);
                //An edge with the same level already exits
                if(it != depthMap.end()){
                    return false;
                }
                depthMap[currentDepth]=n;
        }
        if(depthMap.empty()){
            return false;
        }
        //Check if there is no gap between two nodes.
        for(map<unsigned int ,node >::iterator it = depthMap.begin(); it != depthMap.end();){
            node from = it->second;
            node to = (++it)->second;

            if(it != depthMap.end() ){

                //If there is no edge between two nodes a gap exists so the selection is invalid.
                if(!graph->hasEdge(from,to,true)){
                    return false;
                }
                Iterator<edge>* edges = graph->getOutEdges(from);
                unsigned int count = 0;
                bool moreThanOneEdge = false;
                //Check if there is more than one edge between the source and the destination node
                while(edges->hasNext()){
                    edge e = edges->next();
                    if(graph->target(e) == to){
                        ++count;
                        if(count > 1){
                            //More than one edge invalid structure.
                            moreThanOneEdge = true;
                            break;
                        }
                    }
                }
                delete edges;
                if(moreThanOneEdge){
                    return false;
                }
            }
        }
        return true;
}

void PorgyTlpGraphStructure::getMatchingPropertyNames(tlp::Graph *g, std::unordered_set<std::string> &propertiesNames) {
    //Add the matching properties to the display list
    std::vector<std::string> matchingPropertiesList;
    if (g->getRoot()->attributeExist(PorgyConstants::MATCHING_PROPERTIES)) {
        g->getRoot()->getAttribute<std::vector<std::string> >(PorgyConstants::MATCHING_PROPERTIES, matchingPropertiesList);
    }

    for(const std::string& st: matchingPropertiesList) {
        propertiesNames.insert(st);
    }
}

void PorgyTlpGraphStructure::getMatchingPropertyNamesExtended(tlp::Graph *g, std::unordered_set<std::string> &propertiesNames) {
    std::unordered_set<std::string> matchingPropertiesList;
    getMatchingPropertyNames(g, matchingPropertiesList);

    //Extend the normal output with the "*_state" values

    for(const std::string& st: matchingPropertiesList) {
        propertiesNames.insert(st);
        propertiesNames.insert(PorgyConstants::MATCHING_PROPERTIES_PREFIX + st + PorgyConstants::MATCHING_PROPERTIES_SUFFIX);
    }
}

void PorgyTlpGraphStructure::setMatchingPropertyNames(tlp::Graph *g, std::unordered_set<std::string> &propertiesNames) {
  vector<string> vect(propertiesNames.begin(), propertiesNames.end());
  g->getRoot()->setAttribute<std::vector<std::string> >(PorgyConstants::MATCHING_PROPERTIES, vect);
}

void PorgyTlpGraphStructure::getStandardPropertyNames(tlp::Graph *g, std::vector<string> &propertiesNames) {
    std::string tmp;
    propertiesNames.clear();
    forEach (tmp, g->getRoot()->getProperties()) {
        propertiesNames.push_back(tmp);
    }
    forEach (tmp, g->getRoot()->getProperties()) {
        std::vector<std::string>::iterator itP = std::find(propertiesNames.begin(), propertiesNames.end(), PorgyConstants::MATCHING_PROPERTIES_PREFIX + tmp + PorgyConstants::MATCHING_PROPERTIES_SUFFIX);
        if (itP != propertiesNames.end()) {
            propertiesNames.erase(itP);
        }
    }
}
