/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <tulip/ForEach.h>
#include <tulip/StringProperty.h>
#include <tulip/BooleanProperty.h>
#include <tulip/LayoutProperty.h>
#include <tulip/IntegerProperty.h>

#include <portgraph/PortNode.h>
#include <portgraph/PortGraph.h>
#include <portgraph/Port.h>
#include <portgraph/Bridge.h>

using namespace std;
using namespace tlp;

PortGraph::PortGraph (const set<node> v, Graph *g):_graph(g){

    //récupération des sommets centraux
    for(auto& n : v) {
        if(PortNodeBase::isCenter(n,_graph))
            portnode.push_back(new PortNode(*this, n));
    }

    node n;
    Iterator<PortNode *> *it = getPortNodes();
    while(it->hasNext()) {
        PortNode *pn = it->next();
        Iterator<Port::PortPtr> *it2 = pn->getPorts();
        while(it2->hasNext()) {
            Port::PortPtr p=it2->next();
            forEach(n, _graph->getInOutNodes(p->getNode())) {
                if(PortBase::isPort(n, _graph))
                    p->connectPortTo(getPortContainer(n));
            }
        }delete it2;
    }delete it;
}

PortGraph::PortGraph (Graph *g):_graph(g){
    node n;
    //récupération des sommets centraux
    forEach(n, g->getNodes()){
        if(PortNodeBase::isCenter(n,_graph))
            portnode.push_back(new PortNode(*this, n));
    }
    //tous les portnodes sont créés. je cherche les ports connectés
    Iterator<PortNode *> *it = getPortNodes();
    while(it->hasNext()) {
        PortNode *pn=it->next();
        Iterator<Port::PortPtr> *it2 = pn->getPorts();
        while(it2->hasNext()) {
            Port::PortPtr p = it2->next();
            forEach(n, _graph->getInOutNodes(p->getNode())) {
                if(PortBase::isPort(n, _graph))
                    p->connectPortTo(getPortContainer(n));
            }
        }delete it2;
    }delete it;
}

PortGraph::~PortGraph() {
    PortNode *pn;
    stableForEach(pn, getPortNodes()) {
        delete pn;
    }
}

PortNode* PortGraph::operator[](const tlp::node &n) const {
    if(Bridge::isBridge(n, _graph))
        return nullptr;
    else {
        assert(container.find(n) != container.end());
        return container.at(n);
    }
}

PortNode*& PortGraph::operator[](const tlp::node &n) {
    return container[n];
}

PortNode* PortGraph::getOnePortNode() const {
    return *(portnode.begin());
}

tlp::Iterator<PortNode *> *PortGraph::getInOutPortNodes(PortNode* pn, tlp::BooleanProperty *ban) const {
    dequeportnode p;
    Port::PortPtr pt, pt2;
    if(ban==nullptr) {
        forEach(pt, pn->getPorts()) {
            forEach(pt2, pt->getConnectedPorts()) {
                p.push_back((*this)[pt2->getNode()]);
            }
        }
    }
    else {
        forEach(pt, pn->getPorts()) {
            forEach(pt2, pt->getConnectedPorts()) {
                PortNode* pn2 = (*this)[pt2->getNode()];
                if(!ban->getNodeValue(pn2->getCenter()))
                    p.push_back(pn2);
            }
        }
    }
    return new CopyStlIterator<PortNode*, dequeportnode>(p);
}

Iterator<PortNode *> *PortGraph::getPortNodes(BooleanProperty *ban, bool random_order) const {     //renvoie un itérateur sur l'ensemble des portnodes
    dequeportnode p;
    if(ban==nullptr)
        p=portnode;
    else {
        for(auto& pr : portnode) {
            if(!ban->getNodeValue(pr->getCenter()))
                p.push_back(pr);
        }
    }
    if(random_order) {
        shuffle(p.begin(), p.end(), PorgyTlpGraphStructure::gen);
    }
    return new CopyStlIterator<PortNode *, dequeportnode>(p);
}

bool PortGraphDecorator::isCenter(Graph *g, const node n) {
    assert(g->isElement(n));
    return (g->getProperty<IntegerProperty>(PorgyConstants::PORTNODE_TYPE)->getNodeValue(n)==PorgyConstants::CENTER);
}

//static method
bool PortGraph::isPortGraphEdge(const edge e, Graph *g) {
    const pair<node, node> &ends = g->ends(e);
    return (PortBase::isPort(ends.first, g)&&PortBase::isPort(ends.second, g)&&!Bridge::isBridge(ends.second, g)&&!Bridge::isBridge(ends.first, g));
}

void PortGraph::addPortContainer(const node &n, Port::PortPtr p) {
    container_port[n] = p;
}

const Port::PortPtr PortGraph::getPortContainer(const node &n) const {
    std::map<tlp::node, Port::PortPtr>::const_iterator it = container_port.find(n);
    assert(it!=container_port.end());
    return it->second;
}

const pair<PortNode*, PortNode*> PortGraph::ends(const tlp::edge e) const {
    const pair<node, node> &ends = _graph->ends(e);
    return make_pair((*this)[ends.first], (*this)[ends.second]);
}

//PortNode* PortGraph::getPortNode(const tlp::node n) const {
//    return Bridge::isBridge(n, _graph)?nullptr:(*this)[n];
//}

void PortGraph::drawSameSizes(SizeProperty *NodeSize, LayoutProperty *layout) const {
    map <string, pair<unsigned, unsigned> > optimalSizes;
    pair<unsigned, unsigned> pair1;
    pair<unsigned, unsigned> pair2;
    PortNode *pn;
    optimalSizes[""] = make_pair(1,1);
    forEach(pn, getPortNodes()) {
        if(Bridge::isBridge(pn->getCenter(), _graph))
            continue;
        if (!pn->getName().empty()) {
            pair1 = pn->getOptimalSize(layout);
            pair2 = optimalSizes[pn->getName()];

            pair2.first  = max(pair1.first,  pair2.first);
            pair2.second = max(pair1.second, pair2.second);

            optimalSizes[pn->getName()] = pair2;
        }
    }
    forEach(pn, getPortNodes()) {
        if(Bridge::isBridge(pn->getCenter(), _graph)) {
            continue;
        }
        if (optimalSizes.end() != optimalSizes.find(pn->getName())){ // found
            pn->setPositions(optimalSizes[pn->getName()].first, optimalSizes[pn->getName()].second, layout);
            pn->resizeCenter(optimalSizes[pn->getName()].first, optimalSizes[pn->getName()].second, NodeSize);
        }
        else {
            pn->setPositions(layout);
            pn->resizeCenter(NodeSize, layout);
        }
    }
}

PortGraphDecorator::PortGraphDecorator(Graph *g):GraphDecorator(g){

}

/**
 * @brief PortGraphModel::numberOfNodes
*/
unsigned PortGraphDecorator::numberOfNodes() const {
    unsigned nb=0, nb_tulip=0;
    if(graph_component->attributeExist("number_of_nodes")) {
        graph_component->getAttribute<unsigned>("number_of_nodes",nb);
        graph_component->getAttribute<unsigned>("number_of_nodes_tulip",nb_tulip);
    }

    if(nb_tulip!=graph_component->numberOfNodes()) {
        node n;
        forEach(n, graph_component->getNodes()) {
            if(PortGraphDecorator::isCenter(graph_component,n))
                nb++;
        }
        graph_component->setAttribute<unsigned>("number_of_nodes", nb);
        graph_component->setAttribute<unsigned>("number_of_nodes_tulip", graph_component->numberOfNodes());
    }
    return nb;
}
