/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <portgraph/PortGraphModel.h>
#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/PortNode.h>
#include <portgraph/PortGraphRule.h>
#include <portgraph/Port.h>

#include <tulip/ForEach.h>
#include <tulip/DoubleProperty.h>
#include <tulip/BooleanProperty.h>
#include <tulip/IntegerProperty.h>
#include <tulip/LayoutProperty.h>
#include <tulip/ConnectedTest.h>
#include <tulip/Algorithm.h>

using namespace tlp;
using namespace std;

//////////////
PortGraphModel::PortGraphModel (tlp::Graph *g):PortGraph(g){}

//////////////
PortGraphModel::PortGraphModel (const std::set<tlp::node> v, tlp::Graph *g):PortGraph(v,g){}


//////////////
tlp::BooleanProperty *PortGraphModelDecorator::getSelectionProperty() {
    return PorgyTlpGraphStructure::getModelsRoot(graph_component)->getLocalProperty<BooleanProperty>("viewSelection");
}

bool PortGraphModelDecorator::redraw(const std::string& plugin, tlp::DataSet& parameters ,std::string &errMsg,tlp::PluginProgress* progress){
    bool res=false;
    std::list<std::string> plugins=layoutPluginsList();
    if(std::find(plugins.begin(),plugins.end(),plugin) == plugins.end()){
        errMsg = "The choosen plugin is not in the list";
    }else{
        if(numberOfNodes()>1) {
            Observable::holdObservers();
            res = graph_component->applyPropertyAlgorithm(plugin, graph_component->getProperty<LayoutProperty>("viewLayout"), errMsg,progress,&parameters);
            Observable::unholdObservers();
        }
    }
    return res;
}


//////////////
void PortGraphModel::SortPortNodebyPosition(const BooleanProperty *prop) {
    dequeportnode v;
    PortNode *pn;
    forEach(pn, getPortNodes()) {
        if(prop->getNodeValue(pn->getCenter())==true)
            v.push_front(pn);
        else
            v.push_back(pn);
    }
    setPortNode(v);
}

//////////////

/**
 * @brief The PartialMap class
 * Used to find instances of the LHS of a rule
 * @param current : currently mapped vertices. At the beginning, no vertex is mapped. current is extended progressively
 * @param M : Matrix given in adgency list.
 * Gives the possible mappings for each node of the LHS of the rule
 */

class PartialMap {

    typedef set<node> possiblemap;
    typedef map<node,node> nodemap;
    typedef pair<const node, node> pair_node;
    typedef map<node,possiblemap> partialmap;

    //matrix containing the possibilities for the mapping
    partialmap M ;

    //number of nodes already mapped
    unsigned nb_mapped ;

    //current mapping of the chosen nodes
    nodemap current ;

    PortGraph* left ;
    PortGraph* modele ;

    //PluginProgress for feedback on the avancement
    PluginProgress *pp ;
    int progress ;
    int progress_step ;
    int max_progress ;

public:

    PartialMap(PortGraph* left, PortGraph* modele, PluginProgress * pp)
        : nb_mapped(0), left(left), modele(modele), pp(pp), progress(0), progress_step(INT_MAX), max_progress(INT_MAX) {}

    explicit PartialMap(const PartialMap &partmap):M(partmap.M), nb_mapped(partmap.nb_mapped), current(partmap.current),
        left(partmap.left), modele(partmap.modele), pp(partmap.pp),progress(partmap.progress), progress_step(partmap.progress_step), max_progress(partmap.max_progress) {}

    bool initialisation(bool exact, BooleanProperty *banp);
#ifndef NDEBUG
    void display();
#endif
    bool refine(bool exact);
    void backtrack(int numtofind, int &nbr_found, bool exact, BooleanProperty *pos, BooleanProperty *banp);

private:
    node chooseNode() const;
    bool verifyPos(BooleanProperty *pos) const;
    bool verifyIso(bool exact) const;
    void addSub(int num_solutions, BooleanProperty *position, BooleanProperty *banp) const;
} ;

bool PartialMap::initialisation(bool exact, BooleanProperty *banp){

#ifndef NDEBUG
    cerr << __PRETTY_FUNCTION__ << " : Initialisation begin" << endl;
#endif

    node n;
    forEach(n, left->getGraph()->getNodes()) {
        M[n] = possiblemap();
#ifndef NDEBUG
        cerr << __PRETTY_FUNCTION__ <<" LHS rule node :" << n << endl;
#endif
    }

#ifndef NDEBUG
    cerr <<__PRETTY_FUNCTION__  <<" : LHS number of nodes :" << M.size() << endl ;
#endif


    //Begin with adding the portNodes possible images
    PortNode *rule_pn;
    forEach(rule_pn, left->getPortNodes()) {
#ifndef NDEBUG
        cerr << __PRETTY_FUNCTION__ << " : looking for candidates for PortNode :" << rule_pn->getCenter() << " "<< rule_pn->getName() << endl ;
#endif
        Iterator<PortNode *> *it2 = modele->getPortNodes(banp, true) ;
        while(it2->hasNext()) {
            PortNode *modele_pn = it2->next() ;
            //only add compatible PortNodes (ie : same name, same state, compatible degrees)
            if(PortNode::check(rule_pn,modele_pn,exact)) {
#ifndef NDEBUG
                cerr <<__PRETTY_FUNCTION__ << " : Found candidate :" << modele_pn->getCenter() << " " << modele_pn->getName()<< endl ;
#endif
                M.at(rule_pn->getCenter()).insert(modele_pn->getCenter()) ;
            }
//#ifndef NDEBUG
//            else {
//                cerr <<__PRETTY_FUNCTION__ << " : Refused candidate :" << modele_pn->getCenter() << " " << modele_pn->getName() << " of state " << modele_pn->getState() << endl ;
//            }
//#endif
        }
        delete it2 ;
    }

#ifndef NDEBUG
    cerr << "PortNodes finished" << endl;
    cerr << "Ports begin : " << endl;
#endif

    //Then add the possbilities for the ports
    forEach(rule_pn, left->getPortNodes()) {
        Iterator<Port::PortPtr> *it_port = rule_pn->getPorts();
        while(it_port->hasNext()) {
            Port::PortPtr rule_port = it_port->next() ;
#ifndef NDEBUG
            if(!exact)
                cerr << __PRETTY_FUNCTION__ << "Port considered : " << rule_port->getNode() << " " << rule_port->getName() << " of arity " << rule_port->getArity() << endl;
            else
                cerr << __PRETTY_FUNCTION__ << "Port considered : " << rule_port->getNode() << " " << rule_port->getName() << endl;
#endif
            for(auto n: M.at(rule_pn->getCenter())) {
                Iterator<node> * it_vois = modele->getGraph()->getInOutNodes(n) ;
                while(it_vois->hasNext()) {
                    node p2 = it_vois->next() ;
                    Port::PortPtr modele_port = modele->getPortContainer(p2) ;
                    if(Port::check(rule_port,modele_port, exact)) {
#ifndef NDEBUG
                        cerr << __PRETTY_FUNCTION__ << "Candidate mapping : " << p2 << endl;
#endif
                        M.at(rule_port->getNode()).insert(p2) ;
                    }
#ifndef NDEBUG
                    else {
                        cerr << __PRETTY_FUNCTION__ << "Candidate refused : " << p2 << " " << modele_port->getName() << " of degree :" << modele_port->deg() << endl;
                    }
#endif
                }
                delete it_vois ;
            }
        }
        delete it_port ;
    }

    bool b = refine(exact) ;
#ifndef NDEBUG
    cerr << __PRETTY_FUNCTION__ << " : Initialisation end" << endl;
#endif
    //End with a call to refine
    return b ;
}

#ifndef NDEBUG
//Debug function to display the content of current and of M
void PartialMap::display(){
    cerr << "Display of the current partialmap" << endl;
    cerr << "Number of nodes mapped for now :" << nb_mapped << endl;
    cerr << "Current mapping : " << endl;
    for(const auto& p : current) {
        cerr << "node " << p.first <<" to node " << p.second << endl ;
    }

    cerr << "Possibilities for other nodes :" << endl;
    for(const auto &p: M) {
        if(current.find(p.first) == current.end()) {
            cerr << "Candidates for " << p.first << " are : ";
            for(auto& n: p.second) {
                cerr << n << " ; " ;
            }
            cerr << endl ;
        }
    }
}
#endif

node PartialMap::chooseNode() const{
    unsigned int k = UINT_MAX ;
    node m, n;
    forEach(m, left->getGraph()->getNodes()) {
        assert(M.find(m) != M.end());
        if(current.find(m) == current.end() && M.at(m).size() < k ) {
            k = M.at(m).size() ;
            n = m ;
        }
    }
#ifndef NDEBUG
    if(!n.isValid()) {
        cerr << __PRETTY_FUNCTION__ << " ERROR: All vertices already mapped" << endl;
        assert(false);
    }
#endif

    return n;
}

bool PartialMap::verifyPos(BooleanProperty *pos) const {
    if(pos == nullptr)
        return true;

    for(const auto& pnode: current) {
        if(pos->getNodeValue(pnode.second))
            return true ;
    }
    return false ;
}

bool PartialMap::verifyIso(bool exact) const {
    if(!exact) {
        //We verify that all edges in the rule have an image in the model
        edge e;
        Graph *g = left->getGraph();
        forEach(e, g->getEdges()) {
            const pair<node, node> &ends = g->ends(e);
            vector<edge> e_rule = g->getEdges(ends.first,ends.second,false) ;
            vector<edge> e_modele = modele->getGraph()->getEdges(current.at(ends.first), current.at(ends.second),false) ;
            if(!PorgyTlpGraphStructure::checkEdgeState(modele->getGraph(), e_rule, e_modele, exact)) {
#ifndef NDEBUG
                cerr << __PRETTY_FUNCTION__ << " : The map given isn't an isomorphisme" << endl;
#endif
                return false ;
            }
        }
    }
    else {
        node u;
        forEach(u, left->getGraph()->getNodes()) {
            Iterator<node> *it2 = left->getGraph()->getNodes();
            while(it2->hasNext()) {
                node v = it2->next();
                vector<edge> e_rule = left->getGraph()->getEdges(u,v, false) ;
                vector<edge> e_modele = modele->getGraph()->getEdges(current.at(u), current.at(v), false) ;
                if(!PorgyTlpGraphStructure::checkEdgeState(modele->getGraph(), e_rule, e_modele, exact)) {
#ifndef NDEBUG
                    cerr << __PRETTY_FUNCTION__ << " : The map given isn't an isomorphism" << endl;
#endif
                    return false ;
                }
            }
            delete it2;
        }

    }

#ifndef NDEBUG
    cerr << __PRETTY_FUNCTION__ << " : Isomorphism verified : ok" << endl;
#endif
    return true ;
}

void PartialMap::addSub(int num_solutions, BooleanProperty *position, BooleanProperty *banp) const {

    string rulename = left->getGraph()->getSuperGraph()->getName();
    Graph *sub = modele->getGraph()->addSubGraph("sg_"+rulename+ "_"+to_string(num_solutions)+"-"+((position!=nullptr)?position->getName():PorgyConstants::NO_P));
    PorgyTlpGraphStructure::setPPropertyNameAttribute(sub, ((position!=nullptr)?position->getName():PorgyConstants::NO_P));
    PorgyTlpGraphStructure::setBanPropertyNameAttribute(sub, ((banp!=nullptr)?banp->getName():PorgyConstants::NO_BAN));
    PorgyTlpGraphStructure::setRuleNameForSubGraph(sub, rulename);
    IntegerProperty *tag = sub->getLocalProperty<IntegerProperty>(PorgyConstants::TAG);
    tag->setAllNodeValue(-1);
    tag->setAllEdgeValue(-1);

#ifndef NDEBUG
    cerr << __PRETTY_FUNCTION__ << ": New subgraph found: " << sub->getName() << endl;
#endif

    node n;
    forEach(n, left->getGraph()->getNodes()) {
        assert(!sub->isElement(current.at(n)));
        sub->addNode(current.at(n));
        assert(tag->getNodeValue(current.at(n)) == tag->getNodeDefaultValue());
        tag->setNodeValue(current.at(n),n);
    }

    edge e_rule;
    forEach(e_rule, left->getGraph()->getEdges()) {
        const pair<node, node> &ends = left->getGraph()->ends(e_rule);
        for(auto& e: modele->getGraph()->getEdges(current.at(ends.first),current.at(ends.second),false)) {
            if(tag->getEdgeValue(e) == tag->getEdgeDefaultValue()) {
                assert(e.isValid());
                sub->addEdge(e);
                tag->setEdgeValue(e, e_rule);
                break;
            }
        }
    }
}

bool PartialMap::refine(bool exact) {

#ifndef NDEBUG
    display();
    cerr << __PRETTY_FUNCTION__ << ": Begin refine" << endl;
#endif

    bool modified = false ;

    for(auto &p: M) {
        vector<set<node>::iterator> to_delete;
        for(set<node>::iterator model_node_iter = p.second.begin() ; model_node_iter != p.second.end() ; ++model_node_iter) {
            bool found = false ;
            Iterator<node> *itervoisin = left->getGraph()->getInOutNodes(p.first) ;
            while(itervoisin->hasNext())
            {
                bool voisin_ok = false ;
                node rule_vois = itervoisin->next() ;

                for(auto& model_vois: M.at(rule_vois)) {
                    vector<edge> edges_rule = left->getGraph()->getEdges(p.first,rule_vois,false) ;
                    vector<edge> edges_modele = modele->getGraph()->getEdges(*model_node_iter, model_vois, false) ;
                    if(PorgyTlpGraphStructure::checkEdgeState(modele->getGraph(), edges_rule, edges_modele, exact)){
                        //If there is an edge, with the same state as in the rule
                        voisin_ok =true ;
                        break ;
                    }
                }
                if(!voisin_ok)
                {
                    //There is a neighbour of the node in the rule which can't be mapped to a neighbourg of the image in the model
                    found = true ;
                    break ;
                }
            }
            delete itervoisin ;
            if(found) {
#ifndef NDEBUG
                cerr << __PRETTY_FUNCTION__ << " : Found that node " << p.first << " could not be mapped to node " << *model_node_iter << endl;
#endif
                modified = true ;
                //We suppress the node from the possibilities
                to_delete.push_back(model_node_iter);
            }
        }
        for(const auto &i: to_delete) {
            p.second.erase(i);
        }
    }

   //We verify that all nodes of the rule have at least one possible mapping
    for(auto &p: M) {
        if(p.second.empty())
        {
#ifndef NDEBUG
            cerr << __PRETTY_FUNCTION__ << " : Found no mapping from here" << endl;
#endif
            return false ;    //One node has no possible image, we stop
        }
    }

    if(exact) {
        node n;
        forEach(n, modele->getGraph()->getNodes()) {
            bool ok = false ;
            for(partialmap::iterator iter = M.begin() ; iter != M.end() ; ++iter) {
                if(iter->second.find(n) != iter->second.end()) {
                    ok = true;
                    break;
                }
            }
            if(!ok)
                return false ;
        }
    }

    //The operation is repeated untill fixpoint is reached
    return modified?refine(exact):true;
}

void PartialMap::backtrack(int numtofind, int & nbr_found, bool exact, BooleanProperty *pos, BooleanProperty *banp) {

    node v = chooseNode();
    const possiblemap &set_v = M.at(v);
    int nb_possibilities =  set_v.size() ;

    if(nb_possibilities == 0)
        return ;

#ifndef NDEBUG
    display() ;
    cerr << __PRETTY_FUNCTION__ << " : Looking for images of node " << v << endl;
    cerr << __PRETTY_FUNCTION__ << " : Number of possibilities : " << set_v.size() << endl;
#endif

    progress_step /= nb_possibilities;

    //When only one node remaining
    if(nb_mapped == left->getGraph()->numberOfNodes()-1) {
#ifndef NDEBUG
        cerr << __PRETTY_FUNCTION__ << " : Last node to be mapped" << endl;
#endif

        nb_mapped ++ ;
        for(auto& w: set_v) {
            current[v] = w ;
#ifndef NDEBUG
            cerr << __PRETTY_FUNCTION__ << " : Trying to send node " << v << " to node " << w << endl ;
            cerr << __PRETTY_FUNCTION__ << " : All vertices mapped, checking the isomorphism" << endl;
#endif
            //All nodes have been mapped. Checking if we truly have found an isomorphism

            if(verifyPos(pos) && verifyIso(exact)) { //only uses the content of current (M is false)
                addSub(nbr_found,pos,banp);
                ++nbr_found;
                if(nbr_found == numtofind) {
                    return;
                }
            }
            progress += progress_step ;
            pp->progress(progress, max_progress) ;
        }

        progress_step *= nb_possibilities ;
        current.erase(v) ;
        --nb_mapped;
        return;
    }

    //General case
    for(auto& w: set_v) {
//#ifndef NDEBUG
//        cerr << __PRETTY_FUNCTION__ << " : Trying to send node " << v << " to node " << w << endl;
//#endif

        PartialMap *partmap = new PartialMap(*this) ;
        partmap->current[v] = w ;
        partmap->nb_mapped += 1;

        partmap->M.at(v).clear();
        partmap->M.at(v).insert(w);

        //We suppress to other node the possibility to map to w
        for(auto& p: partmap->M) {
            if(p.first!=v) {
                set<node>::iterator iter=p.second.find(w);
                if(iter!=p.second.end()) {
                    p.second.erase(iter);
                }
            }
        }

        //Case lots of vertices few edges
        node n_vois;
        forEach(n_vois, left->getGraph()->getInOutNodes(v)) {
            if(partmap->current.find(n_vois) == partmap->current.end() && partmap->M.at(n_vois).size() > modele->getGraph()->deg(w)) {
//#ifndef NDEBUG
//                cerr << __PRETTY_FUNCTION__ << " : Restraining images of " << n_vois << " to neighbors of node " << w << endl;
//#endif
                set<node> node_set(partmap->M.at(n_vois).begin(), partmap->M.at(n_vois).end());
                partmap->M.at(n_vois).clear() ;
                Iterator<node> *iter_modele = modele->getGraph()->getInOutNodes(w) ;
                while(iter_modele->hasNext()) {
                    node node_modele = iter_modele->next() ;
//#ifndef NDEBUG
//                    cerr << __PRETTY_FUNCTION__ << " : Node considered : " << node_modele << endl;
//#endif
                    if(node_set.find(node_modele) != node_set.end()) {
                        partmap->M.at(n_vois).insert(node_modele);
//#ifndef NDEBUG
//                        cerr << __PRETTY_FUNCTION__ << " : Image Found : " << node_modele << endl;
//#endif
                    }
                }
                delete iter_modele ;
            }
        }

        if (partmap->refine(exact)) {
//#ifndef NDEBUG
//            cerr << __PRETTY_FUNCTION__ << ": Choice of mapping node " << v << " to node" << w << " may lead to solutions" << endl;
//#endif
            partmap->backtrack(numtofind,nbr_found,exact,pos,banp);
            if(nbr_found >= numtofind && numtofind != -1)
                return;
        }
//#ifndef NDEBUG
//        else {
//            cerr << __PRETTY_FUNCTION__ << ": Choice of mapping node " << v << " to node " << w << " leads to no solutions" << endl;
//        }
//#endif
        delete partmap ;
        progress += progress_step ;
        pp->progress(progress, max_progress) ;
    }

//#ifndef NDEBUG
//    cerr << __PRETTY_FUNCTION__ << " : All possibilites tested in this branch " << endl;
//#endif
    progress_step *= nb_possibilities ;
}

bool PortGraphModel::find(tlp::Graph* left, tlp::DataSet *dataset, bool exact, PluginProgress * pp){

//#ifndef NDEBUG
//    cerr << __PRETTY_FUNCTION__ << ": Looking for subgraph of the rule " << left->getSuperGraph()->getName() <<" in the graph :" << getGraph()->getName() << endl;
//#endif

    int numtofind = -1 ;
    BooleanProperty *banp = nullptr;
    BooleanProperty *pos = nullptr;
    int nb_found = 0;

    if(nullptr != dataset) {
        dataset->get(PorgyConstants::maximum_number_of_instances, numtofind); //generate only the required LHS instances
        dataset->get(PorgyConstants::BAN, banp);
        dataset->get(PorgyConstants::POSITION, pos);
    }

    pp->setComment("Trying to instantiate the LHS of "+left->getSuperGraph()->getName() +" on "+_graph->getName());

    //on initialise notre mapping
    PortGraph *pg_left = new PortGraphModel(left);
    PartialMap partmap(pg_left, this, pp) ;
    if(! partmap.initialisation(exact, banp))
    {
#ifndef NDEBUG
        cerr << __PRETTY_FUNCTION__ << ": No solutions after initialisation" << endl;
#endif
        return false;
    }

    partmap.backtrack(numtofind, nb_found, exact, pos,banp);
    delete pg_left;
    if(nullptr != dataset)
        dataset->set("number of instances",nb_found);
#ifndef NDEBUG
    cerr << __PRETTY_FUNCTION__ << ": Search over,  " << nb_found << " solutions found" << endl;
#endif

    return (nb_found != 0);
}

//////////////
static void build_edges_vector(Graph *g, vector<edge> &v, BooleanProperty *ban=nullptr) {
    edge e;
    forEach(e, g->getEdges()) {
        if(PortGraph::isPortGraphEdge(e, g)) {
            //check if source and target are both in the object
            //cannot do that in an assert because in some cases the portgraph is not built upon the whole graph returned by _graph
            const pair<node, node> &ends = g->ends(e);
            if(g->isElement(ends.first)&&g->isElement(ends.second))
                if((ban==nullptr)||(!ban->getEdgeValue(e)))
                    v.push_back(e);
        }
    }
}

//////////////
Iterator<tlp::edge> *PortGraphModel::getEdges(BooleanProperty *ban) {
    vector<edge> v;
    build_edges_vector(_graph, v, ban);
    return new CopyStlIterator<edge, vector<edge> >(v);
}

PortGraphModelDecorator::PortGraphModelDecorator(Graph *g):PortGraphDecorator(g){

}

/**
 * @brief PortGraphModel::numberOfEdges
*/
unsigned PortGraphModelDecorator::numberOfEdges() const {
    unsigned nb=0, nb_tulip=0;
    if(graph_component->attributeExist("number_of_edges")) {
        graph_component->getAttribute<unsigned>("number_of_edges",nb);
        graph_component->getAttribute<unsigned>("number_of_edges_tulip",nb_tulip);
    }

    if(nb_tulip!=graph_component->numberOfEdges()) {
        vector<edge> v;
        build_edges_vector(graph_component, v);
        nb = v.size();
        graph_component->setAttribute<unsigned>("number_of_edges", nb);
        graph_component->setAttribute<unsigned>("number_of_edges_tulip", graph_component->numberOfEdges());
    }
    return nb;
}

///////////////
list<std::string> PortGraphModelDecorator::layoutPluginsList()const{
    std::list<std::string> list;
    list.push_back(PorgyConstants::REDRAW_GRAPH);
    return list;
}
