/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <tulip/StringProperty.h>
#include <tulip/BooleanProperty.h>
#include <tulip/LayoutProperty.h>
#include <tulip/ForEach.h>

#include <portgraph/PortGraphRule.h>
#include <portgraph/PortNode.h>
#include <portgraph/Bridge.h>
#include <portgraph/PorgyTlpGraphStructure.h>

using namespace std;
using namespace tlp;

PortGraphRule::PortGraphRule (Graph *g):PortGraph(g),bridge(nullptr){
//#ifndef NDEBUG
//    string errMsg;
//    if(!PorgyTlpGraphStructure::isPorgyGraph(g, errMsg)) {
//        cerr << "Fatal Error: " << errMsg << endl;
//        assert(false);
//    }
//#endif
    assert(PorgyTlpGraphStructure::isRuleGraph(g));

    node n;    
        forEach(n, g->getNodes()) {            
            if(PortNodeBase::isCenter(n,getGraph())&& Bridge::isBridge(n, getGraph())) {
                bridge = new Bridge(*this, n);
                break;
            }
        }
        //Be sure to find the bridge
        assert(bridge != nullptr);
}

bool PortGraphRuleDecorator::redraw(const std::string& plugin, tlp::DataSet& parameters ,std::string &errMsg,tlp::PluginProgress* progress){
    bool res=false;
    std::list<std::string> plugins=layoutPluginsList();
    if(std::find(plugins.begin(),plugins.end(),plugin) == plugins.end()){
        errMsg = "The choosen plugin is not in the list";
    }else{
        if(numberOfNodes()!=1) {   //la règle a plus qu'un bridge
            Observable::holdObservers();
            res = graph_component->applyPropertyAlgorithm(plugin, graph_component->getProperty<LayoutProperty>("viewLayout"), errMsg,progress,&parameters);
            Observable::unholdObservers();
        }
        else{
            errMsg = "The rule is empty";
        }
    }
    return res;
}

std::pair<tlp::Graph *,tlp::Graph *> PortGraphRule::extractLHSRHS(Graph *g, PluginProgress *pp, string &errMsg) {
    PortGraphRuleDecorator dec(g);
    IntegerProperty *side = dec.getSideProperty();
    DataSet parameters;
    parameters.set("Property",side);
    StringCollection elementsType("nodes;edges;");
    elementsType.setCurrent("nodes");
    parameters.set("Type",elementsType);
    parameters.set("Connected",false);
    if(!g->applyAlgorithm("Equal Value",errMsg,&parameters,pp))
        return make_pair<Graph*, Graph*>(nullptr, nullptr);
    Graph *sg, *rhs = nullptr, *lhs = nullptr;
    forEach(sg, g->getSubGraphs()) {
        stringstream sstr; //use the code from EqualValueClustering.cpp to build graph name
        sstr << PorgyConstants::RULE_SIDE << ": ";
        sstr.width(8);
        sstr << PorgyConstants::SIDE_RIGHT;
        if(sg->getName()==sstr.str())
            rhs = sg;
        sstr.str(string());
        sstr.clear();
        sstr << PorgyConstants::RULE_SIDE << ": ";
        sstr.width(8);
        sstr << PorgyConstants::SIDE_LEFT;
        if(sg->getName()==sstr.str())
            lhs = sg;
    }
//    assert(lhs!=nullptr); //at least, we should have lhs
    return make_pair(lhs, rhs);
}

void PortGraphRule::cleanLHSRHS(Graph *r) {
    Graph *g;
    stableForEach(g, r->getSubGraphs()) {
        r->delSubGraph(g);
    }
}

PortGraphRule::~PortGraphRule() {
    delete bridge;
}

tlp::BooleanProperty *PortGraphRuleDecorator::getSelectionProperty() {
    return PorgyTlpGraphStructure::getRulesRoot(graph_component)->getLocalProperty<BooleanProperty>("viewSelection");
}

node PortGraphRule::findRightEquiv(const node& n) const {
    return bridge->getRight(getPortContainer(n));
}

Iterator<PortNode *> *PortGraphRule::getMember(const PorgyConstants::RuleSide side) {
    PortGraphRuleDecorator dec(_graph);
    vector<PortNode *> v;
    PortNode *pn;
    forEach(pn, PortGraph::getPortNodes()) {
        if(dec.getSide(pn->getCenter())==side)
            v.push_back(pn);
    }
    return new CopyStlIterator<PortNode *, vector<PortNode *> >(v);
}

tlp::Iterator<PortNode *> *PortGraphRule::getPortNodes(BooleanProperty *) const {
    PortGraphRuleDecorator dec(_graph);
    vector<PortNode *> v;
    PortNode *pn;
    forEach(pn, PortGraph::getPortNodes()) {
        if((dec.getSide(pn->getCenter())==PorgyConstants::SIDE_RIGHT)||(dec.getSide(pn->getCenter())==PorgyConstants::SIDE_LEFT))
            v.push_back(pn);
    }
    return new CopyStlIterator<PortNode *, vector<PortNode *> >(v);
}

Iterator<edge> *PortGraphRule::getEdges(tlp::BooleanProperty *) {
    PortGraphRuleDecorator dec(_graph);
    vector<edge> v;
    edge e;
    forEach(e, getGraph()->getEdges()) {
        if(((dec.getSide(e)==PorgyConstants::SIDE_RIGHT)||(dec.getSide(e)==PorgyConstants::SIDE_LEFT))&&(PortGraph::isPortGraphEdge(e, getGraph()))) {
            v.push_back(e);
        }
    }
    return new CopyStlIterator<edge, vector<edge> >(v);
}

Iterator<edge> *PortGraphRule::getMemberEdges(const PorgyConstants::RuleSide side) {
    PortGraphRuleDecorator dec(_graph);
    vector<edge> v;
    edge e;
    forEach(e, getGraph()->getEdges()) {
        if(side==PorgyConstants::SIDE_BRIDGE) {
            if((dec.getSide(e)==side)) { //useless?
                v.push_back(e);
            }
        }
        if((dec.getSide(e)==side)&&(PortGraph::isPortGraphEdge(e, getGraph()))) {
            v.push_back(e);
        }
    }
    return new CopyStlIterator<edge, vector<edge> >(v);
}

StringCollection PortGraphRule::getLayoutAlgorithm() {
    string layout_algo;
    getGraph()->getAttribute(PorgyConstants::LAYOUTALGORITHM,layout_algo);
    StringCollection tmp(LAYOUT_ALGO);
    for(unsigned int i=0;i<tmp.size();++i) {
        if(tmp.at(i)==layout_algo) {
            tmp.setCurrent(i);
            break;
        }
    }
    return tmp;
}

PorgyConstants::RuleSide PortGraphRuleDecorator::getSide(const tlp::node &n) {
    assert(PorgyTlpGraphStructure::isRuleGraph(graph_component));
    assert(graph_component->isElement(n));
    return static_cast<PorgyConstants::RuleSide>(getSideProperty()->getNodeValue(n));
}

PorgyConstants::RuleSide PortGraphRuleDecorator::getSide(const tlp::edge &e) {
    assert(PorgyTlpGraphStructure::isRuleGraph(graph_component));
    assert(graph_component->isElement(e));
    return static_cast<PorgyConstants::RuleSide>(getSideProperty()->getEdgeValue(e));
}

std::list<std::string> PortGraphRuleDecorator::layoutPluginsList()const{
    std::list<std::string> list;
    list.push_back(PorgyConstants::REDRAW_RULE);
    return list;
}

PortGraphRuleDecorator::PortGraphRuleDecorator(Graph *g):PortGraphDecorator(g) {

}
