/**
 *
 * This file is part of Porgy (http://tulip.labri.fr/TulipDrupal/?q=porgy)
 *
 * from LaBRI, University of Bordeaux, Inria and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef PORGYGRAPHSTRUCTURE_H
#define PORGYGRAPHSTRUCTURE_H

#include <string>
#include <random>

#include <portgraph/PorgyConstants.h>

#include <tulip/Iterator.h>
#include <tulip/StlIterator.h>
#include <tulip/StringCollection.h>

namespace tlp {
class BooleanProperty;
}

/**
 * @brief Iterate on a copy of a stl container.
 **/
template<typename VALUE, typename CONTAINER>
class CopyStlIterator: public tlp::Iterator<VALUE> {
    CONTAINER container;
    tlp::StlIterator<VALUE, typename CONTAINER::iterator> stlIterator;

public:
    CopyStlIterator(const CONTAINER &container):container(container), stlIterator(this->container.begin(), this->container.end()) {
    }

    VALUE next() {
        return stlIterator.next();
    }
    ///Return true if it exists a next element
    bool hasNext() {
        return stlIterator.hasNext();
    }
};

struct PorgyTlpGraphStructure {

    static tlp::Graph* setPorgyStructure(tlp::Graph* root = nullptr);

    static bool hide(tlp::Graph *g);

    static std::mt19937 gen;

    /**
    * @brief Transform a classic graph into a portgraph.
    *
    * @param graph A graph of a Porgy graph structure.
    * @param pluginProgress The plugin progress manager.
    * @param multiPort A boolean indicating if the generated portgraph will get multiple ports (one per link) [default - true].
    * @param mergeInOut A boolean indicating if the in- and outcoming edges will use the same port [default - true].
    **/
    static void Graph2PortGraph(tlp::Graph *graph, tlp::PluginProgress *pluginProgress, bool multiPort = true, bool mergeInOut = true);

    static bool isPorgyGraph(const tlp::Graph* graph, std::string&);
    /**
         * @brief Returns the associate root Graph (i.e. TraceRoot, ModelsRoot or RulesRoot)
         * @param graph: a graph of a Porgy graph structure
         */
    static tlp::Graph *getRootGraph(const tlp::Graph* graph);

    static bool isTraceGraph(const tlp::Graph*);

    static bool isRuleUsed(tlp::Graph* g);
    static bool isGraphUsed(tlp::Graph* g);

    static inline  tlp::Graph* getModelsRoot(const tlp::Graph* g) {
        tlp::Graph *r = g->getRoot();
        return (r==nullptr)?nullptr:r->getDescendantGraph(PorgyConstants::GRAPHS_GRAPH_NAME);
    }

    static inline bool isModelGraph(const tlp::Graph* g){
        return getModelsRoot(g)->isDescendantGraph(g);
    }

    static inline tlp::Graph* findModel(const tlp::Graph* g, const std::string &name) {
        return getModelsRoot(g)->getDescendantGraph(name);
    }

    static inline tlp::Graph* getRulesRoot(const tlp::Graph* g) {
        tlp::Graph *r = g->getRoot();
        return (r==nullptr)?nullptr:r->getDescendantGraph(PorgyConstants::RULES_GRAPH_NAME);
    }

    static inline tlp::Graph* findRule(const tlp::Graph *g, const std::string &name) {
        tlp::Graph *r = getRulesRoot(g);
        return (r==nullptr)?nullptr:r->getDescendantGraph(name);
    }

    static inline bool isRuleGraph(const tlp::Graph* g){
        tlp::Graph *r = getRulesRoot(g);
        return r!=nullptr&&r->isDescendantGraph(g);
    }

    static inline tlp::Graph* getTraceRoot(const tlp::Graph* g){
        tlp::Graph *r = g->getRoot();
        return (r==nullptr)?nullptr:r->getRoot()->getDescendantGraph(PorgyConstants::TRACE_GRAPH_NAME);
    }

    static bool checkEdgeState(const tlp::Graph *root, const tlp::edge &rule_e, const tlp::edge &graph_e);
    static bool checkEdgeState(const tlp::Graph *root, const std::vector<tlp::edge> &rule_e, const std::vector<tlp::edge> &graph_e, bool exact);

    //P property set and get
    static void setPPropertyNameAttribute(tlp::Graph *g, const std::string &name);
    static std::string getPPropertyNameAttribute(tlp::Graph *g);

    static void setBanPropertyNameAttribute(tlp::Graph *g, const std::string &name);
    static std::string getBanPropertyNameAttribute(tlp::Graph *g);

    //rulename set and get
    static void setRuleNameForSubGraph(tlp::Graph *g, const std::string &);
    static std::string getRuleNameForSubGraph(tlp::Graph *g);

    /**
         * @brief Find the main trace graph in the given graph hierachy and return it.
  *
  * @param graph A graph of a Porgy graph structure.
         * @return A tlp::Graph * if the traces graph is found.
  **/
    static tlp::Graph *getMainTrace(const tlp::Graph *graph);

    static tlp::Graph * clone(tlp::Graph * old_graph, std::string pName, std::string BanName);

    static bool graphIsASimplePath(tlp::Graph* graph);

    static void getMatchingPropertyNames(tlp::Graph *g, std::unordered_set<std::string> &propertiesNames);

    static void getMatchingPropertyNamesExtended(tlp::Graph *g, std::unordered_set<std::string> &propertiesNames);

    static void setMatchingPropertyNames(tlp::Graph *g, std::unordered_set<std::string> &propertiesNames);

    static void getStandardPropertyNames(tlp::Graph *g, std::vector<std::string> &propertiesNames);

};
#endif // PORGYGRAPHSTRUCTURE_H
